/*  Krister Swenson,Mark Marron                                     Fall 2002
The HP graph along with the overlap graph matrix of Anne Bergeron's paper 
"A very elementary presentation of the hannenhalli-pevzner theory".
*/

#include "Ograph.h"

using namespace std; 
using namespace __gnu_cxx; 



//Build the identity graph of length len.
Ograph::Ograph(int len)
  {
  if(len < 1)
    len = 1;
  int *temp = makeID(len);
  construct(temp, len);
  free(temp);
  }


/* Build the graph as though we were given the sequence "1".
*/
Ograph::Ograph()
  {
  int v[1];
  v[0] = 1;
  construct(v, 1, false);
  }

/* Create a graph that represents the given -sequence-.
  ** If given an invalid sequence then just build the graph as though we were
     given the sequence "1".
*/
Ograph::Ograph(const int sequence[], const int size, const bool invert)
  { construct(sequence, size, invert); }
Ograph::Ograph(const vector<int> sequence, const bool invert)
  {
  int temp[sequence.size()];
  for(unsigned i=0; i < sequence.size(); i++)
    temp[i] = sequence[i];
  construct(temp, sequence.size(), invert);
  }

void Ograph::construct(const int sequence[], const int size, const bool inv)
  {
  if(validPerm(sequence, size))  //If we're given a valid permutation
    {                            //then set up the sequence:
    seqsize = size + 2;
    graphsize = size*2 + 2;
    matsize = graphsize/2;


    seq = new int[seqsize];     //Set up the sequence:
    seqinds = new int[seqsize];
    seq[0] = 0;
    seqinds[0] = 0;
                  //We've removed Joel's invert sequence thing.  Hopefully his
                  //sorting sequences still turn out OK (just inverted).
    if(inv)                     //If we want to build the inveted sequence:
      {
      int gene;
      for(int i=1; i <= size; i++)//Invert sequence so seq is genome 1 from an
        {                         //identity-permutation-numbered genome 2. Then
        gene = sequence[i-1];     // the reversal sequence will convert genome 1
        if (gene >=0)             // into genome 2.
          {                       //  `--Joel Added this, not sure why?
          seq[gene] = i;
          seqinds[i] = gene;
          }
        else
          {
          seq[-gene] = -i;
          seqinds[i] = -gene;
          }
        }
      }
    else
      {
      for(int i=1; i <= size; i++)
        {
        seq[i] = sequence[i-1];
        seqinds[abs(seq[i])] = i;
        }
      }
    seq[seqsize-1] = seqsize-1;
    seqinds[seqsize-1] = seqsize-1;
    }
  else                           //Otherwise, just make the sequence be "0 1 2":
    {
    cerr << "(Ograph)Warning: Invalid sequence given. Creating default.\n";
    seqsize = 3;
    graphsize = 4;
    matsize = 2;
    seq = new int[3];
    seqinds = new int[3];
    seq[0]=0;seq[1]=1;seq[2]=2;
    seqinds[0]=0;seqinds[1]=1;seqinds[2]=2;
    }


  gr = (struct Node*)malloc(graphsize * sizeof(struct Node));
  findGraphArray();
  findOverlapMat();
  comps = 0;           //Initialize.
  updateCompInfo();    //Find the cycles(and components) in the initial graph.
  }

/* Copy constructor.
*/
Ograph::Ograph(const Ograph& g)
  { copy(g); }

/* Destructor
*/
Ograph::~Ograph()
  { del(); }

/* Delete the memory for this.
*/
void Ograph::del()
  {
  free(gr);
  
  delete comps;
  
  for(int i=0; i < matsize; i++)
    delete [] matrix[i];      //Delete the rows.
  free(matrix);               //Free the columns.

  delete [] p;
  delete [] s;

  delete [] seq;
  delete [] seqinds;
  }


/* Assignment Operator.
*/
void Ograph::operator=(const Ograph& g)
  { del(); copy(g); }



/*Return true if the given graph is equivalent to -this-.
*/
bool Ograph::operator==(const Ograph& g) const
  {
  if(seqsize != g.seqsize)
    return false;

  for(int i=0; i<seqsize; i++)
    if(seq[i] != g.seq[i])
      return false;

  return true;
  }



/* Copy the memory from the object.
*/
void Ograph::copy(const Ograph& g)
  {
  graphsize = g.graphsize;
  gr = (struct Node*)malloc(graphsize * sizeof(struct Node));
  for(int i=0; i < graphsize; i++)
    {
    gr[i].sign = g.gr[i].sign;
    gr[i].reality = g.gr[i].reality;
    gr[i].desire = g.gr[i].desire;
    gr[i].label = g.gr[i].label;
    }

  comps = new CompInfo(g.graphsize);
  *comps = *(g.comps);

  seqsize = g.seqsize;
  seq = new int[seqsize];
  seqinds = new int[seqsize];
  for(int i=0; i < seqsize; i++)
    {
    seq[i] = g.seq[i];
    seqinds[i] = g.seqinds[i];
    }

  matsize = g.matsize;
  matrix = (bool **)malloc(matsize*sizeof(bool*));
  for(int i=0; i < matsize; i++)
    {
    matrix[i] = new bool[matsize];
    for(int j=0; j < matsize; j++)
      matrix[i][j] = g.matrix[i][j];
    }

  p = new bool[matsize];
  s = new int[matsize];
  for(int i=0; i < matsize; i++)
    {
    p[i] = g.p[i];
    s[i] = g.s[i];
    }

  inRev = g.inRev;

  nintcon = g.nintcon;
  cTOi.resize(g.cTOi.size());
  std::copy(g.cTOi.begin(), g.cTOi.end(), cTOi.begin());
  }

int Ograph::getGrSize() const
  { return graphsize; }

int Ograph::getMatSize() const
  { return matsize; }


/* Return true if the sequence is not sorted.
*/
bool Ograph::notSorted() const
  { return  (comps->getNumCycles() != seqsize-1); }


/*Return true if this is a positive permutation.
*/
bool Ograph::isPositive() const
  {
  for(int i=0; i<seqsize; i++)
     {
     if(seq[i] < 0)
       return false;
     }

  return true;
  }


/* Fill -gr- according to the contents of -seq-.
*/
void Ograph::findGraphArray() {
  int i;
  int size = seqsize - 2;
  vector<int> 
    nodeLoc(graphsize);	      //Location of node i in the graph:0,-1,1,-2,2... 

  nodeLoc[0] = 0;             //Initialize the frame(two added nodes):
  nodeLoc[graphsize-1] = graphsize-1;

  gr[0].sign = 0;             //sign is meaningless for the first node.
  gr[0].label = 0;            //my addition
  gr[0].reality.id = 1;
  for(i=1; i <= size; i++)
    {
    if(seq[i] > 0)
      {
      nodeLoc[abs(seq[i]*2)] = 2*i;
      nodeLoc[abs(seq[i]*2)-1] = (2*i)-1;

      gr[(2*i)-1].sign = NEG;
      gr[2*i].sign = POS;
      }
    else
      {
      nodeLoc[abs(seq[i]*2)-1] = 2*i;
      nodeLoc[abs(seq[i]*2)] = (2*i)-1;

      gr[(2*i)-1].sign = POS;
      gr[2*i].sign = NEG;
      }

    //gr[(2*i)-1].label = -sequence[i-1];   //THIS IS SetubalMed CONVENTION:
    //gr[2*i].label = sequence[i-1];
    if(seq[i] < 0)		           //THIS IS Bergeron CONVENTION:
      {
      gr[(2*i)-1].label = -2*seq[i];
      gr[2*i].label = (-2*seq[i])-1;
      }
    else
      {
      gr[(2*i)-1].label = (2*seq[i])-1;
      gr[2*i].label = 2*seq[i];
      }

    gr[(2*i)-1].reality.id = 2*(i-1); 
    gr[2*i].reality.id = (2*(i+1))-1; 
    }
  gr[graphsize-1].sign = 0;   //sign is meaningless for the last node.
  gr[graphsize-1].reality.id = graphsize-2;
  gr[graphsize-1].label = graphsize-1;


  gr[0].desire.id = nodeLoc[1];                      //The first.
  gr[graphsize-1].desire.id = nodeLoc[graphsize-2];  //The last.
  for(i=1; i <= size; i++)
    {
    gr[nodeLoc[(i*2)-1]].desire.id = nodeLoc[(i*2)-2];
    gr[nodeLoc[i*2]].desire.id = nodeLoc[(i*2)+1];
    }
}


// Different values for -direction- in findCycles() :
#define FORWARDS 1
#define BACKWARDS 0
#define NEITHER 2
// Different values for -nextedge- in findCycles() :
#define REALITY 0
#define DESIRE 1

/* Fill -comps- with information for all the cycles of the graph including
the cycle numbers as well as a mark on the cycles that are good (oriented).
Mark the component info as being stale (i.e. comps->classCycles() must be 
called to update the component info if desired).
*/
void Ograph::findCycles()
  {
			      //True for nodes that are not yet part of a cycle.
  vector<bool> unchecked(graphsize);
  for(int i=0; i < graphsize; i++)          //Init:
    unchecked[i] = true;

  int start = 0;              //The spot to start looking for a cycle at.
  int cyclenum = 0;           //The current cycle we're looking at.
  int nextedge = REALITY;     //The next edge-type to be checked.
  if(comps)
    delete comps;
  comps = new CompInfo(graphsize);
  while(start < graphsize)
    {
    int directn = NEITHER;    //Cycle direction induced by the current edge.
    int current = start;      //We're looking at the edge coming from this node.
    while(unchecked[current])
      {                                       //Add this node to the cycle.
      comps->addToCycle(current, abs(seq[(int)ceil(current/2.0)]), cyclenum);
      unchecked[current] = false;             //Mark this node as checked.

      if(nextedge == REALITY)      //For reality edges we need to
        {                          //worry about the goodness of the cycle:
        if(directn == NEITHER)                //This is the first reality edge
          {
          if(gr[current].reality.id > current)// so we set the direction:
            {
            gr[gr[current].reality.id].reality.direction =
            gr[current].reality.direction = (directn = FORWARDS);
            }
          else
            {
            gr[gr[current].reality.id].reality.direction =
            gr[current].reality.direction = (directn = BACKWARDS);
            }
          }
        else if(directn == FORWARDS)          //If cycle direction is forwards
          {
          if(gr[current].reality.id < current)// and current edge is backwards
            {
            comps->setGood(cyclenum);         // we know the cycle is good.
            gr[gr[current].reality.id].reality.direction =
            gr[current].reality.direction = BACKWARDS;
            }
          else
            {
            gr[gr[current].reality.id].reality.direction =
            gr[current].reality.direction = FORWARDS;
            }
          }
        else if(directn == BACKWARDS)         //The cycle direction is backwards
          {
          if(gr[current].reality.id > current)// and current edge goes forward
            {
            comps->setGood(cyclenum);         // so we know the cycle is good.
            gr[gr[current].reality.id].reality.direction =
            gr[current].reality.direction = FORWARDS;
            }
          else
            {
            gr[gr[current].reality.id].reality.direction =
            gr[current].reality.direction = BACKWARDS;
            }
          }
        else
          cerr << "\n++++ We have a bad direction value\n";

        nextedge = DESIRE;         //The next edge will be a desire edge.
        current = gr[current].reality.id;     //Update current to the next node.
        }
      else                         //For desire eges,
        {
        nextedge = REALITY;              //the next edge will be a reality edge.
        current = gr[current].desire.id; //Update -current- to the next node.
        }
      }
    cyclenum++;
                                   //Find the next start spot:
    while(!unchecked[start] && (++start < graphsize));
    }
  comps->markStale();
  }

/*Return the number of hurdles over all components.
  ** Precondition: the sequence is a positive permutation!
*/
int Ograph::getNumHurdles() const
  { return comps->getNumHurdles(); }

/*Return the number of cycles over all components.
*/
int Ograph::getNumCycles() const
  { return comps->getNumCycles(); }


/* Mark the given values as deletions.  Return the number of operations
needed to delete them.
** For now just mark the edge as containing at least one deletion.
   `-> This is a simplified approximation of El-Mabrook's method.
       (if hurdles exist, it's not completely accurate [0 or 1 del per merge])
*/
int Ograph::markDels(stack<Pair> &dels)
  {
  while(!dels.empty())
    {
    //cerr << "DEL:" << dels.top() << " | " << 2*dels.top().getRight() << endl;
    //cerr << "  seqinds:" << gr[seqinds[abs(dels.top().getRight())]*2].label
    //     << endl;
    int node = abs(dels.top().getRight());
    gr[seqinds[node]*2].reality.deletion = dels.top().getLeft();
    comps->setCycleDelVal(node*2);
    dels.pop();
    }

  int numdels = 0;                 //For each cycle in the graph,
  for(int i=0; i < comps->getNumCycles(); i++)
    if(comps->getCycleDelVal(i))   //if the cycle has a deletion, then consider
      numdels++;                   //all the dels in the cycle to be 1.

  return numdels;
  }



/* Return a copy of the sequence without the frame.
Caller is responsible for the memory.
*/
int* Ograph::getSeq() const
  {
  int i;
  int* ret = (int*)calloc(seqsize-2, sizeof(int));

  for(i=0; i<seqsize-2; i++)
    ret[i] = seq[i+1];

  return ret;
  }
vector<int> Ograph::getSeqVec() const
  {
  vector<int> ret;

  for(int i=0; i<seqsize-2; i++)
    ret.push_back(seq[i+1]);

  return ret;
  }

/*Return an array of the sequence (without the frame) named according to
the given sequence.  The given sequence must be the length of this sequence
without the frame elements 0 and n-1.
  ** Caller responsible for the memory.
*/
int* Ograph::getRenSeq(const int s2[], int s2len) const
  {
  int *s1 = &seq[1];         //Make s1 so that it points to the second element
                             // of seq (ignor the 0 at the beginining).
  int s1len = seqsize-2;     //The length of seq without the frame elements.
  int* ret = (int*)calloc(s1len, sizeof(int));

  assert(s1len == s2len);    //Ensure the sequence lengths match.

  int index[s2len+1];        //index[i] is the index where i is located in s2.
                             // (so should be the value that i becomes in s1)
  for(int i=0; i<s2len; i++)     //Build the index:
    {
    if(s2[i] < 0)                //If the number is negative
      index[abs(s2[i])] = -(i+1);//then store its index as negative
    else                         //otherwise,
      index[s2[i]] = (i+1);      //store it as positive.
    }

  for(int j=0; j<s1len; j++)     //Use the index.
    {
    if(s1[j]>0)
      ret[j] = index[s1[j]];
    else
      ret[j] = -index[abs(s1[j])];
    }

  return ret;
  }


/* Return the sequence length without the frame elements (0 and n+1).
This is the size of the array returned in getSeq().
*/
int Ograph::getSeqLen() const
  { return seqsize-2; }

/* Return a character array of the sequence without the frame.
Caller is responsible for the memory.
*/
char* Ograph::getSeqStr() const
  {
  string s = getSeqString();
  char* ret = (char*)calloc(s.length()+1, sizeof(char));
  s.copy(ret, s.length());
  ret[s.length()] = '\0';

  return ret;
  }

/* Return a string of the sequence without the frame.
*/
string Ograph::getSeqString() const
  {
  string ret = ToString(seq[1]);

  for(int i=2; i<seqsize-1; i++)
    ret += " " + ToString(seq[i]);

  return ret;
  }


/*Return a string of the sequence (without the frame) named according to
the given sequence.  The given sequence must be the length of this sequence
without the frame elements (0, n-1).
  ** Caller responsible for the memory.
*/
char* Ograph::getRenStr(const int s2[], int s2len) const
  {
  int *s1 = &seq[1];         //Make s1 so that it points to the second element
                             // of seq (ignor the 0 at the beginining).
  int s1len = seqsize-2;     //The length of seq without the frame elements.

  assert(s1len == s2len);    //Ensure the sequence lengths match.

  int index[s2len];          //index[i] is the index where i is located in s2.
                             // (so should be the value that i becomes in s1)
  for(int i=0; i<s2len; i++)     //Build the index:
    {
    if(s2[i] < 0)                //If the number is negative
      index[abs(s2[i])] = -(i+1);//then store its index as negative
    else                         //otherwise,
      index[s2[i]] = i+1;        //store it as positive.
    }

  char* ret = (char*)calloc(seqsize*3, sizeof(char));
  ret[0] = '\0';
  for(int j=0; j<s1len; j++)     //Use the index.
    {
    if(s1[j]>0)                  //If this element is positive then the sign
      sprintf(ret, "%s %i", ret, index[abs(s1[j])]);  //depends on the index.
    else                         //otherwise,
      sprintf(ret, "%s %i", ret, -index[abs(s1[j])]); //it's the inverse.
    }

  return ret;
  }

/* Return a string of the inverse of the sequence (without the frame).
  ** Caller responsible for the memory.
*/
char* Ograph::getInvStr() const
  {
  int ts[seqsize-2];

  for(int i=0; i<seqsize-2; i++)
    {
    if(seq[i+1] > 0)
      ts[abs(seq[i+1])-1] = i+1;
    else
      ts[abs(seq[i+1])-1] = -(i+1);
    }


  char* ret = (char*)calloc(seqsize*3, sizeof(char));
  ret[0] = '\0';
  for(int i=0; i<seqsize-2; i++)
    sprintf(ret, "%s %i", ret, ts[i]);  //depends on the index.

  return ret;
  }

/* Return the inverse graph of the current one.
*/
Ograph Ograph::getInvGraph() const
  {
  int inv[seqsize-2];
  for(int i=1; i<seqsize-1; i++)  //Copy seq to inv.
    inv[i-1] = seq[i];
  invertSeq(inv, seqsize-2);      //Invert inv.

  Ograph ret(inv, seqsize-2);     //Build the graph
  return ret;                     // that we return.
  }


/* Return the inversion distance WITHOUT CONSIDERING HURDLES from this sequence
to the identity.
  ** No inversions are actually done (constant time op).
*/
int Ograph::getDistNH() const 
  {
  return (seqsize-1) - getNumCycles();
  }

/* Return the inversion distance from this sequence to the identity.
  ** No inversions are actually done.
  ** THIS runs resolve() on a copy of the graph BECAUSE findHurdles
     (in CompInfo) only works on positive permutations.
*/
int Ograph::getDist() const 
  {
  Ograph tmp = *this;
  vector<Pair> garbage;
  int dist = tmp.resolve(garbage);

  // return ((seqsize-1) - getNumCycles()) + ((getNumHurdles()+1)/2);
  return dist;
  }


/* Do all reversals necessary to make this graph into a bunch of two node
cycles, including hurdle cutting and merging.  Return the inversion distance.
Fill the given vector with the inversions done.
*/
int Ograph::resolve(vector<Pair> &revs)
  {
  int ret=0;

  if(comps->getNumCycles() == seqsize-1)
    return ret;

  while(!allZeroParity())	//Do reversals until we hit a hurdle:
    {
    revs.push_back(doRev());
    ++ret;
    }

  if(!allZeroScore())		//If there are hurdles,
    {				//resolve them:
    updateCompInfo();
    ret+=resolveHurdles(revs);	//Get rid of hurdles. (Joel)

    while(!allZeroParity())	//do the remainder of the reversals:
      {
      revs.push_back(doRev());
      ++ret;
      }

    if(!allZeroScore())	     //If there are still hurdles, there is a problem.
      {
      cerr << endl << "++++ERROR: Problem with Handling Hurdles!" << endl;
      cout << endl << "++++ERROR: Problem with Handling Hurdles!" << endl;
      }
    }

  return ret;
  }


/* Update the information about the cycles and components and -gr-.
*/
void Ograph::updateCompInfo()
{
  findGraphArray();
  findCycles();
  comps->classCycles();
  comps->findHurdles();
}

/* Return a lists of all pairs of reality edges that are on different cycles.
*/
void Ograph::getUnsortingInvs(vector<Pair> &ops) const
  {
  Pair p;

  for(int i=1; i < graphsize; i+=2)     //For each reality edge
    for(int j=i+2; j < graphsize; j+=2) //compare it to each r edge after it:
      if(comps->getCycleNum(i) != comps->getCycleNum(j))
        {                               //If the edges are from different cycles
        p = Pair((i+1)/2, j/2);
        ops.push_back(p);               //then add the pair.
        }
  }



/* Return a vector of all pairs of divergent reality edges that represent safe
inversions.  These pairs indicate the inversions that can be done to decrease
the inversion distance.
*/
int Ograph::getSafeInvs(vector<Pair> &ops) const
  {
  Pair p;

  for(int i=1; i < graphsize; i+=2)
    {
    if(gr[i].reality.direction == FORWARDS)       //If this edge is FORWARDS
      {
      for(int j=i+2; j < graphsize; j+=2)         //for everyone on this cycle
        if(comps->getCycleNum(j) == comps->getCycleNum(i) &&
           gr[j].reality.direction == BACKWARDS)  //that is backwards
          {
          p = Pair((i+1)/2, j/2);
          if(isSafe(p))
            ops.push_back(p);                     //add the pair.
          }
      }
    else
      {
      for(int j=i+2; j < graphsize; j+=2)
        if(comps->getCycleNum(j) == comps->getCycleNum(i) &&
           gr[j].reality.direction == FORWARDS)
          {
          p = Pair((i+1)/2, j/2);
          if(isSafe(p))
            ops.push_back(p);                     //add the pair.
          }
      }
    }

  return (int)ops.size();
  }

/* Return a vector of all pairs of divergent reality edges that represent 
cycle-splitting inversions.
**This isn't the complete list of sorting inversion if there are hurdles around.
*/
int Ograph::getCSInvs(vector<Pair> &ops) const
  {
  Pair p;

  for(int i=1; i < graphsize; i+=2)
    {
    if(gr[i].reality.direction == FORWARDS)       //If this edge is FORWARDS
      {
      for(int j=i+2; j < graphsize; j+=2)         //for everyone on this cycle
        if(comps->getCycleNum(j) == comps->getCycleNum(i) &&
           gr[j].reality.direction == BACKWARDS)  //that is backwards
          {
          p = Pair((i+1)/2, j/2);
          ops.push_back(p);                       //add the pair.
          }
      }
    else
      {
      for(int j=i+2; j < graphsize; j+=2)
        if(comps->getCycleNum(j) == comps->getCycleNum(i) &&
           gr[j].reality.direction == FORWARDS)
          {
          p = Pair((i+1)/2, j/2);
          ops.push_back(p);
          }
      }
    }

  return (int)ops.size();
  }


/* Return a vector of all pairs of divergent reality edges (with one of the
endpoints being that which is given) that represent cycle splitting
inversions.  These pairs indicate the inversions that can be done to decrease
the inversion distance.  This isn't the complete list of sorting inversion if
there are hurdles around.
*/
void Ograph::getCSInvs(vector<Pair> &ops, const int endpoint) const
  {
  Pair p;
  int i=endpoint*2+1;

  assert(i < graphsize);

  if(gr[i].reality.direction == FORWARDS)       //If this edge is FORWARDS
    {
    for(int j=1; j < graphsize; j+=2)           //for everyone on this cycle
      if(comps->getCycleNum(j) == comps->getCycleNum(i) &&
         gr[j].reality.direction == BACKWARDS)  //that is backwards
        {
        if(i<j)
          p = Pair((i+1)/2, j/2);
        else
          p = Pair((j+1)/2, i/2);
        ops.push_back(p);                       //add the pair.
        }
    }
  else                                          //Otherwise it's BACKWARDS...
    {
    for(int j=1; j < graphsize; j+=2)
      if(comps->getCycleNum(j) == comps->getCycleNum(i) &&
         gr[j].reality.direction == FORWARDS)
        {
        if(i<j)
          p = Pair((i+1)/2, j/2);
        else
          p = Pair((j+1)/2, i/2);
        ops.push_back(p);
        }
    }
  } 


/*Fill a vector with all oriented (adjacency creating) inversions.
Return the size of the vector.
  ** Unsafe inversions are included in this list.
*/
int Ograph::getOrientedInvs(vector<Pair> &ops) const
  {
  for(int i=0; i < graphsize; i++)
    {
    int other = gr[i].desire.id;
    if((i < other) &&                             //If we're at left endpoint,
       ((i+other)%2 == 0) &&                      //the edge is oriented,
       !((i%2) && (gr[i-1].desire.id == other-1)))//and it's not a duplicate,
      ops.push_back(Pair((i/2)+1,other/2));       //then add the inversion.
    }

  return (int)ops.size();
  }



/*Return the ratio of right-facing reality edges to the size of the cycle for 
the largest cycle.
*/
float Ograph::getOrientationBalanceForLargest() const
  {
  int count=0;
  vector<int> sizes = comps->getCycleSizes();
  int size = sizes[0]*2;
  int cyclenum = -1;
  for(int i=0; i<graphsize; i++)
    {
    if(cyclenum < 0 && comps->getCycleSize(i) == size)
      cyclenum = comps->getCycleNum(i);
    if(comps->getCycleNum(i) == cyclenum && gr[i].reality.direction == FORWARDS)
      count++;
    }

  return (float)((float)count/(float)size);
  }



/*Return a vector with an entry per (large enough) cycle.  Each entry contains
the ratio of right-facing reality edges to the size of the cycle.
*/
vector<float> Ograph::getOrientationBalance() const
  {
  vector<int> count;
  vector<int> sizes;
  count.resize(comps->getNumCycles());
  sizes.resize(comps->getNumCycles());
  for(int i=0; i<graphsize; i++)
    {
    if(gr[i].reality.direction == FORWARDS)
      {
      count[comps->getCycleNum(i)]++;
      sizes[comps->getCycleNum(i)] = comps->getCycleSize(i);
      }
    }

/*
  vector<int> m;
  m.resize(count.size());
  for(unsigned int i=0; i < count.size(); i++)
    {
    int largest=0;
    int index=i;
    unsigned int j=i;
    while(j < count.size())  //Find the next largest:
      {
      if(largest < count[j])
        {
        largest = count[j];
        index = j;
        }
      j++;
      }
    m[i] = j;
    }
  for(unsigned int i=0; i < sizes.size(); i++)
    cerr << sizes[m[i]] << endl;
*/


  vector<float> vf;
  for(unsigned int i=0; i < count.size(); i++)
    if(sizes[i] > graphsize/20)
{ cerr << count[i] << " divide by " << sizes[i] << endl;
      vf.push_back((float)count[i]/(float)sizes[i]);
}


  return vf;
  }


/*Return the length of the largest cycle.
*/
int Ograph::getLargestCycleSize() const
  { return comps->getLargestCycleSize(); }


/*Return the length of the cycles, sorted from largest to smallest.
*/
vector<int> Ograph::getCycleSizes() const
  { return comps->getCycleSizes(); }




/* Return a vector of all pairs of non-divergent reality edges from the same
cycle.  These pairs indicate the inversions that can be done that have no
effect on the inversion distance.
*/
void Ograph::getBenignInvs(vector<Pair> &ops) const
  {
  Pair p;

  for(int i=1; i < graphsize; i+=2)
    {
    if(gr[i].reality.direction == FORWARDS)       //If this edge is FORWARDS
      {
      for(int j=i+2; j < graphsize; j+=2)         //for everyone on this cycle
        if(comps->getCycleNum(j) == comps->getCycleNum(i) &&
           gr[j].reality.direction == FORWARDS)   //that is also FORWARDS
          {
          p = Pair((i+1)/2, j/2);
          ops.push_back(p);                       //add the pair.
          }
      }
    else
      {
      for(int j=i+2; j < graphsize; j+=2)
        if(comps->getCycleNum(j) == comps->getCycleNum(i) &&
           gr[j].reality.direction == BACKWARDS)
          {
          p = Pair((i+1)/2, j/2);
          ops.push_back(p);
          }
      }
    }
  }







/*Give a maximum independent set of commuting inversions in -mis-.  The size
of mis is returned.
*/
list<Pair> Ograph::getCommutingMIS() const
  {
  list<Pair> mis;

        //Take the edge orientation information from the graph:
  int *forwards = new int[seqsize];    //The forward facing reality edges.
  int fsize=0;
  int *backwards = new int[seqsize];   //The backward facing reality edges.
  int bsize=0;
  for(int i=0; i < seqsize-1; i++)
    {
    if(gr[2*i].reality.direction == FORWARDS)
      forwards[fsize++] = i;
    else
      backwards[bsize++] = i;
    }
  if(bsize == 0)                       //If there are no oriented inversions,
    return mis;                        //then return an empty set.

       //Build the chord model (indexed starting at 0):
  int csize = 0;
  vector<int> cmodel;                    //The chord (interval) model.
  cmodel.resize(fsize*bsize*2);
  Pair *intTOinv = new Pair[fsize*bsize];//Map interval labels to inversions.

  int msize;
  if(fsize > bsize)
    msize = fsize;
  else
    msize = bsize;
  int notused=0;                       //Index larger than any so far.
                                       
  int **mapstacks = new int*[seqsize-1];
  for(int i=0; i<seqsize-1; i++)       //One stack per edge.
    mapstacks[i] = new int[msize];
  int *mapinds = new int[seqsize-1];   //The size of each stack.
  for(int i=0; i < seqsize-1; i++)
    mapinds[i] = 0;

                              //Build the chord model as an adjacency list:
  int find=0, bind=0;
  for(int i=bsize-1; i>=0; i--)        //For BACKWARD edges right of the first
    if(comps->getCycleNum(backwards[i]*2) == comps->getCycleNum(1))
      {                                //that are on the same cycle, 
      cmodel[csize++] = notused;       //store the new left endpoint
      mapstacks[0][mapinds[0]++] = notused++;
      }                                //and stack it on this edge.
  find++;
  while(find < fsize || bind < bsize)  //While we have some reality edges left:
    {
    if(bind >= bsize || (find < fsize && forwards[find] < backwards[bind]))
      {                                //If we're looking at a FORWARD edge.
      int cnum = comps->getCycleNum(forwards[find]*2);
      for(int i=bind-1; i >= 0; i--)   //For each BACKWARD edge to the left
        {                              //that is on the same cycle,
        if(comps->getCycleNum(backwards[i]*2) == cnum)
          {                            //store the right endpoint
          cmodel[csize++] = mapstacks[backwards[i]][--mapinds[backwards[i]]];
          intTOinv[cmodel[csize-1]].set(backwards[i]+1, forwards[find]);
          }                            //and save the interval to inversion map.
        }
      for(int i=bsize-1; i>=bind; i--) //For each BACKWARD edge to the right
        if(comps->getCycleNum(backwards[i]*2) == cnum)
          {                            //that is on our cycle,
          cmodel[csize++] = notused;   //store the new left endpoint
          mapstacks[forwards[find]][mapinds[forwards[find]]++] = notused++;
          }                            //and stack it on this edge.
      find++;
      }
    else                               //Otherwise, we're at a BACKWARD edge.
      {
      int cnum = comps->getCycleNum(backwards[bind]*2);
      for(int i=find-1; i >= 0; i--)   //For each FORWARD edge to the left
        {                              //that is on the same cycle,
        if(comps->getCycleNum(forwards[i]*2) == cnum)
          {                            //store the right endpoint
          cmodel[csize++] = mapstacks[forwards[i]][--mapinds[forwards[i]]];
          intTOinv[cmodel[csize-1]].set(forwards[i]+1, backwards[bind]);
          }                            //and save the interval to inversion map.
        }
      for(int i=fsize-1; i>=find; i--) //For each FORWARD edge to the right
        if(comps->getCycleNum(forwards[i]*2) == cnum)
          {                            //that is on our cycle,
          cmodel[csize++] = notused;   //store the new left endpoint
          mapstacks[backwards[bind]][mapinds[backwards[bind]]++] = notused++;
          }                            //and stack it on this edge.
      bind++;
      }
    }
  cmodel.resize(csize);

       //Get the MIS:
  list<int> misintervals;
  getMISCG(cmodel, misintervals);
  for(list<int>::iterator i=misintervals.begin(); i != misintervals.end(); i++)
    mis.push_back(intTOinv[*i]);


/*  
for(int i=0; i<csize; i++)  //Exercise all rotations of the chord model:
  {
  int tmp = cmodel[0];
  for(int j=0; j<csize-1; j++)
    cmodel[j] = cmodel[j+1];
  cmodel[csize-1] = tmp;
  
  //for(int k=0; k<csize; k++)
  //  cerr << cmodel[k] << " ";
  //cerr << endl;
  getMISCG(cmodel, misintervals);
  }
*/

  delete [] intTOinv;
  delete [] forwards;
  delete [] backwards;
  delete [] mapinds;
  for(int i=0; i<seqsize-1; i++)
    delete [] mapstacks[i];
  delete [] mapstacks;

  return mis;
  }





/*Print the traces by getting one word at a time.
 *Do it by only visiting prefixes that correspond to appending scenarios.
 */
void Ograph::printAppendingTraces(ostream &out)
  {
  vector<Trace> traces = getAppendingTraces();
  std::copy(traces.begin(), traces.end(), ostream_iterator<Trace>(out, "\n"));
  }


/*Return all traces corresponding to this permutation.
 */
vector<Trace> Ograph::getAppendingTraces()
  {
  vector<Trace> retval;
  Trace t;
  getAppendingTraces(retval, t);
  return retval;
  }

/*Put in -retval- all traces corresponding to this permutation.
 */
void Ograph::getAppendingTraces(vector<Trace>& retval, Trace& t)
  {
  if(getDistNH() == 0)       //If we're sorted,
    {
    retval.push_back(t);     //then record the trace and
    return;                  //return.
    }
                             //Get the list of next reversals.
  list<Pair> nextrevs = getAppendingSet(t);
  //cerr << getSeqString() << endl;  //Print the permutation we're visiting.

  for(list<Pair>::iterator r=nextrevs.begin(); r!=nextrevs.end(); r++)
    {
    revAll(*r);
    t.addElement(invTOset(*r));
    getAppendingTraces(retval, t);
    t.removeLastElement();
    revAll(*r);
    }
  }




/*Recursively print the traces.
*/
/*
void Ograph::printTraces(vector< list<Element> > &nf)
  {
  if(getDistNH() == 0)
    {
    printNF(cerr, nf);
    cerr << endl;
    return;
    }

    //Find a maximal independant set of noninterfering inversions.
  list< list<Pair> > nmislist = getMalWords();
  for(list< list<Pair> >::iterator is=nmislist.begin();is!=nmislist.end();is++)
    {
    if(isNonint(*is))
      {
      Ograph g = *this;
      g.resetMalISinfo();
      g.applyCommutingSet(*is, true);
      nf.push_back(invsTOsets(*is));
      g.printTraces(nf);
      nf.pop_back();
      }
    else
      {
      cerr << "skipping:";
      printInvSet(cerr, invsTOsets(*is));
      cerr << endl;
      }
    }
  }
*/

/* Reset the datastructures that hold info about maximal independent sets.
*/
void Ograph::resetMalISinfo()
  { nintcon = Graph(); cTOi.clear(); }

/* Print the given normal form.
*/
void Ograph::printNF(ostream& out, const vector< list<Element> > nf)
  {
  vector< list<Element> >::const_iterator i = nf.begin();
  printInvSet(out, *i++);
  while(i != nf.end())
    {
    out << " < ";
    printInvSet(out, *i++);
    }
  }


/*Print the traces that correspond to this permutation.
 */
inline bool char_greater(const char c1, const char c2)
  {
  if(c1 == c2)
    return false;
  if(c1 == '}')
    return false;
  if(c2 == '}')
    return true;

  return c1 > c2;
  }
inline bool string_greater(const string &s1, const string &s2)
  {
  if(s1.size() > s2.size())        //If s1 is longer
    {
    for(unsigned i=0; i<s2.size(); i++)    //then check each character:
      {
      if(char_greater(s1[i], s2[i]))       //If s1 is larger
        return false;                      //then return.
      else if(char_greater(s2[i], s1[i]))  //If s2 is larger
        return true;                       //then return.
      }
    return false;                          //s2 is a prefix of s1.
    }
  else                                     //Otherwise, s2 is longer.
    {
    for(unsigned i=0; i<s1.size(); i++)    //Check each character:
      {
      if(char_greater(s1[i], s2[i]))       //If s1 is larger
        return false;                      //then return.
      else if(char_greater(s2[i], s1[i]))  //If s2 is larger
        return true;                       //then return.
      }
    return true;                           //s1 is a prefix of s2.
    }
  }


/*Print the traces starting from the inverse permutation by getting one word
 * at a time.
 */
void Ograph::printFromInvWordTraces(ostream &out)
  {
  Ograph inverse = getInvGraph();             //Get the inverse permutation.
  Trace empty;
  vector<Trace> traces;

  inverse.getWordTraces(empty, traces);       //Get the traces for the inverse.

  vector<int> perm = getSeqVec();
  for(vector<Trace>::iterator i=traces.begin(); i!=traces.end(); i++)
    i->rename(perm);                          //Invert and remap the traces.

  std::copy(traces.begin(), traces.end(), ostream_iterator<Trace>(out, "\n"));
  }


/*Print the traces by getting one word at a time.
 */
void Ograph::printWordTraces(ostream &out)
  {
  Trace empty;
  vector<Trace> traces;
  getWordTraces(empty, traces);
  vector<string> tstrings;
  std::copy(traces.begin(), traces.end(), ostream_iterator<Trace>(out, "\n"));
/*
  for(vector<Trace>::const_iterator i=traces.begin(); i!=traces.end(); i++)
    tstrings.push_back(i->getString());

  sort(tstrings.begin(), tstrings.end(), string_greater);
  std::copy(tstrings.begin(), tstrings.end(), ostream_iterator<string>(out, "\n"));
*/
  }


/* Print the traces corresponding to maximal safe words.
 */
void Ograph::printMalWordTraces(ostream &out)
  {
  Trace empty;
  vector<Trace> traces;
  getMalWordTraces(empty, traces);
  std::copy(traces.begin(), traces.end(), ostream_iterator<Trace>(out, "\n"));
  }


/* Print the traces corresponding to maximal safe words starting from the
 * inverse permutation.
 */
void Ograph::printInvMalWordTraces(ostream &out)
  {
  Ograph inverse = getInvGraph();             //Get the inverse permutation.
  Trace empty;
  vector<Trace> traces;

  inverse.getMalWordTraces(empty, traces);    //Get the traces for the inverse.

  vector<int> perm = getSeqVec();
  for(vector<Trace>::iterator i=traces.begin(); i!=traces.end(); i++)
    i->invert(perm);                          //Invert and remap the traces.

  std::copy(traces.begin(), traces.end(), ostream_iterator<Trace>(out, "\n"));
  }


/*Put in -retval- all traces corresponding to this permutation.
 *  ** assume the graph's state corresponds to the given trace.
 */
void Ograph::getWordTraces(Trace &trace, vector<Trace> &retval,
                           const list<Element> &prevword)
  {
  if(getDistNH() == 0)       //If we're sorted
    {
    retval.push_back(trace); //then save the trace
    return;                  //and return.
    }
                             //Get the next words that don't commute with
  vector< list<Pair> > words = getWords(prevword); //the previous word.

                             //Find the suffix to trace < word.
  for(vector< list<Pair> >::iterator w=words.begin(); w!=words.end(); w++)
    {
    list<Element> sinvs = invsTOsets(*w);
    trace.appendWord(Word(sinvs));             //Append the word to the trace.
    applyCommutingSet(*w);                     //Apply the word to the graph.
    getWordTraces(trace, retval, sinvs);       //Recurse using this word.
    trace.removeLast();                        //Remove the last word.
    reApplyCommutingSet(*w);                   //Re-apply the word to the graph.
    }
  }


/*Get the traces that correspond to maximal safe words.
 *  ** assume the graph's state corresponds to the given trace.
 */
void Ograph::getMalWordTraces(Trace &trace, vector<Trace> &retval)
  {
  if(getDistNH() == 0)       //If we're sorted
    {
    retval.push_back(trace); //then save the trace
    return;                  //and return.
    }

  vector< list<Pair> > words = getMalWords(); //Get the next words.
  for(vector< list<Pair> >::iterator w=words.begin(); w!=words.end(); w++)
    {
    trace.appendWord(Word(invsTOsets(*w))); //Append the word to the trace.
    applyCommutingSet(*w);                  //Apply the word to the graph.
    getMalWordTraces(trace, retval);        //Recurse using this word.
    trace.removeLast();                     //Remove the last word.
    reApplyCommutingSet(*w);                //Re-apply the word to the graph.
    }
  }
      
/* Print the words that come from independent sets of noninterfering
 * inversions
 */
void Ograph::printWords(ostream& out)
  {
  vector<Word> words = getWordsW();
  std::copy(words.begin(), words.end(), ostream_iterator<Trace>(out, "\n"));
  }

/* Print the words that come from maximal independent sets of noninterfering
 * inversions.
 */
void Ograph::printMalWords(ostream& out, bool onlysafe)
  {
  vector<Word> words = getMalWordsW(onlysafe);
  std::copy(words.begin(), words.end(), ostream_iterator<Trace>(out, "\n"));
  }

/*Return a vector of all maximal safe first words in a normal form for
this graph.
*/
vector<Word> Ograph::getMalWordsW(bool onlysafe)
  {
  vector<int> ccm = getCommutingCM(cTOi);   //The commuting constraints.
  vector<int> scm = getSortingCM(ccm, cTOi);//The sorting constraints.
  removeUnsafe(ccm, cTOi);
  Graph cg(ccm);                            //The commuting constraints graph.
  Graph sg(scm);                            //The sorting constraints graph.
  Graph ug = cg.getUnion(sg);               //The union of the two.
                                       //Get all maximal independent sets.
  vector< list<int> > mislist = ug.getAllMalIS();


  vector<Word> words;                  //The maximal safe first words.
  for(vector< list<int> >::iterator i=mislist.begin(); i!=mislist.end(); i++)
    if(!onlysafe || isNonint(cTOinv(*i, cTOi)))  //If the inversions are safe
      words.push_back(Word(cTOset(*i, cTOi)));   //then save the Word.


  return words;
  }


/*Return a vector of all safe first reversals in a normal form that don't
 * commute with the given word.
*/
vector< list<Pair> > Ograph::getWords(const list<Element> &word)
  {
  vector<int> ccm = getCommutingCM(cTOi);   //The commuting constraints.
  vector<int> scm = getSortingCM(ccm, cTOi);//The sorting constraints.
  removeUnsafe(ccm, cTOi);
  Graph cg(ccm);                            //The commuting constraints graph.
  Graph sg(scm);                            //The sorting constraints graph.
  Graph ug = cg.getUnion(sg);               //The union of the two.

       //Remove the chords corresponding to inversions that commute with the
       //last word of the trace:
  if(!word.empty())
    for(unsigned i=0; i < cTOi.size(); i++)
      {                      //If the pair for this chord doesn't overlap
      if(commutesWith(word, invTOset((Pair)cTOi[i])))
        ug.remVertex(i);     //then remove it from the graph.
      }

                                       //Get all maximal independent sets.
  vector< list<int> > mislist = ug.getAllIS();

  vector< list<Pair> > words;          //The maximal safe first words.
  //int listsize = mislist.size();
  //int counter=0;
  //cerr << listsize << " sets to check...";
  for(vector< list<int> >::iterator i=mislist.begin(); i!=mislist.end(); i++)
    {
    //cerr << "|";
    //if(counter++ == listsize/10)
      //{
      //cerr << ".";
      //counter=0;
      //}
    list<Pair> pinvs = cTOinv(*i, cTOi);
    if(isNonint(pinvs))                          //If the inversions are skafe
      {
      Word winvs = Word(cTOset(*i, cTOi));       //then get the Word.
      if(hasInterferencePath(winvs, pinvs, *i))  //If there's a path interfering
        words.push_back(pinvs);                  //with the remaining invs,
      }                                          //then keep the word.
    }
  //cerr << endl;

  return words;
  }


/*Return true if the two sets intersect (one is not the subset
 * of the other)
 */
bool Ograph::intersect(const set<int> &s1, const set<int> &s2)
  {
  set<int> s;
  set_intersection(s1.begin(), s1.end(), s2.begin(), s2.end(),
                   insert_iterator< set<int> >(s,s.begin()));
  return !(s.empty() || (s.size() == s1.size() || s.size() == s2.size()));
  }


/*Return true if the given element commutes with the given list of elements.
 */
bool Ograph::commutesWith(const list<Element> &slist, const Element &s)
  {
  if(s.size() == 1)     //If this reversal is length 1
    return true;        //then it commutes with anything.

  for(list<Element>::const_iterator i=slist.begin(); i!=slist.end(); i++)
    if(intersect(*i, s))
      return false;

  return true;
  }


/*Return true if the given Pair commutes with the given list of pairs.
 */
bool Ograph::commutesWith(const list<Pair> &plist, const Pair &p)
  {
  if(p.getLeft() == p.getRight())  //If this reversal is length 1
    return true;                   //then it commutes with anything.

  for(list<Pair>::const_iterator i=plist.begin(); i!=plist.end(); i++)
    if(intersect(*i, p))
      return false;

  return true;
  }


/*Return true if the two intervals intersect (one is not the subset of the
 * other)
 */
bool Ograph::intersect(const Pair &p1, const Pair &p2)
  {
  if(p1.getLeft() > p2.getRight() || p2.getLeft() > p1.getRight())
    return false;    //If they're disjoint.
  if((p1.getLeft() >= p2.getLeft() && p1.getRight() <= p2.getRight()) ||
     (p2.getLeft() >= p1.getLeft() && p2.getRight() <= p1.getRight()))
    return false;    //If one is a subinterval of the other.

  return true;
  }


/*Return a vector of all first words in a normal form for this graph.
*/
vector<Word> Ograph::getWordsW()
  {
  vector< list<Pair> > invlist = getWords();
  
  vector<Word> retlist;
  for(vector< list<Pair> >::const_iterator i=invlist.begin(); i!=invlist.end();
      i++)
    retlist.push_back(Word(invsTOsets(*i)));

  return retlist;
  }


/*Return true if the given word is a complete first word of a trace for this
 * permutation.
 */
bool Ograph::hasInterferencePath(const Word &winvs, list<Pair> &pinvs,
                                 const list<int> &chords)
  {
  applyCommutingSet(pinvs);   //Apply the set so that we can get the next
                              //inversions that may create a new word.
  if(!notSorted())            //If the permutation is sorted
    {                         //then we've found a one word full normal form.
    reApplyCommutingSet(pinvs);
    return true;
    }
  Trace t = Trace(winvs);     //The trace starting with this word.
                              //Get the next safe inversions that don't commute
  list<Pair> interinv = getExtensionSet(t);
                              //with the word -winvs-.
  if(interinv.empty())
    {
    reApplyCommutingSet(pinvs);
    return false;
    }

  for(list<Pair>::const_iterator i=interinv.begin(); i!=interinv.end(); i++)
    {
    Element sinv = invTOset(*i);
    revAll(*i);
    t.appendWord(Word(sinv));         //Add the element.
    if(hasInterferencePath(t))        //If this has an interference path then
      {
      revAll(*i);                     //undo the inversions.
      reApplyCommutingSet(pinvs);
      return true;                    //We know it's a first word.
      }
    t.removeAllButFirst();
    revAll(*i);                       //Otherwise, undo the inversion and
    }                                 //continue to check the others.

  reApplyCommutingSet(pinvs);
  return false;
  }


/*Return the next safe inversions that would be the last element if
 * added to the normal form -t-.
 */
list<Pair> Ograph::getAppendingSet(const Trace &t)
  {
  list<Pair> retval;
  vector<Pair> safeinvs;
  getSafeInvs(safeinvs);            //Get the next inversions.

  for(vector<Pair>::const_iterator i=safeinvs.begin(); i!=safeinvs.end(); i++)
    if(t.goesAtEnd(invTOset(*i)))   //If the new inversion doesn't penetrate
      retval.push_back(*i);         //the trace, then save it.

  return retval;
  }


/*Return the next safe inversions that would be the last element if of a
 * word other than the first word if added to the normal form -t-.
 */
list<Pair> Ograph::getExtensionSet(const Trace &t)
  {
  list<Pair> retval;
  vector<Pair> safeinvs;
  getSafeInvs(safeinvs);            //Get the next inversions.

//cerr << "\nperm:" << getSeqVec() << endl;
//cerr << "trace:" << t << endl;
  for(vector<Pair>::const_iterator i=safeinvs.begin(); i!=safeinvs.end(); i++)
    {
    Element e = invTOset(*i);       //If the new inversion doesn't penetrate
    if(t.goesAtEnd(e) && !t.goesInFirstWord(e))
//{ cerr << "found one:" << invTOset(*i) << endl;
      retval.push_back(*i);         //the trace, then save it.
//}
    }

  return retval;
  }

/*Return the next safe inversions that don't enter the first word of the given
 * normal form.
 */
list<Pair> Ograph::getNoncommSet(const Trace &t)
  {
  list<Pair> retval;
  vector<Pair> safeinvs;
  getSafeInvs(safeinvs);            //Get the next inversions.

  for(vector<Pair>::const_iterator i=safeinvs.begin(); i!=safeinvs.end(); i++)
    {
    Element sinv = invTOset(*i);
    if(!t.goesInFirstWord(sinv))    //If the new inversion doesn't penetrate
      retval.push_back(*i);         //the first word, then save it.
    }
  return retval;
  }


/*Return true if there exist no sequence of reversals that enters the first
 * word of -t-.
 */
bool Ograph::hasInterferencePath(Trace &t)
  {
  if(!notSorted())      //If it is sorted
    return true;        //then we found an interferance path.
  list<Pair> interinv = getExtensionSet(t);
  if(interinv.empty())
    return false;

  for(list<Pair>::const_iterator i=interinv.begin(); i!=interinv.end(); i++)
    {
    Element sinv = invTOset(*i);
    revAll(*i);
    Trace tmp = t;
    tmp.addElement(sinv);             //Add the element.
    if(hasInterferencePath(tmp))      //If this has an interference path then
      {
      revAll(*i);                     //undo the inversion and
      return true;                    //we know it's a first word.
      }
    revAll(*i);                       //Otherwise, undo the inversion and
    }                                 //continue to check the others.

  return false;
  }



/*Return a vector of all maximal safe first words in a normal form for
this graph.
*/
vector< list<Pair> > Ograph::getMalWords(bool onlysafe)
  {
  vector<int> ccm = getCommutingCM(cTOi);   //The commuting constraints.
  vector<int> scm = getSortingCM(ccm, cTOi);//The sorting constraints.
  removeUnsafe(ccm, cTOi);
  Graph cg(ccm);                            //The commuting constraints graph.
  Graph sg(scm);                            //The sorting constraints graph.
  Graph ug = cg.getUnion(sg);               //The union of the two.
                                       //Get all maximal independent sets.
  vector< list<int> > mislist = ug.getAllMalIS();

  vector< list<Pair> > wordv;
  for(vector< list<int> >::iterator i=mislist.begin(); i!=mislist.end(); i++)
    {
    list<Pair> word = cTOinv(*i, cTOi); //Get the word.
    if(!onlysafe || isNonint(word))     //If it safe
      wordv.push_back(word);            //then save the word.
    }

/*
  for(vector< list<int> >::iterator i=mislist.begin(); i!=mislist.end(); i++)
    wordv.push_back(cTOinv(*i, cTOi)); //Turn the chords into inversion Pairs.

  for(vector< list<Pair> >::iterator w=wordv.begin(); w!=wordv.end(); w++)
    {
    if(!isNonint(*w))                  //For every set that has an unsafe inv:
      {
cerr << "++++++++++++++++++++++" << endl;


        //First find those that are part of an unsafe subset:
      Ograph g = *this;
      g.applyCommutingSet(retval, true);
      cerr << "Unsafe inversions exist:" << g.getSeqString() << endl;
    
      map<ipair,int> iTOc = invertMap(cTOi);
      set<int> unsafe;                //The set of unsafe inversions.
cerr << "MIS:";
g.printInvSet(cerr, invsTOsets(retval));
cerr << endl;
      for(list<Pair>::iterator inv = retval.begin(); inv != retval.end(); inv++)
        {
        g.revAll(*inv);
        if(!g.isSafe(*inv))           //If -inv- isn't safe
          unsafe.insert(iTOc[*inv]);  //then put it in the set.
        g.revAll(*inv);
        }
//cerr << "anintcon:" << nintcon << endl;
//cerr << "gnintcon:" << g.nintcon << endl;

  //cerr << "\tunsafe set:";
  //printInvSet(cerr, invsTOsets(cTOinv(unsafe, cTOi)));
  //cerr << endl;

    //Now build the sets:
  //ASSUME THERE IS ONLY ONE SET FOR NOW!!!!

      //Now record the added maximal independent sets:
      Graph nccopy = nintcon;
cerr << "set-b:" << mis << endl;
cerr << "unsafe:" << unsafe << endl;
  //set<int> misset = listTOset(mis);
      set<int>::iterator inv = unsafe.begin();
      while(inv != unsafe.end())
        {
        int node = *inv;
    
        set<int>::iterator tmp = inv++;
        list<int> tmplist;
        tmplist.push_back(node);
        mis.remove(node);                                 //Remove unsafe inv.
        hash_set<int> neighbors = nccopy.remVertex(node);
                                                          //Remove the vertex.
                                                          //Get any additions.
        list<int> another = nccopy.getMalISWith(mis, tmplist);
        cerr << "newset:";
        printInvSet(cerr, invsTOsets(cTOinv(another, cTOi)));
        cerr << endl;
    
        nccopy.addVertex(node, neighbors);                //Restore the vertex.
        tmplist.clear();
        mis.push_back(node);                              //Restore the list.
        }
      }
cerr << "set-e:" << mis << endl;
    }
*/

  return wordv;
  }




/*Return the next maximal independent set of noninterfering inversions.
*/
/*
list<Pair> Ograph::getNextNintMalIS()
  {
  int _NOT_FINISHED_YET__MAYBE_UNNECESSARY;
  if(nintcon.empty())
    {
    vector<int> ccm = getCommutingCM(cTOi);   //The commuting constraints.
    vector<int> scm = getSortingCM(ccm, cTOi);//The sorting constraints.
    removeUnsafe(ccm, cTOi);
    Graph cg(ccm);                   //The commuting constraints graph.
    Graph sg(scm);                   //The sorting constraints graph.
    nintcon = cg.getUnion(sg);       //The union of the two.
    }
                                     //Get the next maximal independent set.
  list<int> mis = nintcon.getNextMalIS();

  list<Pair> retval;
  if(mis.empty())                    //If the MIS is empty then 
    return retval;                   //return the empty list of Pairs.
  retval = cTOinv(mis, cTOi);        //Turn the chords into inversion Pairs.
//cerr << "mnintcon:" << nintcon << endl;
  if(isNonint(retval))               //If we found a noninterfering set
    return retval;                   //then immediately return it.
cerr << "++++++++++++++++++++++" << endl;


    //First find those that are part of an unsafe subset set:
  Ograph g = *this;
  g.applyCommutingSet(retval, true);
  cerr << "Unsafe inversions exist:" << g.getSeqString() << endl;

  map<ipair,int> iTOc = invertMap(cTOi);
  set<int> unsafe;                //The set of unsafe inversions.
cerr << "MIS:";
g.printInvSet(cerr, invsTOsets(retval));
cerr << endl;
  for(list<Pair>::iterator inv = retval.begin(); inv != retval.end(); inv++)
    {
    g.revAll(*inv);
    if(!g.isSafe(*inv))           //If -inv- isn't safe
      unsafe.insert(iTOc[*inv]);  //then put it in the set.
    g.revAll(*inv);
    }
//cerr << "anintcon:" << nintcon << endl;
//cerr << "gnintcon:" << g.nintcon << endl;

  //cerr << "\tunsafe set:";
  //printInvSet(cerr, invsTOsets(cTOinv(unsafe, cTOi)));
  //cerr << endl;

    //Now build the sets:
  //ASSUME THERE IS ONLY ONE SET FOR NOW!!!!

    //Now record the added maximal independent sets:
  Graph nccopy = nintcon;
cerr << "set-b:" << mis << endl;
cerr << "unsafe:" << unsafe << endl;
  //set<int> misset = listTOset(mis);
  set<int>::iterator inv = unsafe.begin();
  while(inv != unsafe.end())
    {
    int node = *inv;

    set<int>::iterator tmp = inv++;
    list<int> tmplist;
    tmplist.push_back(node);
    mis.remove(node);                                 //Remove the unsafe inv.
    hash_set<int> neighbors = nccopy.remVertex(node); //Remove the vertex.
                                                      //Get any additions.
    list<int> another = nccopy.getMalISWith(mis, tmplist);
    cerr << "newset:";
    printInvSet(cerr, invsTOsets(cTOinv(another, cTOi)));
    cerr << endl;

    nccopy.addVertex(node, neighbors);                //Restore the vertex.
    tmplist.clear();
    mis.push_back(node);                              //Restore the list.
    }
cerr << "set-e:" << mis << endl;

  return retval;
  }
*/



/*Convert a list of inversions in endpoint format to a list of inversions
in set format.
*/
list< set<int> > Ograph::invsTOsets(const list<Pair> invs) const
  {
  list< set<int> > retval;
  for(list<Pair>::const_iterator i = invs.begin(); i != invs.end(); i++)
    retval.push_back(invTOset(*i));

  //sort(retval.begin(), retval.end());
  return retval;
  }


/*Convert the inversion in endpoint format to an inversion in set format.
*/
set<int> Ograph::invTOset(const Pair &inv) const
  {
  vector<int> perm = getSeqVec();
  set<int> tmp;
  for(int j = inv.getLeft(); j <= inv.getRight(); j++)
    tmp.insert(abs(perm[j-1]));
  return tmp;
  }


/*Convert a vector of inversions in endpoint format to a vector of inversions
in set string format.
*/
vector<string> Ograph::invTOsetstr(const vector<Pair> invs) const
  {
  vector<ipair> ipinvs;
  for(vector<Pair>::const_iterator i = invs.begin(); i != invs.end(); i++)
    ipinvs.push_back(ipair(i->getLeft(), i->getRight()));

  return invTOsetstr(ipinvs);
  }

/*Convert a vector of inversions in endpoint format to a vector of inversions
in set string format.
*/
vector<string> Ograph::invTOsetstr(const list<Pair> invs) const
  {
  vector<ipair> ipinvs;
  for(list<Pair>::const_iterator i = invs.begin(); i != invs.end(); i++)
    ipinvs.push_back(ipair(i->getLeft(), i->getRight()));

  return invTOsetstr(ipinvs);
  }

/*Convert a vector of inversions in endpoint format to a vector of inversions
in set string format.
*/
vector<string> Ograph::invTOsetstr(const vector<ipair> invs) const
  {
  vector<int> perm = getSeqVec();

  vector<string> retval;
  for(vector<ipair>::const_iterator i = invs.begin(); i != invs.end(); i++)
    {
    set<int> tmp;
    for(int j = i->first; j <= i->second; j++)
      tmp.insert(abs(perm[j-1]));

    retval.push_back(getInvSetStr(tmp));
    }

  return retval;
  }


/* Ouput a list of sets that represent inversions.
*/
void Ograph::printInvSet(ostream& output, const list< set<int> > slist)
  {
  for(list< set<int> >::const_iterator set = slist.begin();
                                       set != slist.end(); set++)
    output << getInvSetStr(*set);
  }



/*Print the graph represening the interference constraints.
 * 
 * */
void Ograph::printNonintGraph() const
  {
  cout << getNonintGraph() << endl;          //Print the graph out.
  }


/*Get the graph representing the interference constraints.
 *
 */
Graph Ograph::getNonintGraph() const
  {
  vector<ipair> chordTOinv;
  vector<int> ccm = getCommutingCM(chordTOinv);    //The commuting constraints.
  vector<int> scm = getSortingCM(ccm, chordTOinv); //The sorting constraints.
  removeUnsafe(ccm, chordTOinv);
  Graph cg(ccm, invTOsetstr(chordTOinv)); //The commuting constraints graph.
  Graph sg(scm, invTOsetstr(chordTOinv)); //The sorting constraints graph.
  return cg.getUnion(sg);                 //The union of the two.
  }



/*Return a maximum independent set of noninterfering inversions.
*/
list<Pair> Ograph::getNonintMIS() const
  {
  vector<ipair> chordTOinv;
  vector<int> ccm = getCommutingCM(chordTOinv);    //The commuting constraints.
  vector<int> scm = getSortingCM(ccm, chordTOinv); //The sorting constraints.
  removeUnsafe(ccm, chordTOinv);
  Graph cg(ccm, invTOsetstr(chordTOinv)); //The commuting constraints graph.
  Graph sg(scm, invTOsetstr(chordTOinv)); //The sorting constraints graph.
  Graph ug = cg.getUnion(sg);             //The union of the two.

  list<int> mis = ug.getMISEX();     //Get the maximum independent set.

  list<Pair> retval;
  if(mis.empty())                    //If the MIS is empty then 
    return retval;                   //return the empty list of Pairs.
  retval = cTOinv(mis, chordTOinv);  //Turn the chords into inversion Pairs.
  if(isNonint(retval))               //If we found a noninterfering set
    return retval;                   //then immediately return it.


                                     //The inversions may not all be safe:
  map<ipair, int> invTOchord = invertMap(chordTOinv);
  list<int> newmis;
  for(int k=0; k < comps->getNumComps(); k++)
    {                                          //For each HP component:
    list<Pair> pcmis=comps->invsONcomp(k, retval); //The invs for this HP comp.
    list<int> cmis = invTOc(pcmis,invTOchord);     //The corresponding chords.


    if(!isNonint(pcmis))                       //If there is an unsafe inversion
      {
      //cerr << "Unsafe Inversion " << *findInterfering(pcmis)
           //<< " exists in\n\t" << pcmis;
      //cerr << "\n\tSubstituting..." << endl;

      Graph cc = ug.getCompsWith(cmis);        //get the pertinent components 
      list<int>::iterator mp = cmis.begin();   //and find a valid substitution:
      while(cc.ISreplace(cmis, mp) &&          //While we can replace one
            !isNonint(cTOinv(cmis,chordTOinv)))//and still have conflicts,
        mp++;                                  //move on to the next op.
    
      if(mp == cmis.end())
        {                         //If there isn't one, exhaustively
        cerr << "Trying exhaustive search..." << endl;
        if(!findNISofsizek(cmis.size(), cc, cmis, chordTOinv))
          {                       //search for a same size set.
          cerr << "????Didn't find IS of size " << cmis.size() << "." << endl;
          //This is the spot that needs to be changed if we want it to work
          // for more then 1 unsafe inverstion per component.
          int _ASSUME_ONE_UNSAFE_PER_COMPONENT;
          list<Pair> tmppl = cTOinv(cmis, chordTOinv);
          int rmchord = invTOchord[*findInterfering(tmppl)];
          cerr << "removing " << chordTOinv[rmchord] << " from "
               << cTOinv(cmis, chordTOinv) << endl;
          cmis.erase(find(cmis.begin(), cmis.end(), rmchord));
          }                       //If there isn't one, remove unsafe inversion.
        else
          cerr << "!!!!Found new IS of size " <<cmis.size()<<":"<< cmis << endl;
        }
      else
        {
        cerr << "replacing:" << cTOinv(cmis,chordTOinv) << endl;
        }
      }

    newmis.insert(newmis.end(), cmis.begin(), cmis.end());
    }

  retval.clear();
  for(list<int>::iterator i = newmis.begin(); i != newmis.end(); i++)
    retval.push_back(Pair(chordTOinv[*i].first, chordTOinv[*i].second));

  return retval;
  }




/*Return a mutual maximum independent set of noninterfering inversions.
*/
list<Pair> Ograph::getMutualNonintMIS(const Ograph &og) const
  {
  list<Pair> retval;         //Inversions to return.

  vector<ipair> chordTOinv1, chordTOinv2;
  vector<int> ccm1 = getCommutingCM(chordTOinv1);    //Commuting constraints.
  vector<int> scm1 = getSortingCM(ccm1, chordTOinv1);//The sorting constraints.
  removeUnsafe(ccm1, chordTOinv1);

  vector<int> ccm2 = og.getCommutingCM(chordTOinv2);
  vector<int> scm2 = og.getSortingCM(ccm2, chordTOinv2);
  og.removeUnsafe(ccm2, chordTOinv2);

  vector<ipair> chordTOinv = restrictTOcommon(ccm1, scm1, chordTOinv1,
                                              ccm2, scm2, chordTOinv2);

  Graph cg1(ccm1);                   //The commuting constraints graph.
  Graph sg1(scm1);                   //The sorting constraints graph.
  Graph cg2(ccm2);                   //The commuting constraints graph.
  Graph sg2(scm2);                   //The sorting constraints graph.

  Graph fg = cg1.getUnion(sg1);      //Build constraints for both permutations:
  fg = fg.getUnion(cg2);             //Add the commuting constraints from 2.
  fg = fg.getUnion(sg2);             //Add the sorting constraints from 2.

  list<int> mis = fg.getMISEX();     //Get the maximum independent set.

  if(mis.empty())                    //If the MIS is empty then 
    return retval;                   //return the empty list of Pairs.
  retval = cTOinv(mis, chordTOinv);  //Turn the chords into inversion Pairs.
  if(isNonint(retval) && og.isNonint(retval))
    return retval;                   //If we found a noninterfering set, return.



                                     //The inversions may not all be safe:
  map<ipair, int> invTOchord = invertMap(chordTOinv);
  list<int> newmis;
  for(int k=0; k < comps->getNumComps(); k++)
    {                                          //For each HP component:
    list<Pair> pcmis1=comps->invsONcomp(k, retval); //The invs for this HP comp.
    list<Pair> pcmis=og.comps->invsONcomp(pcmis1, retval);
    if(!shareInvs(cTOinv(newmis,chordTOinv), pcmis))
      {
      list<int> cmis = invTOc(pcmis,invTOchord);    //The corresponding chords.
  
      if(!isNonint(pcmis) || !og.isNonint(pcmis))
        {                                      //If there is an unsafe inversion
        // cerr << "Unsafe Inversion(s) " << *findInterfering(pcmis) << " "
             // << *og.findInterfering(pcmis) << " exist in\n\t" << pcmis;
        // cerr << "\n\tSubstituting..." << endl;
  
        Graph cc = fg.getCompsWith(cmis);      //get pertinent -fg- components 
        list<int>::iterator mp = cmis.begin(); //and find a valid substitution:
        while(cc.ISreplace(cmis, mp) &&             //While we have replaced one
              (!isNonint(cTOinv(cmis,chordTOinv)) ||//and still have conflicts,
               !og.isNonint(cTOinv(cmis,chordTOinv))))
          mp++;                                     //move on to the next op.
      
        if(mp == cmis.end())
          {                       //If there isn't a replacement, exhaustively
          //cerr << "Trying exhaustive search..." << endl;
          if(!findNISofsizek(cmis.size(), cc, cmis, chordTOinv, &og))
            {                     //search for a same size set.
            //cerr << "????Didn't find IS of size "<<cmis.size()<<"."<<endl;
            //If there is more then one unsafe inversion on this subgraph
            //there will be a problem:
            int _ASSUME_ONE_UNSAFE_PER_COMPONENT;
            if(!tryISwithout(cc, cmis, chordTOinv, &og))
              {              //If there isn't one die:
              cerr << "ERROR: there will be an unsafe inversion in the set!\n";
              }
            //else
              //cerr << "removed one:" << cTOinv(cmis, chordTOinv) << endl;
            }
          //else
            //cerr << "!!!!Found new IS of size "<<cmis.size()<<":"<<cmis<<endl;
          }
        // else
          // {
          // cerr << "replacing:" << cTOinv(cmis,chordTOinv) << endl;
          // }
        }
  
      newmis.insert(newmis.end(), cmis.begin(), cmis.end());
      }
    }


  retval.clear();       //Convert to inversions while removing duplicates:
  hash_set<int> beenseen;
  for(list<int>::iterator i = newmis.begin(); i != newmis.end(); i++)
    if(beenseen.find(*i) == beenseen.end())
      {
      beenseen.insert(*i);
      retval.push_back(Pair(chordTOinv[*i].first, chordTOinv[*i].second));
      }

  return retval;
  }









/*Try to find a noninterfering independent set of size k.
Return true and modify -is- if we find one.
*/
bool Ograph::findNISofsizek(int k, const Graph &g, list<int> &is,
                            const vector<ipair> &chordTOinv,
                            const Ograph* og) const
  {
                   //Now check for all possible ISs of size k:
  list<Pair> pset;
  for(unsigned int i=0; i < chordTOinv.size(); i++)
    if(g.hasVertex(i))
      pset.push_back(Pair(chordTOinv[i].first, chordTOinv[i].second));

  unsigned long long nCHk = choose(pset.size(), k);
  if(nCHk > NINT_TOO_MANY_TO_SEARCH)
    {
    cerr << "Skipping search on (" << pset.size() << " choose " << k << ") = "
         <<  nCHk << " sets!" << endl;
    return false;
    }

  list<Pair> mis;
  if(tryISofsizek(k, mis, pset.begin(), pset.end(), og))
    {                                       //If we find an IS of size k
    is.clear();
    for(unsigned int i=0; i<chordTOinv.size(); i++)
      if(find(mis.begin(), mis.end(), chordTOinv[i]) != mis.end())
        is.push_back(i);                    //then copy the chords to -is-

    return true;                            //and return true.
    }

  return false;
  }


/*Recursively try to find an independent set by adding -k- more elements
 from between -from- and -end- to -set-.  When k=0, find the IS of set.
If a set if found true is returned and -set- contains the set.
*/
bool Ograph::tryISofsizek(int k, list<Pair> &set, list<Pair>::iterator from,
                          const list<Pair>::const_iterator &end,
                          const Ograph* og) const
  {
  if(k)                                        //If we're not at the bottom,
    {
    if(from == end)                            //check if we're out of elements
      return false;                            //to test,

    set.push_back(*from);                      //add it to the set,
    if(tryISofsizek(k-1, set, ++from, end, og))//and try it with the remainder.
      return true;
    set.pop_back();                            //If it didn't work, remove it

    if(tryISofsizek(k, set, from, end, og))    //and try without it.
      return true;
    }
  else                                         //Otherwise, test the set.
    {
    if(og == NULL)
      return isNonint(set);
    else
      return isNonint(set) && og->isNonint(set);
    }

  return false;
  }



/*Return true if the given set of inversions commutes.
*/
bool Ograph::isCommuting(const list<Pair> &set) const
  {
  for(list<Pair>::const_iterator i = set.begin(); i != set.end(); i++)
    {
    list<Pair>::const_iterator j = i; j++;
    while(j != set.end())
      if(i->overlaps(*j++))
        return false;
    }

  return true;
  }



/*Return true if the given set is noninterfering.
*/
bool Ograph::isNonint(list<Pair> set) const
  { return isCommuting(set) && (findInterfering(set) == set.end()); }

/*Return set.end() if the given commuting set is noninterfering.
Otherwise return a pointer to an offending inversion in -set-.
-set- is not modified.
*/
list<Pair>::iterator Ograph::findInterfering(list<Pair> &set) const
  {
  Ograph g = *this;
  return g.applyCommutingSet(set);
  }



/*True if the union of commuting and noninterfering constraints can be
represented by a chord model.
*/
bool Ograph::isCG() const
  {
  vector<ipair> chordTOinv;
  vector<int> ccm = getCommutingCM(chordTOinv);

  vector<int> scm = getSortingCM(ccm, chordTOinv);
  Graph cg(ccm);   //The graph with commuting constraints.
  Graph sg(scm);   //The graph with sorting (interference) constraints.
  Graph ug = cg.getUnion(sg);  //The graph that is the union of the two.

  int _REMOVE_TEMPORARY_ASSERTS_HERE;
  assert(sg.checkCM(scm));  //TEMPORARY!
  assert(cg.checkCM(ccm));  //TEMPORARY!
  return ug.isCG();
  }


/*Return a permutation that is a shrink of the current one where each component
is a single cycle.
Fill in -removed- with the elements removed.
*/
vector<int> Ograph::getShrink(hash_set<int> &removed) const
  {        //Assure that we have a single cycle (create newseq by shrinks):
  int newseqsize = 0;
  int shrinksdone = 0;
  vector<int> newseq;                //The permutation with (d,r)-shrinks.
  newseq.resize(seqsize-2);
  Ograph tmp = *this;
  for(int i=1; i < tmp.seqsize-1; i++)
    {
    if((tmp.comps->getCycleNum((tmp.seqinds[i]*2)-1) ==
        tmp.comps->getCycleNum(tmp.seqinds[i]*2)) ||
       (tmp.comps->getCompNum((tmp.seqinds[i]*2)-1) !=
        tmp.comps->getCompNum(tmp.seqinds[i]*2)))
      {                              //For element abs(i) on the same cycle or
      if(tmp.seq[tmp.seqinds[i]] > 0)//on different components, don't shrink.
        newseq[tmp.seqinds[i]-1] = tmp.seq[tmp.seqinds[i]]-shrinksdone;
      else
        newseq[tmp.seqinds[i]-1] = tmp.seq[tmp.seqinds[i]]+shrinksdone;
      newseqsize++;
      }
    else                             //Otherwise,
      {
      removed.insert(tmp.seq[tmp.seqinds[i]]);
      newseq[tmp.seqinds[i]-1] = 0;  //Turn the element into 0.
      shrinksdone++;
      tmp.comps->shrinkCycles(tmp.comps->getCycleNum((tmp.seqinds[i]*2)-1),
                              tmp.comps->getCycleNum(tmp.seqinds[i]*2));
      }
    }
  shrinksdone = 0;
  for(int i=0; i < seqsize-2; i++)
    {
    if(newseq[i] == 0)
      shrinksdone++;
    else
      newseq[i-shrinksdone] = newseq[i];
    }
  newseq.resize(newseqsize);

  return newseq;
  }

/*Return the chord model for the sorting constraints labeled based on the
given chord to inversion mapping -chordTOinv-.
  To do this we rearrange the chord model for the commuting constraints
according to the given -chordTOinv- map.
  ** NOTE: not the way the paper describes.
*/
vector<int> Ograph::getSortingCM(const vector<int> &ccm,
                                 const vector<ipair> &chordTOinv) const
  {
       //Get a shrink of this permutation.
  hash_set<int> removed;
  vector<int> shperm = getShrink(removed);
  Ograph shg(shperm);     //The HP graph for the shrunken permutation.

       //Now get a chord model representing noninterfering constraints
       //based on this shrink:
  vector<int> cm;         //The chord model.

       //Map the reality edges to the intervals on the given chord model:
  vector<ipair> edgecount;//Cycle i has edgecount[i].first FORWARD edges.
  edgecount.resize(comps->getNumCycles());
  for(int i=0; i < graphsize-1; i+=2)           //For each reality edge
    {
    if(gr[i].reality.direction == FORWARDS)     //increment the appropriate
      edgecount[comps->getCycleNum(i)].first++; //counter for the cycle.
    else
      edgecount[comps->getCycleNum(i)].second++;
    }
  vector<ipair> redgeTOcm;//Map reality edge to chord model interval [x,y).
  int cmpos = 0;          //The position in the commuting chord model.
  for(int i=0; i < graphsize-1; i+=2)       //For each reality edge
    {                                       //the # of chords is the # of edges
    int left = cmpos;
    if(gr[i].reality.direction == FORWARDS) //of opposite orientation.
      cmpos += edgecount[comps->getCycleNum(i)].second;
    else
      cmpos += edgecount[comps->getCycleNum(i)].first;
    redgeTOcm.push_back(ipair(left,cmpos));
    }

       //Find the mapping of reality edges in shg to those in this graph:
  vector<ipair> redgemap; //Map reality edges in -shg- to a range [x,y] in this.
  for(int i=0; i < seqsize-2;)
    {
    redgemap.push_back(ipair(i,0));   //Set the left endpoint of the next one.
                                      //Find the next nonremoved element:
    while((i++ < seqsize-2) && removed.find(seq[i]) != removed.end());
    redgemap.rbegin()->second = i-1;  //Set the right endpoint.
    }
  if(redgemap.size() < (unsigned)seqsize-1)   //If last element was not removed
    redgemap.push_back(ipair(seqsize-2, seqsize-2));    //add the last edge.


       //Now follow the cycle on shg and copy intervals of ccm accordingly:
  hash_set<int> compseen;         //The components we've already looked at.
  for(int i=1; i < shg.graphsize; i+=2)//For each reality edge
    {
    if(compseen.find(shg.comps->getCompNum(i)) == compseen.end())
      {                                //on an unvisited component:
      compseen.insert(shg.comps->getCompNum(i));

      int node = i;        //The current node.
      do{                  //For each reality edge in the shortened cycle:
        bool forwards = (shg.gr[node].reality.direction == FORWARDS);
        if(forwards)       //If it's forward facing
          {
          for(int i = redgemap[(int)floor(node/2)].first;   //visit the edges of
              i <= redgemap[(int)floor(node/2)].second; i++)//full graph forward
            for(int j = redgeTOcm[i].first; j < redgeTOcm[i].second; j++)
              cm.push_back(ccm[j]);                         //and copy chords.
          }
        else               //Otherwise, it's backwards facing so
          {
          for(int i = redgemap[(int)floor(node/2)].second; //visit the edges of
              i >= redgemap[(int)floor(node/2)].first; i--)//full graph backward
            for(int j = redgeTOcm[i].second-1; j >= redgeTOcm[i].first; j--)
              cm.push_back(ccm[j]);                        //and copy chords.
          }
        node = shg.gr[shg.gr[node].desire.id].reality.id;  //Move to next node.
        }
      while(node != i);
      }
    }

  removeUnsafe(cm, chordTOinv);

  return cm;
  }


/*Remove the unsafe inversions from the given chord model.
*/
void Ograph::removeUnsafe(vector<int> &cm,const vector<ipair> &chordTOinv) const
  {
  hash_set<int> toberemoved;
  int index=0;
  for(vector<ipair>::const_iterator i = chordTOinv.begin();
                                    i < chordTOinv.end(); i++)
     {
     if(!isSafe(*i))
       toberemoved.insert(index);
     index++;
     }

  list<int> cml;
  std::copy(cm.begin(), cm.end(), back_inserter(cml));

  list<int>::iterator i = cml.begin();
  while(i != cml.end())
    {
    list<int>::iterator j = i;
    i++;
    if(toberemoved.find(*j) != toberemoved.end())
      cml.erase(j);
    }

  cm.clear();
  std::copy(cml.begin(), cml.end(), back_inserter(cm));
  }

/*Sort the elements by the position they occur in the sequence.
*/
void Ograph::sortByPosition(list<int> &l) const
  {
  if(l.size() == 1)
    return;

  list<int> l2;               //Put the back half of -l- into -l2-.
  list<int>::iterator middle = l.begin();
  for(unsigned int i=0; i < l.size()/2; i++)
    middle++;
  l2.splice(l2.begin(), l, middle, l.end());
  sortByPosition(l);
  sortByPosition(l2);
  merge(l, l2);
  }

/*Merge the sorted lists based on position of the element in the sequence.
*/
void Ograph::merge(list<int> &l, list<int> & l2) const
  {
  list<int>::iterator i2 = l2.begin();
  for(list<int>::iterator i = l.begin(); i != l.end(); i++)
    {
    list<int>::iterator left = i2;
    while(i2 != l2.end() && seqinds[abs(*i2)] < seqinds[abs(*i)]) i2++;
    l.splice(i, l2, left, i2);
    }
  l.splice(l.end(), l2, i2, l2.end());
  }



/*Return the chord model for the commuting constraints.
  Fill -intTOinv- with the map from chord to inversion Pairs.
*/
vector<int> Ograph::getCommutingCM(vector<ipair> &intTOinv) const
  {
  vector<int> retval;                //Vectorize the answer (retrofit code).

        //Take the edge orientation information from the graph:
  int *forwards = new int[seqsize];  //The forward facing reality edges.
  int fsize=0;
  int *backwards = new int[seqsize]; //The backward facing reality edges.
  int bsize=0;
  for(int i=0; i < seqsize-1; i++)
    {
    if(gr[2*i].reality.direction == FORWARDS)
      forwards[fsize++] = i;
    else
      backwards[bsize++] = i;
    }
  if(bsize == 0)                     //If there are no oriented inversions,
    return retval;                   //then return an empty chord model.

       //Build the chord model (indexed starting at 0):
  int csize = 0;
  int *cmodel = new int[fsize*bsize*2];//The chord (interval) model.
  intTOinv.resize(fsize*bsize);

  int msize;
  if(fsize > bsize)
    msize = fsize;
  else
    msize = bsize;
  int notused=0;                       //Index larger than any so far.
                                       
  int **mapstacks = new int*[seqsize-1];
  for(int i=0; i<seqsize-1; i++)       //One stack per edge.
    mapstacks[i] = new int[msize];
  int *mapinds = new int[seqsize-1];   //The size of each stack.
  for(int i=0; i < seqsize-1; i++)
    mapinds[i] = 0;

                              //Build the chord model as an adjacency list:
  int find=0, bind=0;
  for(int i=bsize-1; i>=0; i--)        //For BACKWARD edges right of the first
    if(comps->getCycleNum(backwards[i]*2) == comps->getCycleNum(1))
      {                                //that are on the same cycle, 
      cmodel[csize++] = notused;       //store the new left endpoint
      mapstacks[0][mapinds[0]++] = notused++;
      }                                //and stack it on this edge.
  find++;
  while(find < fsize || bind < bsize)  //While we have some reality edges left:
    {
    if(bind >= bsize || (find < fsize && forwards[find] < backwards[bind]))
      {                                //If we're looking at a FORWARD edge.
      int cnum = comps->getCycleNum(forwards[find]*2);
      for(int i=bind-1; i >= 0; i--)   //For each BACKWARD edge to the left
        {                              //that is on the same cycle,
        if(comps->getCycleNum(backwards[i]*2) == cnum)
          {                            //store the right endpoint
          cmodel[csize++] = mapstacks[backwards[i]][--mapinds[backwards[i]]];
          intTOinv[cmodel[csize-1]] = ipair(backwards[i]+1, forwards[find]);
          }                            //and save the interval to inversion map.
        }
      for(int i=bsize-1; i>=bind; i--) //For each BACKWARD edge to the right
        if(comps->getCycleNum(backwards[i]*2) == cnum)
          {                            //that is on our cycle,
          cmodel[csize++] = notused;   //store the new left endpoint
          mapstacks[forwards[find]][mapinds[forwards[find]]++] = notused++;
          }                            //and stack it on this edge.
      find++;
      }
    else                               //Otherwise, we're at a BACKWARD edge.
      {
      int cnum = comps->getCycleNum(backwards[bind]*2);
      for(int i=find-1; i >= 0; i--)   //For each FORWARD edge to the left
        {                              //that is on the same cycle,
        if(comps->getCycleNum(forwards[i]*2) == cnum)
          {                            //store the right endpoint
          cmodel[csize++] = mapstacks[forwards[i]][--mapinds[forwards[i]]];
          intTOinv[cmodel[csize-1]] = ipair(forwards[i]+1, backwards[bind]);
          }                            //and save the interval to inversion map.
        }
      for(int i=fsize-1; i>=find; i--) //For each FORWARD edge to the right
        if(comps->getCycleNum(forwards[i]*2) == cnum)
          {                            //that is on our cycle,
          cmodel[csize++] = notused;   //store the new left endpoint
          mapstacks[backwards[bind]][mapinds[backwards[bind]]++] = notused++;
          }                            //and stack it on this edge.
      bind++;
      }
    }

  for(int i=0; i<csize; i++)
    retval.push_back(cmodel[i]);
  intTOinv.resize(csize/2);

  delete [] cmodel;
  delete [] forwards;
  delete [] backwards;
  delete [] mapinds;
  for(int i=0; i<seqsize-1; i++)
    delete [] mapstacks[i];
  delete [] mapstacks;

  return retval;
  }






//%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Inversion %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


/* Do a reversal. Return the interval(inclusive) over which the reversal was
done.
*/
Pair Ograph::doRev()
  {
  int maxScoreVert;		//First to the -matrix- updates:
  Pair currRev;			//Current reversal

  maxScoreVert = getMaxScoreVert();

  adjScoreArray(maxScoreVert);     //Update the score array.
  adjAdjList(maxScoreVert);        //Update the matrix values for neighbors.
  adjParity(maxScoreVert);         //Update the parity values.

  bool* v = matrix[maxScoreVert];
  for(int j = 0; j < matsize; ++j) //Update the value for this vertex.
    v[j] = 0;

                               //Now do the -seq- updates:
  int mini = seqinds[maxScoreVert];//Get the interval that the reversal is over.
  int maxi = seqinds[maxScoreVert+1];
  if(mini > maxi)                  //Make sure it's proper...
    {
    int t = mini;
    mini = maxi;
    maxi = t;
    }

  if(seq[mini] + seq[maxi] > 0)
    {
    rev(mini, maxi - 1);    //Update sequence. Do revAll() to maintain gr.
    currRev = Pair(mini-1, maxi-2);
    }
  else
    {
    rev(mini + 1, maxi);    //Update sequence. Do revAll() to maintain gr.
    currRev = Pair(mini, maxi-1);
    }
#ifdef SORTING_SEQUENCE
  if (globalSortSequence) {
    globalSortSequence->addRev(currRev);
  }
#endif
  return currRev;
  }


/* Reverse -seq- and -seqinds- over the given interval(inclusive).
Take sequence indices.
*/
void Ograph::rev(int min, int max) {
  int t;

  assert(min>0 && max<seqsize-1);   //Make sure we reverse over valid indicies.

  while(min <= max) {
    t = seq[min];
    seq[min] = -seq[max];
    seq[max] = -t;

    seqinds[abs(seq[min])] = min;
    seqinds[abs(seq[max])] = max;
    min++;
    max--;
  }
}

/* Reverse -seq-, -seqinds-, and -gr- over the given interval(inclusive).
   Take a Pair.  Update the component info.
*/
void Ograph::revAll(const Pair& p)
  { revAll(p.getLeft(), p.getRight()); }
void Ograph::revAll(const ipair& p)
  { revAll(p.first, p.second); }

/* Reverse -seq-, -seqinds-, and -gr- over the given interval(inclusive).
   Take sequence indices.  Update the component info.
*/
void Ograph::revAll(int min, int max)
  {
  revAllNoUpdate(min, max); 
  comps->classCycles();
  }

/* Reverse -seq-, -seqinds-, and -gr- over the given interval(inclusive).
   Take sequence indices.  Don't update the component info (for efficiency).
  **Efficiency: Update cycles in a smarter way.
*/
void Ograph::revAllNoUpdate(int min, int max) {
  Node tNode;
  int t;

  rev(min, max);

  min = min*2-1;
  max = max*2;

  while(min < max) {
    t = gr[max].desire.id;
    gr[ gr[min].desire.id ].desire.id = max;
    gr[ t                 ].desire.id = min;

    tNode = gr[min];
    {
      if (min%2==0) {
        tNode.reality.id = max-1;
        gr[max].reality.id = min+1;
      }
      else {
        tNode.reality.id = max+1;
        gr[max].reality.id = min-1;
      }
    }
    gr[min] = gr[max];
    gr[max] = tNode;

    ++min;
    --max;
  }

  findCycles();    // **Efficiency - see above.
}



/* Do the given reversal as though it were done on the inverse graph...
in other words, do an inversion on the identity and make the appropriate
changes to the breakpoint graph.  As usual, the given pair represents the 
1-indexed, inclusive indices.
   Update everything.
   ** We just annihilate everything and rebuild the object based on our
      calculated new sequence.
*/
void Ograph::revAllInv(const Pair& p)
  {
  int len = seqsize-2;
  int tmp[len];
  for(int i=0; i<len; i++)   //Copy the sequence.
    tmp[i] = seq[i+1];

  int map[p.getRight()+1];   //We won't use the first p.getLeft() indices.
  int c=0;
  for(int i=p.getLeft(); i<=p.getRight(); i++) //Build the map.
    map[i] = p.getRight()-c++;


  for(int i=0; i<len; i++)     //Copy the sequence.
    if((abs(tmp[i]) >= p.getLeft()) && (abs(tmp[i]) <= p.getRight()))
      {
      if(tmp[i] > 0)               //If the value here is positive
        tmp[i] = -map[tmp[i]];     // then put the negation of it's map
      else                         // otherwise,
        tmp[i] = map[abs(tmp[i])]; // put it's map.
      }

  del();               //Delete everything.
  construct(tmp, len); //Construct with the new sequence.
  }




/* Return true if the given pair represents a safe reversal on the
current graph.
  ** This check isn't totally right at the moment because it just compares
     the number of unoriented component before and after the inversion.
     If the particular component that becomes unoriented seperates two others,
     then this is not a hurdle so is technically "safe".
      `--->  If the hurdles are cleared at the begining instead of when we 
             can't sort any longer then this wouldn't be an issue.
*/
bool Ograph::isSafe(const Pair& p) const
  {
  Ograph tmp = *this;

        //Do the quick check first:
        //  (will the signs of the new adjacencies be different?)
  int loutside = seq[p.getLeft()-1];
  int linside = seq[p.getLeft()];
  int routside = seq[p.getRight()+1];
  int rinside = seq[p.getRight()];
  if(((loutside<0 && rinside<0) || (loutside>=0 && rinside>=0)) &&
     ((routside<0 && linside<0) || (routside>=0 && linside>=0)))
    return true;  //If the signs of the would-be adjacencies are the same now.

        //Now do the real check if it passes:
  int UCcount = tmp.comps->getBadCompCount(); //Get the current UC count.
  int Ccount = tmp.getNumCycles();            //Get the number of cycles.
  tmp.revAll(p);                              //Do the inversion.
  if(UCcount < tmp.comps->getBadCompCount() ||//If we have more unoriented comps
     Ccount >= tmp.getNumCycles())            //or didn't create a new cycle
    return false;                             //then return false.

  return true;
  }
bool Ograph::isSafe(const ipair& p) const
  { return isSafe(Pair(p.first, p.second)); }


/* Return true if an adjacency is created by the given reversal.
*/
bool Ograph::createsAdjacency(const ipair r) const
  {
  if((-seq[r.second] - seq[r.first-1] == 1) ||
     (seq[r.second+1] - -seq[r.first] == 1))
    return true;
  return false;
  }


/* Apply the given set of commuting inversions to the permutation.
If force is -false-, then stop and return a pointer to an unsafe inversion in
-set- if one exists.
-set- is not modified.
   ** sort the inversions by length so that we can apply shortest first and
      need not worry about what the indices mean (since they commute).
      (if we were to invert [1,4] first, [1,1] would become [4,4])
*/
inline bool comp_by_length(const Pair &p1, const Pair &p2)
  { return p1.size() > p2.size(); }
list<Pair>::iterator Ograph::applyCommutingSet(list<Pair> &set, bool force)
  {
  vector<Pair> setv;                               //Random access for sorting.
  std::copy(set.begin(), set.end(), back_inserter(setv));
  sort(setv.begin(), setv.end(), comp_by_length);  //Sort inversions by length.

  for(vector<Pair>::iterator i = setv.begin(); i != setv.end(); i++)
    {
    if(!force && !isSafe(*i))                      //If we find an unsafe op
      return find(set.begin(), set.end(), *i);     //then return a pointer.
    revAll(*i);                                    //Apply the inversions.
    }

  return set.end();
  }


//Re-apply the given set of commuting inversions to the permutation.
// (the indices in the -set- are valid only if applied from longest to
//  shortest -- this undoes the action done by applyCommutingSet())
inline bool comp_by_length_opposite(const Pair &p1, const Pair &p2)
  { return p1.size() < p2.size(); }
void Ograph::reApplyCommutingSet(const list<Pair> &set)
  {
  vector<Pair> setv;                               //Random access for sorting.
  std::copy(set.begin(), set.end(), back_inserter(setv));
  sort(setv.begin(), setv.end(), comp_by_length_opposite);
                                                   //Sort inversions by length.
  for(vector<Pair>::iterator i = setv.begin(); i != setv.end(); i++)
    revAll(*i);                                    //Apply the inversions.
  }




/*Do an unoriented reversal. Return the interval(inclusive) over which reversal
was done.
***
 * Precondition: inRev must be all falses with a length of seqsize
 * Postcondition: inRev will be all falses

***/
Pair Ograph::doUnorRev(int left, int right) {
  int node, i, len;
  vector<int> nodes;
  Pair currRev;

  revAll(left, right);            // Do reversal in -seq-, -seqinds-, and -gr-
  comps->findHurdles();

  for (i=left; i <= right; i++) { // mark graph nodes in reversal into inRev
    node = abs(seq[i]);
    inRev[node] = true;
  }

  len = seqsize-1;
  for (i=0; i < len; ++i)
    if (inRev[i] != inRev[i+1])
      nodes.push_back(i);

  for (i=left; i <= right; i++) { // unmark reversal's graph nodes in inRev
    node = abs(seq[i]);		  // - (reestablishes postcondition)
    inRev[node] = false;
  }

  adjAdjList(nodes);		// Update the matrix values for neighbors.

  adjParity(nodes);		// Update the parity values.

  currRev = Pair(left-1, right-1);
#ifdef SORTING_SEQUENCE
  if (globalSortSequence) {
    globalSortSequence->addUnorRev(currRev);
  }
#endif
  return currRev;
}



/*Convert a chord list to its corresponding Pair (inversion) list using the
map defined by chordTOinv.
*/
list<Pair> Ograph::cTOinv(const list<int> &set, const vector<ipair> &chordTOinv)
  {
  list<Pair> retval;
  for(list<int>::const_iterator i = set.begin(); i != set.end(); i++)
    retval.push_back(Pair(chordTOinv[*i].first, chordTOinv[*i].second));
  return retval;
  }

/*Convert a chord list to its corresponding set (inversion) using the
map defined by chordTOinv.
*/
list< set<int> > Ograph::cTOset(const list<int> &s, const vector<ipair> &chordTOinv) const
  {
  list< set<int> > retval;
  for(list<int>::const_iterator i = s.begin(); i != s.end(); i++)
    retval.push_back(invTOset(Pair(chordTOinv[*i].first, chordTOinv[*i].second)));
  return retval;
  }


/*Convert a Pair (inversion) list into a chord list.
The mapping is defined by -invTOchord-.
*/
list<int> Ograph::invTOc(const list<Pair> &l, const map<ipair,int> &invTOchord)
  {
  list<int> retlist;
  for(list<Pair>::const_iterator i = l.begin(); i != l.end(); i++)
    retlist.push_back(invTOchord.find(ipair(i->getLeft(), i->getRight()))->second);

  return retlist;
  }

//Invert the int to ipair mapping defined by -chordTOinv-.
map<ipair,int> Ograph::invertMap(const vector<ipair> &chordTOinv)
  {
  map<ipair,int> retval;
  for(unsigned int i=0; i<chordTOinv.size(); i++)
    retval[chordTOinv[i]] = i;

  return retval;
  }



/*Try to find an independent set one smaller than -is- that doesn't include at
least 1 from -is-.
Return true and modify -is- if we find one that is noninterfering for this
and for -og-.
*/
bool Ograph::tryISwithout(const Graph &g, list<int> &is,
                          const vector<ipair> &chordTOinv,
                          const Ograph* og) const
  {
  Graph tg = g;
  for(list<int>::iterator i = is.begin(); i != is.end(); i++)
    {
    hash_set<int> adj = tg.remVertex(*i);//Create a graph without p and
    list<int> nmis = tg.getMISEX(is.size()-1);//look for a IS one smaller.
    if(nmis.size() == is.size()-1 &&
       isNonint(cTOinv(nmis, chordTOinv)) &&
       (og == NULL || og->isNonint(cTOinv(nmis, chordTOinv))))
      {                                      //If we found a good IS
      is = nmis;                             //then we copy it.
      return true;
      }
    tg.addVertex(*i, adj);
    }

  return false;
  }


/*Restrict the given chord models to have only the mutual inversions.
Return the restricted chord to inversion (ipair) mapping.
*/
vector<ipair> Ograph::restrictTOcommon(vector<int> &c1, vector<int> &s1,
                                       const vector<ipair> &cTOi1,
                                       vector<int> &c2, vector<int> &s2,
                                       const vector<ipair> &cTOi2)
  {
  vector<ipair> cTOiM;         //The mutual chord to inversion mapping.
  map<ipair, int> iTOc1 = invertMap(cTOi1);
  map<ipair, int> iTOc2 = invertMap(cTOi2);

                //Find the mutual (intersection of) ops:
  set<ipair> minv;             //Get the mutual inversions:
  for(unsigned int ch=0; ch < cTOi1.size(); ch++)
    {                                    //For each chord in 1,
    vector<ipair>::const_iterator inv = find(cTOi2.begin(), cTOi2.end(),
                                             cTOi1[ch]);
    if(inv != cTOi2.end() &&             //if its inversion is in 2
       find(c1.begin(), c1.end(), iTOc1[*inv]) != c1.end()&&//and it's safe in 1
       find(c2.begin(), c2.end(), iTOc2[*inv]) != c2.end()) //and in 2,
      minv.insert(cTOi1[ch]);            //add it to the set of mutual ops.
    }

  hash_set<int> mch;            //Get the mutual chords:
  for(set<ipair>::iterator i = minv.begin(); i != minv.end(); i++)
    mch.insert(iTOc1[*i]);
  
  hash_map<int,int> cmap1;           //For remapping one chord to another.
  for(unsigned int i=0; i < cTOi1.size(); i++)
    if(mch.find(i) != mch.end())     //If the chord is mutual
      {
      cmap1[i] = cTOiM.size();       //then assign it the next largest label
      cTOiM.push_back(cTOi1[i]);     //and save the mutual chord to inv map.
      }



                //Restrict the chord models for 1 to mutual chords:
  map<ipair, int> iTOcM = invertMap(cTOi1);

  vector<int> newcm;
  for(vector<int>::iterator i=c1.begin(); i!=c1.end(); i++)
    if(mch.find(*i) != mch.end())    //If the current chord is a mutual chord
      newcm.push_back(cmap1[*i]);    //then keep it.
  c1 = newcm;
  newcm.clear();
  for(vector<int>::iterator i=s1.begin(); i!=s1.end(); i++)
    if(mch.find(*i) != mch.end())    //If the current chord is a mutual chord
      newcm.push_back(cmap1[*i]);    //then keep it.
  s1 = newcm;



                //Remap mutual chords in 2 to be consistent with 1:
  hash_map<int,int> cmap2;                //For each c->p mapping in minv,
  for(set<ipair>::iterator i=minv.begin(); i!=minv.end(); i++)
    cmap2[iTOc2[*i]] = cmap1[iTOcM[*i]];  //record the mapping.

  newcm.clear();                       //For each chord endpoint in c2,
  for(vector<int>::iterator i=c2.begin(); i!=c2.end(); i++)
    if(cmap2.find(*i) != cmap2.end())  //if it is mutual,
      newcm.push_back(cmap2[*i]);      //then map it.
  c2 = newcm;
  newcm.clear();                       //For each chord endpoint in s2,
  for(vector<int>::iterator i=s2.begin(); i!=s2.end(); i++)
    if(cmap2.find(*i) != cmap2.end())  //if it is mutual,
      newcm.push_back(cmap2[*i]);      //then map it.
  s2 = newcm;

  return cTOiM;
  }


/*Return true if there is a nonzero intersection.
*/
bool Ograph::shareInvs(const list<Pair> &s1, const list<Pair> &s2)
  {
  list<Pair> set;
  set_intersection(s1.begin(), s1.end(), s2.begin(), s2.end(),
                   insert_iterator< list<Pair> >(set,set.begin()));
  return set.end() != set.begin();
  }





//%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Hurdles %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



/* Get rid of all hurdles. Return number of reversals it takes. (Joel)
*/
int Ograph::resolveHurdles(vector<Pair> &revs) {
  int numHurdles = comps->getNumHurdles();
  Hurdle hurdle;
  int left;
  const int nullHurdle = CompInfo::nullHurdle.comp;
  int hurdleToSkip = nullHurdle;
  bool mergesRemain = true;
  bool cutDone = true;

  inRev.clear();
  inRev.resize(seqsize, false); // establish precondition for doUnorRev()

  // If odd number of hurdles
  if (numHurdles % 2 == 1)
    cutDone = doFirstCut(revs, hurdleToSkip);

  // Merge all other hurdles (non-consecutive merges until last two hurdles)
  if (numHurdles > 1) {
    left = getFirstLeft(hurdleToSkip);

    while (mergesRemain && hurdleToSkip != nullHurdle)
      mergesRemain = mergeNextTwo(revs, left, hurdleToSkip);

    while (mergesRemain)
      mergesRemain = mergeNextTwo(revs, left);
  }

  // if there was a non-fortress with odd # of hurdles and a maximal simple
  // hurdle, cut the maximal hurdle now
  if (!cutDone)
    cutMaxHurdle(revs);

  // findOvMatHaveMem();		// reconstruct the overlap matrix
  for(int i = 0; i < matsize; ++i)
    s[i] = compScore(i);	// recalculate the score array from scratch
  
  return (numHurdles+1)/2;	// return the number of reversals done
}


/* Do the first cut/merge of a graph with an odd number of hurdles
 *  If the cut is performed, returns true.
 *  If cut should be done after all other merges, returns false.
 */
bool Ograph::doFirstCut(vector<Pair> &revs, int &hurdleToSkip) {
  Hurdle h;

  // If genome is a fortress
  if (comps->isFortress()) {
    comps->get1stHurdle();	// skip first hurdle
    h = comps->getNextHurdle();	// get second hurdle
    hurdleToSkip = h.comp;

    // merge 2nd hurdle and its neo-hurdle (outermost overarching component)
    revs.push_back(doUnorRev(comps->getNeoLM(), h.min));
  }
  else {
    h = comps->get1stSimHur();	// first simple hurdle
    hurdleToSkip = h.comp;

    // if h is a maximal hurdle
    if (comps->hasMaximal() && h==comps->get1stHurdle())
      return false;		// cut it after all the merges

    // cut the simple hurdle
    // (if left side of hurdle is node i, reverse nodes between i and i+1)
    revs.push_back(doUnorRev(h.min+1, gr[h.min*2].desire.id/2));
  }

  return true;			// cut complete
}


/* Return the left index of the first hurdle-merging reversal
    (if the hurdle-to-skip is passed, set variable to null)
 */
int Ograph::getFirstLeft(int &hurdleToSkip) {
  bool ignoreMaximal = false;
  int nullHurdle = CompInfo::nullHurdle.comp;
  Hurdle h = comps->get1stHurdle();

  if (h.comp==hurdleToSkip) {
    hurdleToSkip = nullHurdle;
    h = comps->getNextHurdle();
    ignoreMaximal = true;
  }

  if (!ignoreMaximal && comps->hasMaximal())
    return comps->getMaxLM();	// left of first rev is max of left side of the
				//  maximal hurdle
  else
    return h.max;		// left of first rev is 1st hurdle's max
}


/* Merge the next two hurdles (skipping the first-cut hurdle if necessary)
 * Returns true if more merges remain to be done
 */
bool Ograph::mergeNextTwo(vector<Pair> &revs, int &left, int &hurdleToSkip) {
  int mid, right;
  int nullHurdle = CompInfo::nullHurdle.comp;

  Hurdle h = comps->getNextHurdle();

  if (h.comp==hurdleToSkip) {
    hurdleToSkip = nullHurdle;
    h = comps->getNextHurdle();
  }

  if (h.isNull) 
    cerr << "++++Ran out of hurdles mid-merge" << endl;
  mid = h.min;			// mid is min of next hurdle

  h = comps->getNextHurdle();
  if (h.comp==hurdleToSkip && hurdleToSkip!=nullHurdle) {
    hurdleToSkip = nullHurdle;
    h = comps->getNextHurdle();
  }

  if (h.isNull) {		// if last merge
    revs.push_back(doUnorRev(left, mid));
    return false;		// no merges remain
  }
  else {			// if not last merge
    right = h.min;		// right is min of next hurdle
    revs.push_back(doUnorRev(left, right));
    left += right-mid;		// mid (in new position) becomes new left
  }

  return true;			// merges remain
}


/* Merge the next two hurdles
 * Returns true if more merges remain to be done
 */
bool Ograph::mergeNextTwo(vector<Pair> &revs, int &left) {
  int mid, right;
  Hurdle h = comps->getNextHurdle();

  if (h.isNull)
    cerr << "++++Ran out of hurdles mid-merge" << endl;
  mid = h.min;			// mid is min of next hurdle

  h = comps->getNextHurdle();
  if (h.isNull) {		// if last merge
    revs.push_back(doUnorRev(left, mid));
    return false;		// no merges remain
  }
  else {			// if not last merge
    right = h.min;		// right is min of next hurdle
    revs.push_back(doUnorRev(left, right));
    left += right-mid;		// mid (in new position) becomes new left
  }

  return true;			// merges remain
}


/* Cut the maximum hurdle in a non-fortress genome with an odd # of hurdles.
 */
void Ograph::cutMaxHurdle(vector<Pair> &revs) {
  int left = comps->getMaxRM();
  int mid = gr[left*2].desire.id/2;

  // cut starts on the inner right side of the maximal hurdle and may wrap
  // around the end of the genome
  if (left < mid)
    revs.push_back(doUnorRev(left+1, mid));
  else
    revs.push_back(doUnorRev(mid+1, left));
}





//%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Bergeron %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


/*Determine if a vertex is oriented.
*/
bool Ograph::isOriented(int vertex)
  {
  int parity = 0;
  bool* adjArray;
  adjArray = matrix[vertex];
  for(int i = 0; i < matsize; ++i)
    if(adjArray[i])
      ++parity;

  //fprintf(stderr, "%i %i %i \n", vertex, parity, (parity % 2)); 
  return (parity % 2)!=0;
  }

/*Compute the score for a vertex.
*/
int Ograph::compScore(int vertex)
  {
  int i, score;
  bool* adjArray;

  adjArray = matrix[vertex];
  score = 0;

  for(i = 0; i < matsize; ++i)
    if(adjArray[i])
      {
      if(p[i])           //**isOriented**
        --score;
      else
        ++score;
      }
  
  return score;
  }

/*Get the vertex index with the maximal score.
*/
int Ograph::getMaxScoreVert()
  {
  bool noVal = true;
  int maxVert = 0, i;
  for(i = 0; i < matsize; ++i)
    if(p[i] && (noVal || (s[i] > s[maxVert])))
      {
      maxVert = i;
      noVal = false;
      }

  return maxVert;
  }

/*Adjust the score array given vertex v was changed.
*/
void Ograph::adjScoreArray(int maxScoreVert)
  {
  bool* adj = matrix[maxScoreVert];

  for(int i = 0; i < matsize; ++i)
    if(adj[i])
      ++s[i];

  return;
  }

/*Adjust the adjacency lists when a vertex -vert- is changed.
*/
void Ograph::adjAdjList(int vert)
  {
  int i;
  bool *v;
  bool *w;

  v = matrix[vert];

  v[vert] = 1;
  for(i = 0; i < matsize; ++i)
    {
    if((i != vert) && v[i])
      {
      w = matrix[i];

      if(!p[i])         //**isOriented**
        adjForW(i, w, v, -1);
      else
        adjForW(i, w, v, 1);
      }
    }

  return;
  }

/*Adjust the adjacency matrix after an unoriented reversal,
 * given the overlap nodes which have had their parity reversed.
 */
void Ograph::adjAdjList(vector<int> nodes) {
  int row, col;
  int size = nodes.size();
  for (int i=size-1; i > 0; --i) {
    row = nodes[i];
    for (int j=i-1; j >= 0; --j) {
      col = nodes[j];
      matrix[col][row] = matrix[row][col] = !matrix[row][col];
    }
  }
}



/*Adjust the adj for the w in the paper.
*/
void Ograph::adjForW(int wInd, bool *w, bool *v, int gainLoss)
  {
  int i;

  //update the score
  for(i = 0; i < matsize; ++i)
    if(w[i])
      s[i] = s[i] + gainLoss;

  w[wInd] = 1;

/*
/--------------------------------------------------------
  fprintf(stderr, "WIND:%i\nv:", wInd);
  for(int j = 0; j < matsize; ++j)
    fprintf(stderr, "%i ", v[j]);
  
  fprintf(stderr, "w:");
  for(int j = 0; j < matsize; ++j)
    fprintf(stderr, "%i ", w[j]);
  //fprintf(stderr, "\n");
/--------------------------------------------------------
*/

  //adjust w with respect to v
  for(i = 0; i < matsize; ++i)
    w[i] = (v[i] && !w[i]) || (!v[i] && w[i]);

/* 
/--------------------------------------------------------
  fprintf(stderr, "\nw:");
  for(int j = 0; j < matsize; ++j)
    fprintf(stderr, "%i ", w[j]);
  fprintf(stderr, "\n");

  fprintf(stderr, "v:");
  for(int j = 0; j < matsize; ++j)
    fprintf(stderr, "%i ", v[j]);
  fprintf(stderr, "\n");
/--------------------------------------------------------
*/
  
  //do the score thing again
  for(i = 0; i < matsize; ++i)
    if(w[i])
      s[i] = s[i] + gainLoss;
  }


/*Adjust the parity.
*/
void Ograph::adjParity(int v)
  {
  int i;
  bool* vAdj;
  
  vAdj = matrix[v];
  for(i = 0; i < matsize; ++i)
    p[i] = (vAdj[i] && !p[i]) || (!vAdj[i] && p[i]);

  return;
  }



/* Fill -matrix- with Bergeron's overlap matrix for this graph as well as
the score and parity arrays.
*/
void Ograph::findOverlapMat()
  {                                     //Allocate the square matrix:
  matrix = (bool **)malloc(matsize*sizeof(bool*));
  for(int i=0; i < matsize; i++)
    matrix[i] = new bool[matsize];
  s = new int[matsize];                 //Allocate the score and parity arrays:
  p = new bool[matsize];

  findOvMatHaveMem();
  }



/* Fill -matrix- with Bergeron's overlap matrix for this graph as well as
the score and parity arrays. (when memory is already allocated)
*/
void Ograph::findOvMatHaveMem() {
  int i;
  for(i=0; i < matsize; i++)	       //Initialize the matrix:
    for(int j=0; j < matsize; j++)
      matrix[i][j] = 0;
    

  for(i=0; i < graphsize-1; i++)		      //For each
    if(i < gr[i].desire.id)                           //edge
      for(int j = i+1; j < gr[i].desire.id; j++)      //check for overlapping
        if((gr[j].desire.id < i) || (gr[j].desire.id > gr[i].desire.id))
	  {
	    //fprintf(stderr, "\n\n\n--------------------------> %i %i %i\n", matsize, gr[i].label/2, gr[j].label/2);
	    matrix[gr[i].label/2][gr[j].label/2] = true;  //and mark it if so.
	  }

  // fill the score and parity arrays
  for(i = 0; i < matsize; ++i)	       // compScore() uses the parities
    p[i] = isOriented(i);
  for(i = 0; i < matsize; ++i)	       //so do it second.
    s[i] = compScore(i);

  /*
  // print the matrix
  cout << "Ograph matrix:" << endl;
  for(i=0; i < matsize; ++i) {
    for(int j=0; j < matsize; j++)
      cout << matrix[i][j] << " ";
    cout << endl;
  }
  cout << "Parity:" << endl;
  for(i=0; i < matsize; ++i)
    cout << p[i] << " ";
  cout << endl << "Score:" << endl;
  for(i=0; i < matsize; ++i)
    cout << s[i] << " ";
  cout << endl;
  */
}



/*Adjust the parity after an unoriented reversal,
 * given the overlap graph nodes which have had their parity reversed.
 */
void Ograph::adjParity(vector<int> nodes) {
  int node;
  int size = nodes.size();
  for (int i=0; i < size; ++i) {
    node = nodes[i];
    p[node] = !p[node];
  }
}


/*Is the parity 0 for all the verts.
*/
bool Ograph::allZeroParity()
  {
  for(int i = 0; i < matsize; ++i)
    if(p[i])
      return false;

  return true;
  }

/*Is the score 0 for all the verts. (There are hurdles)
*/
bool Ograph::allZeroScore()
  {
  for(int i = 0; i < matsize; ++i)
    if(s[i])
      return false;

  return true;
  }



//%%%%%%%%%%%%%%%%%%%%%%%%%%%%% TRANSPOSITIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


/*Do the specified transposition. Elements are 1 indexed.
   ** We just annihilate everything and rebuild the object based on our
      calculated new sequence.
*/
void Ograph::doTrans(int left, int right, int before)
  {
  vector<int> s;
  s.resize(seqsize-2);
  for(int i=1; i < seqsize-1; i++)
    s[i-1] = seq[i];

  left -= 1;
  right -= 1;
  before -= 1;
  vector<int> p;
  if(before < left)
    {
    p.insert(p.end(), s.begin(), s.begin()+before);
    p.insert(p.end(), s.begin()+left, s.begin()+right+1);
    p.insert(p.end(), s.begin()+before, s.begin()+left);
    p.insert(p.end(), s.begin()+right+1, s.end());
    }
  else
    {
    p.insert(p.end(), s.begin(), s.begin()+left);
    p.insert(p.end(), s.begin()+right+1, s.begin()+before);
    p.insert(p.end(), s.begin()+left, s.begin()+right+1);
    p.insert(p.end(), s.begin()+before, s.end());
    }

  int temp[p.size()];
  for(unsigned i=0; i < p.size(); i++)
    temp[i] = p[i];
  del();
  construct(temp,p.size());
  }

/*Do a transposition that joins three cycles into one, otherwise return false.
   ** Do srandom((unsigned int)time(NULL)) first.
*/
bool Ograph::doThreeMergeTrans()
  {
  int cn = comps->getNumCycles();   //Get the three cycles to merge:
  if(cn < 3)
    return false;
  int r1 = random() % cn;
  int r2 = r1;
  while(r2 == r1)
    r2 = random() % cn;
  int r3 = r2;
  while((r3 == r1) || (r3 == r2))
    r3 = random() % cn;

  vector<int> n;
  n.push_back(comps->getRandomNodeOnCycle(r1));
  if(n[0] % 2)
    n[0] -= 1;
  n.push_back(comps->getRandomNodeOnCycle(r2));
  if(n[1] % 2)
    n[1] -= 1;
  n.push_back(comps->getRandomNodeOnCycle(r3));
  if(n[2] % 2)
    n[2] -= 1;

  sort(n.begin(), n.end());

  int left,right,before;
  if(random() % 2)
    {
    left = n[0]/2 + 1;
    right = n[1]/2;
    before = n[2]/2+1;
    }
  else
    {
    before = n[0]/2 + 1;
    left = n[1]/2 + 1;
    right = n[2]/2;
    }

  doTrans(left,right,before);

  return true;
  }

//Do a transposition that splits two new cycles, otherwise return false.
bool Ograph::doTwoSplitTrans()
  {
/*
  vector<bool> visited;
  visited.resize(graphsize);
  for(int i=0; i<graphsize;)   //Work on cycle starting at i:
    {
    int one=i;                    //Visit possible first endpoints on the cycle:
    visited[one] = true;
    visited[one+1] = true;
    int other = gr[one].desire.id;
    while(other != one+1)
      {
      if((one < other) && (isTwoSplitTrans(one,other)))
        return true;
      other = gr[gr[one].reality.id].desire.id; //Advance the reality edge.
      }
    one = gr[one].desire.id;
    while(++i < graphsize && visited[i]);       //Advance to the next cycle.
    }
*/
  ipair interval;
  int before = comps->getTwoSplitTrans(interval);
  if(before)
    {
    doTrans(interval.first, interval.second, before);
    return true;
    }

  return false;
  }


bool Ograph::isTwoSplitTrans(int left, int right)
  {
  cerr << left << "," << right << endl;
  return false;
  }





//%%%%%%%%%%%%%%%%%%%%%%%%%%%%% OUTPUT %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ostream& operator<<(ostream& output, Ograph& c)
  {
  int i;
  cerr << "| ";
  for(i=0; i < c.seqsize; i++)
    cerr << c.seq[i] << " ";
  cerr << "|";
  for(i=0; i < c.matsize; i++)
    {
    cerr << endl << i << ":";
    for(int j=0; j < c.matsize; j++)
      {
      cerr << c.matrix[i][j] << " ";
      }
    }
  cerr << endl << "-\ns:";
  for(i=0; i < c.matsize; i++)
    cerr << c.s[i] << " ";
  cerr << endl << "p:";
  for(i=0; i < c.matsize; i++)
    cerr << c.p[i] << " ";

  return output;
  }

/* Output the graph is svg format.
void Ograph::printGraphSVG(ostream& filehandle) const
  {
  std::vector<int> g(&seq[1], &seq[seqsize-1]);

  cmb::bp::drawSVG(filehandle, g);
  }
*/


void Ograph::printGraph() const
  {
  int i;

  fprintf(stderr, "--\n");
  fprintf(stderr, "Graph Size = %i\n", graphsize);
  fprintf(stderr, "Seq Size = %i\n", seqsize);
  fprintf(stderr, "Mat Size = %i\n", matsize);
  for(i=0; i < graphsize; i++)
    cerr << gr[i].label << " ";
  cerr << endl;
  for(i=2; i < graphsize; i+=2)
    {
    if(gr[i].label%2)
      {
      cerr << "-";
      cerr << gr[i-1].label/2 << " ";
      }
    else
      cerr << gr[i].label/2 << " ";
    }
  fprintf(stderr, "\n--\n");

/*
  for(i=0; i < graphsize; i++)
    fprintf(stderr, "%d - name:%d real:%d pos?%d des:%d\n", i, gr[i].label, gr[i].reality.id, gr[i].sign, gr[i].desire.id);
*/
  }

/* Print the sequence to stderr.
*/
void Ograph::printSeqStr() const
  {
  for(int i=1; i < seqsize-1; i++)
    cerr << seq[i] << " ";
  }

/* Print the cycles.
*/
void Ograph::printCycles() const
  { comps->printCycles(); }

/* Print the cycles.
*/
void Ograph::printCompInfo() const
  { cerr << *comps << endl; }

/* Print the sequence to stderr with '*' to indicate FORWARDS oriented edges.
*/
void Ograph::printOrientationStr() const
  {
  if(gr[0].reality.direction == FORWARDS)
    cerr << "*";

  for(int i=1; i < seqsize-1; i++)
    {
    cerr << seq[i];
    if(gr[i*2].reality.direction == FORWARDS)
      cerr << "*";
    else
      cerr << " ";
    }
  }

void Ograph::printArray(const int a[], int seqsize) const
  {
  int i;

  for(i=0; i < seqsize; i++)
    fprintf(stderr, "%d ", a[i]);

  fprintf(stderr, "\n");
  }


//-----TEMP DEBUG OUTPUT-----//
void Ograph::printGrInfo() const {
  int i;

  cerr << "----" << endl << "seq:";
  for (i=0; i<seqsize; ++i)
    cerr << seq[i] << " ";
  cerr << endl << "gr.sign:";
  for (i=0; i<seqsize*2-2; ++i)
    cerr << gr[i].sign << " ";

  cerr << endl << "gr.label:";
  for (i=0; i<seqsize*2-2; ++i)
    cerr << gr[i].label << " ";
  cerr << endl << "gr.reality:";
  for (i=0; i<seqsize*2-2; ++i)
    cerr << gr[i].reality.id << " ";
  cerr << endl << "realityref:";
  for (i=0; i<seqsize*2-2; ++i)
    cerr << gr[gr[i].reality.id].label << " ";
  cerr << endl << "gr.reality.direction:";
  for (i=0; i<seqsize*2-2; ++i)
    cerr << gr[i].reality.direction << " ";

  cerr << endl << "gr.reality.deletion:";
  for (i=0; i<seqsize*2-2; ++i)
    cerr << gr[i].reality.deletion << " ";
  cerr << endl << "gr.reality.copies:  ";
  for (i=0; i<seqsize*2-2; ++i)
    cerr << gr[i].reality.copies << " ";

  cerr << endl << "gr.desire: ";
  for (i=0; i<seqsize*2-2; ++i)
    cerr << gr[i].desire.id << " ";

  cerr << endl << endl;

  cerr << *comps << endl;
  cerr << "----" << endl;
}

/* Return a string from the given input.
*/
template < class T >
string Ograph::ToString(const T &arg) const
  {
  ostringstream out;
  out << arg;

  return(out.str());
  }


