/*****************************************************************************
 *  FILE   : SortSeq.C
 *  AUTHOR : Joel V. Earnest-DeYoung
 *  DATE   : 11/2003
 *****************************************************************************
 *
 * Class SortSeq
 *  used in recording the sequence of operations to convert genome 1 into 
 *  genome 2 (as passed to multiMED)
 *
 * I'm using Krister and Mark's convention that
 *   deletions are genes in genome 1 but not in genome 2, and 
 *   insertions are genes in genome 2 but not genome 1.
 * What I call consensus genes are genes found in both genomes.
 * The reversal sequence -revs- converts the identity permutation into the
 *  consensus sequence -consensusG2-
 *
 *****************************************************************************/

#include "SortSeq.h"

// This file is only compiled if the SORTING_SEQUENCE flag is turned on
#ifdef SORTING_SEQUENCE

using namespace std;
using namespace __gnu_cxx;

// SortSeq object with global scope (this acts as a runtime flag for all the
//				     SortSeq routines)
SortSeq *globalSortSequence = NULL;
//SortSeq *globalSortSequence = new SortSeq;


/*****************************************************************************
 * LOCAL SWITCHES
 *****************************************************************************/
// #define VERBOSE


/*****************************************************************************
 * LOCAL FUNCTIONS
 *****************************************************************************/
// Return the input vector reversed and negated
vector<int> reverseGenomeSegment(const vector<int> &seg) {
  vector<int> reversedSeg;
  vector<int>::const_reverse_iterator segIt, segEnd=seg.rend();

  reversedSeg.reserve(seg.size());

  for (segIt = seg.rbegin(); segIt != segEnd; ++segIt) {
    reversedSeg.push_back(-(*segIt));
  }

  return reversedSeg;
}

/*****************************************************************************
 * CLASS DEFINITIONS
 *****************************************************************************/
// INSPOSNODE CLASS

InsPosNode::InsPosNode() {
  twoNew = false;
}

InsPosNode::InsPosNode(int p1) {
  newPos1 = p1;
  twoNew = false;
}

InsPosNode::InsPosNode(int p1, int p2) {
  newPos1 = p1;
  newPos2 = p2;
  twoNew = true;
}


// HURDLEINFO STRUCT
HurdleInfo::HurdleInfo(int l, int r, bool s) {
  left = l;
  right = r;
  isSuperHurdle = s;
}


// SORTSEQ CLASS

SortSeq::SortSeq() {
  firstHurdleRev = lastHurdleRev = -1;
  maxHurdle = maxHInnerLeft = maxHInnerRight = fortressRevPoint = -1;

  // Reserve memory for the sorting sequence 
  // -- These numbers are big enough for Joel's 13 bacterial genomes, though of
  //    course STL data structures will resize themselves if more is needed.
  //    THIS AMOUNT OF MEMORY WILL BE OVERKILL FOR SMALL EVOL. DISTANCES!!
  revs.reserve(1400);
  dels.reserve(300);
  inserts.reserve(400);
}

SortSeq::~SortSeq() {
  clear();
}

void SortSeq::clear() {
  genome1.clear();
  genome2.clear();
  renamedG2.clear();
  consensusG2.clear();
  renamedToGenome2.clear();
  consensusToRenamed.clear();
  revs.clear();
  dels.clear();
  inserts.clear();
  revConsensusG1.clear();
  revConsensusG2.clear();
  insSubsets.clear();
  insertionFinalLocs.clear();
  insertMatrix.clear();
  hurdles.clear();
  firstHurdleRev = lastHurdleRev = -1;
  maxHurdle = maxHInnerLeft = maxHInnerRight = fortressRevPoint = -1;
  insertPos.clear();
  deletePos.clear();
  stickyInsMap.clear();
  combinedConfigs.clear();
  combinedCycConfigs.clear();
  mutExConfigs.clear();
  mutExCycConfigs.clear();
}

void SortSeq::setGenome1(int* seq, int len) {
  setGenome(genome1, seq, len);
}

void SortSeq::setGenome2(int* seq, int len) {
  setGenome(genome2, seq, len);
}

void SortSeq::setRenamedG2(int* seq, int len) {
  int i, rGene, gGene;
  setGenome(renamedG2, seq, len);

  if (int(genome2.size()) != len)
    cerr << endl << "ERROR in SortSeq.C: setRenamedG2()" << endl
	 << " - Renamed genome is not the same size as the original!" << endl
	 << endl;
  renamedToGenome2.clear();
  renamedToGenome2.resize(2*len);

  // map renamedG2 to genome2
  for (i = 0; i < len; ++i) {
    rGene = renamedG2[i];
    gGene = genome2[i];
    renamedToGenome2[rGene] = gGene;
    renamedToGenome2[-rGene] = -gGene;
  }

  consensusToRenamed.clear();
}

void SortSeq::setConsensusG2(int* seq, int len) {
  setGenome(consensusG2, seq, len);
}

int SortSeq::getConsensusSize() const {
  return consensusG2.size();
}

// map a gene in consensusG2 to renamedG2
void SortSeq::mapConsensusToRenamed(int consGene, int renGene) {
  consensusToRenamed[consGene] = renGene;
  consensusToRenamed[-consGene] = -renGene;
}

// Add a reversal
void SortSeq::addRev(Pair rev) {
  revs.push_back(rev);
}

// Add an unoriented reversal. Only happens during hurdle merging or cutting.
void SortSeq::addUnorRev(Pair rev) {
  int index = revs.size();

  revs.push_back(rev);

  if (firstHurdleRev == -1)
    firstHurdleRev = index;
  lastHurdleRev = index;
}

// Add a hurdle.
// The hurdle pair should contain the leftmost and rightmost indices of a
// hurdle in the genome immediately prior to the first hurdle-clearing reversal
// (when converting consensus genome 1 to the identity)
void SortSeq::addHurdle(int left, int right, bool isSuper) {
  hurdles.push_back(HurdleInfo(left, right, isSuper));
}

// Given that a maximal hurdle exists, record the maximum of its left half and
// the minumum of its right half, and figure out which hurdle is the maximal
// hurdle.
void SortSeq::setMaximalHurdle(int innerLeft, int innerRight) {
  int hurdleNum, numHurdles = hurdles.size();

  maxHInnerLeft = innerLeft;
  maxHInnerRight = innerRight;

  maxHurdle = -1;
  for (hurdleNum = 0; hurdleNum < numHurdles && maxHurdle == -1;
       ++hurdleNum) {
    if (hurdles[hurdleNum].left < innerLeft 
	&& hurdles[hurdleNum].right > innerRight)
      maxHurdle = hurdleNum;
  }

  if (maxHurdle == -1)
    cerr << endl << "ERROR in SortSeq.C: setMaximalHurdle()" << endl
	 << " There should be maximal hurdle, but I can't find it" << endl
	 << endl;
}

// The fortress reversal point is the inner left index of the non-hurdle
// unoriented component that was merged with the hurdle it encloses.
// It is set only if the genome is a fortress. Otherwise it remains -1
void SortSeq::setFortressRevPoint(int neoInnerLeft) {
  fortressRevPoint = neoInnerLeft;
}

// Add an insertion.
// Inserts are added in the order they occur in renamedG2. The insert pair
// should contain
//   left: inserted gene's signed gene number in renamedG2
//   right: gene's position in renamedG2
void SortSeq::addInsert(Pair insert) {
  inserts.push_back(insert);
}

// Add a deletion pair
// Pair is [absval+1 of deleted gene's index in renamedG2, closest previous
//  common gene]
void SortSeq::addDelete(Pair del) {
  dels.push_back(del);
}

// Set the insertion matrix, as created in expandInsMatrix
void SortSeq::setInsMatrix(int** iMat, int rows, int cols) {
  insertMatrix.reserve(rows+1);
  for(int i=0; i < rows; ++i)
    insertMatrix.push_back(vector<int>(iMat[i],iMat[i]+cols));
}

// Add an insertion subset.
// Subset contains the numbers of a group of inserts. The numbers for each
// insert is the same as its index in -inserts-.
void SortSeq::addInsSubset(int* subset) {
  int last = subset[0];	   // subset[0] should be the index of its last element
  int rownum, i;
  bool rowMatches;
  vector<int> rows;
  vector< vector<int> >::iterator row;
  vector< vector<int> >::iterator firstRow = insertMatrix.begin();

  insSubsets.push_back(vector<int>(subset+1,subset+last+1));

  for (row=firstRow; row != insertMatrix.end(); ++row) {
    rowMatches = true;
    for (i=1; rowMatches && i <= last; ++i)
      rowMatches = ((*row)[subset[i]] > 0);
    if (rowMatches) {
      rownum = row - firstRow + 1;
      for (i=1; i <= last; ++i) {
  	insertionFinalLocs[subset[i]+1].insert(rownum);
      }
    }
  }

  // increment the insertion numbers in the newly added subset to match the
  // shifted numbering in -insertPos-, -insertionFinalLocs-, etc.
  vector<int>::iterator subsetIt, subsetEnd;
  for (subsetIt = insSubsets.rbegin()->begin(),
	 subsetEnd = insSubsets.rbegin()->end();
       subsetIt != subsetEnd; ++subsetIt) {
    ++(*subsetIt);
  }
}

// Calculates the state of the consensus genes after each reversal, in two ways
// - from consensusG2 to the identity permutation (in revConsensusG2)
// - from "consensusG1" to the identity permutation (in revConsensusG1), where
//   consensusG1 is the inverse of consensusG2 so that
//   if consensusG2[4]==-7 then consensusG1[7]==-4
void SortSeq::computeConsensusPositions() {
  int left, right, temp, i;
  int revNum, numRevs = revs.size();
  int numGenes = consensusG2.size();
  vector<int> genomeG1, genomeG2 = consensusG2;

  revConsensusG1.clear();
  revConsensusG1.resize(numRevs+1);
  revConsensusG2.clear();
  revConsensusG2.reserve(numRevs+1);

  genomeG1.resize(numGenes);
  for (i = 0; i < numGenes; ++i) {
    genomeG1[i] = i+1;
  }

  revConsensusG1[numRevs] = genomeG1;
  revConsensusG2.push_back(genomeG2);

  for (revNum = numRevs-1; revNum >= 0; --revNum) {
    left = revs[revNum].getLeft();
    right = revs[revNum].getRight();
    while(left <= right) {
      temp = genomeG1[left];
      genomeG1[left] = -genomeG1[right];
      genomeG1[right] = -temp;
      temp = genomeG2[left];
      genomeG2[left] = -genomeG2[right];
      genomeG2[right] = -temp;
      ++left;
      --right;
    }
    revConsensusG1[revNum] = genomeG1;
    revConsensusG2.push_back(genomeG2);
  }
}

// Calculate a map that tells what insertions started out (in genome 2) on
// either side of each consensus gene
void SortSeq::computeStickyInsMap() {
  int insNum, numIns = inserts.size();
  vector<int> insSegment;
  int pos, currPos;
  int numConsensus = consensusG2.size();
  // -pos- is the position in consensusG2 that the insertion is just left of
  if (numIns > 0) {
    int gene;
    pos = inserts[0].getRight();
    for (insNum = 0; insNum < numIns; ++insNum) {
      currPos = inserts[insNum].getRight() - insNum;
      if (currPos != pos) {
	if (pos < numConsensus)
	  gene = consensusG2[pos];
	else
	  gene = numConsensus+1;
	stickyInsMap[gene] = insSegment;
	insSegment.clear();
	pos = currPos;
      }
      insSegment.push_back(renamedToGenome2
			   [consensusToRenamed[inserts[insNum].getLeft()]]);
    }
    if (pos < numConsensus)
      gene = consensusG2[pos];
    else
      gene = numConsensus+1;
    stickyInsMap[gene] = insSegment;
  }

#ifdef VERBOSE
  hash_map<int,vector<int> >::const_iterator mIt, mEnd;
  cerr << "INSERTION MAP" << endl;
  for (mIt = stickyInsMap.begin(), mEnd = stickyInsMap.end(); mIt != mEnd;
       ++mIt) {
    cerr << "gene " << mIt->first << ":";
    vector<int>::const_iterator vIt, vEnd;
    for (vIt = mIt->second.begin(), vEnd = mIt->second.end(); vIt != vEnd;
	 ++vIt)
      cerr << *vIt << " ";
    cerr << endl;
  }
#endif // VERBOSE
}

// Calculate all possible positions for the given insertion after each
// reversal in the sorting sequence
//   Reversals here are considered in the order they convert genome 2 into
//   genome 1. In other words, from the end of -revs- to the beginning
void SortSeq::computeInsPositions(int insNum) {
  int revNum;
  int pos;
  int numRevs = revs.size();
  bool removePos1;
  IPosLevel2_iterator mapIt, mapEnd, nextMapEnd;
  IPosLevel2_t *nextMapPtr;
  iisHashMap::iterator finalLocEnd;
  vector<int> elimList;
  hash_set<int>::iterator setIt, setEnd;

  elimList.reserve(consensusG2.size());

  // -pos- is the position in consensusG2 that the insertion is just left of
  pos = inserts[insNum-1].getRight() - insNum + 2;
  reverseInsert(pos, revs[numRevs-1], insNum, 0);

  // calculate positions after each reversal for the insertion
  for (revNum = 1; revNum < numRevs; ++revNum) {
    mapEnd = insertPos[insNum][revNum-1].end();
    for (mapIt = insertPos[insNum][revNum-1].begin();
	 mapIt != mapEnd; ++mapIt) {
      pos = mapIt->second.newPos1;
      reverseInsert(pos, revs[numRevs-revNum-1], insNum, revNum);
      if (mapIt->second.twoNew) {
	pos = mapIt->second.newPos2;
	reverseInsert(pos, revs[numRevs-revNum-1], insNum, revNum);
      }
    }
  }

  // eliminate final positions inconsistent with -insertionFinalLocs-
  elimList.clear();
  mapEnd = insertPos[insNum][numRevs-1].end();
  for (mapIt = insertPos[insNum][numRevs-1].begin();
       mapIt != mapEnd; ++mapIt) {
    setEnd = insertionFinalLocs[insNum].end();
    pos = abs(mapIt->second.newPos1);

    // mark Pos1 for removal if it is not in -insertionFinalLocs-
    removePos1 = (insertionFinalLocs[insNum].find(pos) == setEnd);

    if(mapIt->second.twoNew) {
      pos = abs(mapIt->second.newPos2);

      // if Pos2 isn't in -insertionFinalLocs-, unmark -twoNew-
      if(insertionFinalLocs[insNum].find(pos) == setEnd) {
	mapIt->second.twoNew = false;
      }
    }

    if(removePos1) {
      if(mapIt->second.twoNew) {
	mapIt->second.newPos1 = pos;
	mapIt->second.twoNew = false;
      }
      // if map node should be removed, add it to -elimList-
      else {
	elimList.push_back(mapIt->first);
      }
    }
  }

  // erase all the nodes in -elimList- from -insertPos[insNum][numRevs-1]-
  while (!elimList.empty()) {
    insertPos[insNum][numRevs-1].erase(elimList.back());
    elimList.pop_back();
  }

  // run back along -insertPos- and eliminate inconsistent intermediate
  // positions
  for (revNum = numRevs-2; revNum >= 0; --revNum) {
    elimList.clear();
    nextMapPtr = &(insertPos[insNum][revNum+1]);
    nextMapEnd = nextMapPtr->end();
    mapEnd = insertPos[insNum][revNum].end();
    for (mapIt = insertPos[insNum][revNum].begin();
	 mapIt != mapEnd; ++mapIt) {
      pos = mapIt->second.newPos1;

      // mark Pos1 for removal if it is not in the next higher reversal row
      removePos1 = (nextMapPtr->find(pos) == nextMapEnd);

      if(mapIt->second.twoNew) {
	pos = mapIt->second.newPos2;

	// if Pos2 isn't in -insertionFinalLocs-, unmark -twoNew-
	if(nextMapPtr->find(pos) == nextMapEnd) {
	  mapIt->second.twoNew = false;
	}
      }

      if(removePos1) {
	if(mapIt->second.twoNew) {
	  mapIt->second.newPos1 = pos;
	  mapIt->second.twoNew = false;
	}
	// if map node should be removed, add it to -elimList-
	else {
	  elimList.push_back(mapIt->first);
	}
      }
    }

    // erase all the nodes in -elimList- from -insertPos[insNum][revNum]-
    while (!elimList.empty()) {
      insertPos[insNum][revNum].erase(elimList.back());
      elimList.pop_back();
    }
  }

  // copy the final locations from the "numRevs-1" map to the "numRevs" map
  nextMapPtr = &(insertPos[insNum][numRevs]);
  mapEnd = insertPos[insNum][numRevs-1].end();
  for (mapIt = insertPos[insNum][numRevs-1].begin(); mapIt != mapEnd;
       ++mapIt) {
    (*nextMapPtr)[mapIt->second.newPos1];
    if (mapIt->second.twoNew)
      (*nextMapPtr)[mapIt->second.newPos2];
  }
}

// ***** Helper function for -computeInsPositions()- *****
// Given an insertion's old position and a reversal Pair, determine the
//  insertion's new position(s) and record them in
//  -insertPos[insNum][revNum][pos]-
void SortSeq::reverseInsert(int pos, Pair &rev, int insNum, int revNum) {
  int left = rev.getLeft() + 1;
  int right = rev.getRight() + 2;
  int posVal = abs(pos);
  int opp_sign = (pos < posVal) ? 1 : -1;

  // insertion outside of reversal - keep it unchanged
  if (posVal<left || posVal>right) {
    insertPos[insNum][revNum][pos] = InsPosNode(pos);
  }

  // insertion inside of reversal - shift it and change sign
  else if (posVal>left && posVal<right) {
    insertPos[insNum][revNum][pos]
      = InsPosNode(opp_sign*(left+right-posVal));
  }

  // insertion on endpoint of reversal - split it to both ends
  else {
    insertPos[insNum][revNum][pos]
      = InsPosNode(pos, opp_sign*(left+right-posVal));
  }
}

// Calculate all possible positions for all deletions after each reversal in
// the sorting sequence
void SortSeq::computeDelPositions() {
  vector<Pair>::iterator delIt, delEnd = dels.end();
  vector<Pair>::iterator revIt, revEnd = revs.end();
  dPosLevel1_iterator dPosMapIt, dPosMapEnd, newPos;
  vector<int> tempSegment;
  map<int, set<int> > tempMap;
  map<int, set<int> >::iterator tempMapIt, tempMapEnd;
  set<int>::iterator setIt, setEnd;
  int revNum;
  int pos, left, right;
  bool isHurdle, moveRight;

  if (revConsensusG1.empty())
    computeConsensusPositions();

  // create a vector of maps num-of-reversals+1 long
  deletePos.clear();
  deletePos.resize(revs.size()+1);

  // create sets of all the deletions at their original position in genome1
  for (delIt = dels.begin(); delIt != delEnd; ++delIt) {
    pos = abs(delIt->getRight())+1;
				// deletion starts out right of this position
    tempMap[pos].insert(delIt->getLeft()-1);
				// add deletion indices so they end up sorted
  }
  // copy the sets into the beginning of -deletePos-
  // (translating deletion indices into the original gene numbers in genome1)
  tempMapEnd = tempMap.end();
  for (tempMapIt = tempMap.begin(); tempMapIt != tempMapEnd; ++ tempMapIt) {
    pos = tempMapIt->first;
    setEnd = tempMapIt->second.end();
    for(setIt = tempMapIt->second.begin(); setIt != setEnd; ++setIt) {
      deletePos[0][pos].push_back(genome1[*setIt]);
				// add each deletion in order to deletePos
    }
  }

  revNum = 0;
  for (revIt = revs.begin(); revIt != revEnd; ++revIt) {
    isHurdle = (firstHurdleRev > -1
		&& lastHurdleRev >= revNum
		&& firstHurdleRev <= revNum);
    ++revNum;
    left = revIt->getLeft()+1;
    right = revIt->getRight()+2;

    moveRight =			// move on-edge deletions to the right if
				// - reversal is hurdle-clearing,
				// - reversal produces a length-2 cycle on its
				//   left
      isHurdle
      || (left==1 && revConsensusG1[revNum][0]==1)
      || (revConsensusG1[revNum][left-2] == revConsensusG1[revNum][left-1]-1);
    dPosMapEnd = deletePos[revNum-1].end();
    for (dPosMapIt = deletePos[revNum-1].begin(); dPosMapIt != dPosMapEnd;
	 ++dPosMapIt) {
      pos = dPosMapIt->first;
      if (pos < left || pos > right)		// outside reversal
	deletePos[revNum][pos] = dPosMapIt->second;
      else if (pos > left && pos < right)	// inside reversal
	deletePos[revNum][left+right-pos]
	  = reverseGenomeSegment(dPosMapIt->second);
      else if (pos == left) {			// on left edge of reversal
	if (moveRight)
	  deletePos[revNum][right]
	    = reverseGenomeSegment(dPosMapIt->second);
	else
	  deletePos[revNum][pos] = dPosMapIt->second;
      }
      else {					// on right edge of reversal
	if (moveRight) {
	  tempSegment = dPosMapIt->second;
	  newPos		// get an iterator to the correct position
	    = (deletePos[revNum].insert(dPosLevel1_pair
		(right, dPosLevel2_t()))).first;
	}
	else {
	  tempSegment = reverseGenomeSegment(dPosMapIt->second);
	  newPos		// get an iterator to the correct position
	    = (deletePos[revNum].insert(dPosLevel1_pair
		(left, dPosLevel2_t()))).first;
	}
	newPos->second.insert(newPos->second.end(), tempSegment.begin(),
			      tempSegment.end());
      }
    }
  }
#ifdef VERBOSE
  printDeletePos();
#endif //VERBOSE
}

// Return the number of reversals
int SortSeq::getNumRevs() {
  return revs.size();
}

// Return the number of deletions
int SortSeq::getNumDels() {
  return dels.size();
}

// Return the number of insertions
int SortSeq::getNumInserts() {
  return inserts.size();
}

// Return the number of hurdles
int SortSeq::getNumHurdles() {
  return hurdles.size();
}

// Copy a the given genome sequence into a (private) class genome
void SortSeq::setGenome(vector<int> &genome, int *seq, int len) {
  genome.clear();
  for (int i=0; i < len; ++i)
    genome.push_back(seq[i]);
}

// Returns true if consensusG2 is a fortress
bool SortSeq::isFortress() {
  return (fortressRevPoint > -1);
}

void SortSeq::printInsMatrix() {
  vector< vector<int> >::iterator row;
  vector<int>::iterator col;
  int rowNum;

  for(row = insertMatrix.begin(); row != insertMatrix.end(); ++row) {
    rowNum = row-insertMatrix.begin();
    fprintf(stderr, "%.3d- ", rowNum);
    for(col = row->begin(); col != row->end(); ++col)
      fprintf(stderr, "%d ", *col);
    fprintf(stderr, "-%.3d\n", rowNum);
  }
}

void SortSeq::printInsSubsets() {
  vector< vector<int> >::iterator outerIt, outerEnd=insSubsets.end();
  vector<int>::iterator subsetIt, subsetEnd;

  cerr << "Insertion Subsets:" << endl;
  for (outerIt = insSubsets.begin(); outerIt != outerEnd; ++outerIt) {
    subsetEnd = outerIt->end();
    for (subsetIt = outerIt->begin(); subsetIt != subsetEnd; ++subsetIt) {
      cerr << *subsetIt << " ";
    }
    cerr << endl;
  }
}

void SortSeq::printInsFinalLocs() {
  iisHashMap::iterator it;
  hash_set<int>::iterator setIt, setEnd;
  cerr << "Insertion Final Locations" << endl;
  for (it = insertionFinalLocs.begin(); it != insertionFinalLocs.end(); ++it) {
    setEnd = it->second.end();
    for (setIt = it->second.begin(); setIt != setEnd; ++setIt) {
      cerr << setw(5) << it->first << " = " << setw(5) << *setIt << endl;
    }
  }
}

void SortSeq::printGenomes() {
  vector<int>::const_iterator vIt, vEnd;
  cerr << "Genome 1:" << endl;
  for (vIt = genome1.begin(), vEnd = genome1.end(); vIt != vEnd; ++vIt)
    cerr << *vIt << " ";
  cerr << endl;

  cerr << "Genome 2:" << endl;
  for (vIt = genome2.begin(), vEnd = genome2.end(); vIt != vEnd; ++vIt)
    cerr << *vIt << " ";
  cerr << endl;

  cerr << "Renamed G2:" << endl;
  for (vIt = renamedG2.begin(), vEnd = renamedG2.end(); vIt != vEnd; ++vIt)
    cerr << *vIt << " ";
  cerr << endl;

  cerr << "Consensus G2:" << endl;
  for (vIt = consensusG2.begin(), vEnd = consensusG2.end(); vIt != vEnd; ++vIt)
    cerr << *vIt << " ";
  cerr << endl;
}

void SortSeq::printExecutedReversals() {
  int geneNum, numGenes = consensusG2.size();
  int revIndex, revNum, numRevs = revs.size();
  int left, right;
  if (revConsensusG1.empty() || revConsensusG2.empty())
    computeConsensusPositions();

  cerr << "GLOBALSORTSEQUENCE REVERSAL CHECK" << endl;
  cerr << "REVERSALS" << endl;
  for (revNum = 0; revNum < numRevs; ++revNum) {
    if (revNum == firstHurdleRev) {
      if (revNum > 0)
	cerr << endl;
      cerr << "hurdle-clearing:" << endl;
      }
    cerr << "(" << revs[revNum].getLeft() << "," 
	 << revs[revNum].getRight() << ") ";
    if (revNum == lastHurdleRev)
      cerr << endl;
  }
  cerr << endl;

  cerr << "GENOME 2 TO IDENTITY" << endl;
  for (geneNum = 0; geneNum < numGenes; ++geneNum) {
    cerr << revConsensusG2[0][geneNum] << " ";
  }
  cerr << endl;

  for (revIndex = 1; revIndex <= numRevs; ++revIndex) {
    revNum = numRevs - revIndex;
    left = revs[revNum].getLeft();
    right = revs[revNum].getRight();
    cerr << "reverse (" << left << "," << right << ")";
    if (firstHurdleRev > -1 
 	&& lastHurdleRev >= revNum
	&& firstHurdleRev <= revNum)
      cerr << " [hurdle-clearing]";
    cerr << endl;
    for  (geneNum = 0; geneNum < numGenes; ++geneNum) {
      cerr << revConsensusG2[revIndex][geneNum] << " ";
    }
    cerr << endl;
  }

  cerr << "GENOME 1 TO IDENTITY" << endl;
  for (geneNum = 0; geneNum < numGenes; ++geneNum) {
    cerr << revConsensusG1[0][geneNum] << " ";
  }
  cerr << endl;

  for (revNum = 0; revNum < numRevs; ++revNum) {
    left = revs[revNum].getLeft();
    right = revs[revNum].getRight();
    cerr << "reverse (" << left << "," << right << ")";
    if (firstHurdleRev > -1 
 	&& lastHurdleRev >= revNum
	&& firstHurdleRev <= revNum)
      cerr << " [hurdle-clearing]";
    cerr << endl;
    for  (geneNum = 0; geneNum < numGenes; ++geneNum) {
      cerr << revConsensusG1[revNum+1][geneNum] << " ";
    }
    cerr << endl;
  }
}

// Print out the contents of -insertPos-
void SortSeq::printInsertPos() {
  IPos_iterator insIt, insBegin, insEnd;
  IPosLevel1_iterator revIt, revBegin, revEnd;
  IPosLevel2_iterator mapIt, mapBegin, mapEnd;
  int revNum, numRevs = revs.size();

  if (inserts.empty())
    return;

  cerr << "INSERTPOS DATA STRUCTURE" << endl;

  insBegin = insertPos.begin();
  insEnd = insertPos.end();
  for (insIt = insBegin; insIt != insEnd; ++insIt) {
    cerr << "Inserted gene " << insIt-insBegin << endl;
    revBegin = insIt->begin();
    revEnd = insIt->end();
    revNum = 0;
    for (revIt = revBegin; revIt != revEnd; ++revIt) {
      cerr << "  Reversal " << revIt-revBegin << endl;
      mapBegin = revIt->begin();
      mapEnd = revIt->end();
      for (mapIt = mapBegin; mapIt != mapEnd; ++mapIt) {
	cerr << "   " << setw(4) << mapIt->first;
	if (revNum < numRevs) {
	  cerr << " --> " << setw(4) << mapIt->second.newPos1;
	  if(mapIt->second.twoNew)
	    cerr << "," << setw(4) << mapIt->second.newPos2;
	}
	cerr << endl;
      }
      ++revNum;
    }
  }
}

// Print out the contents of -deletePos-
void SortSeq::printDeletePos() {
  int revNum, numRevs = revs.size();
  dPos_iterator dPosIt, dPosEnd = deletePos.end();
  dPosLevel1_iterator mapIt, mapEnd;
  dPosLevel2_iterator vectIt, vectEnd;

  if (dels.empty())
    return;

  cerr << "DELETEPOS DATA STRUCTURE" << endl;
  revNum = numRevs;
  for (dPosIt = deletePos.begin(); dPosIt != dPosEnd; ++dPosIt) {
    if (revNum < numRevs)
      cerr << "Reversal " << revNum << endl;
    --revNum;
    mapEnd = dPosIt->end();
    for (mapIt = dPosIt->begin(); mapIt != mapEnd; ++mapIt) {
      cerr << " Position " << mapIt->first << ": ";
      vectEnd = mapIt->second.end();
      for (vectIt = mapIt->second.begin(); vectIt != vectEnd; ++vectIt) {
	cerr << *vectIt << " ";
      }
      cerr << endl;
    }
  }
}

// Print out the hurdle sequences as they lie in the original genomes
void SortSeq::printOriginalHurdles() {
  int left, right, pos, i;
  const int renG2Size = renamedG2.size();
  const int consG2Size = consensusG2.size();
  bool isSuper;
  int hurdleNum;
  vector<HurdleInfo>::iterator hurdleIt, hurdleEnd = hurdles.end();
  int min1, max1, innerMin1, innerMax1;	// hurdle endpoints in genome 1
  int min2, max2, innerMin2, innerMax2;	// hurdle endpoints in genome 2
  hash_map<int,int> inverseRenamedG2;
  vector<int> hurdleOnlyGenome;	// consensusG1 immediately before the first
				// hurdle-clearing reversal

  if (hurdles.empty())
    return;

  if (revConsensusG1.empty())
    computeConsensusPositions();

  // Generate an inverse map for renamedG2
  inverseRenamedG2.resize(2*renG2Size);
  for (i = 1; i <= renG2Size; ++i) {
    inverseRenamedG2[renamedG2[i-1]] = i;
    inverseRenamedG2[-renamedG2[i-1]] = -i;
  }

  // Get the state consensusG1 was in immediately before the 1st hurdle-
  // clearing reversal
  hurdleOnlyGenome = revConsensusG1[firstHurdleRev];

  // Print each hurdle
  hurdleNum = 0;
  for (hurdleIt = hurdles.begin(); hurdleIt != hurdleEnd; ++hurdleIt) {
    isSuper = hurdleIt->isSuperHurdle;
    cerr << (isSuper ? "SUPER" : "") << "HURDLE " << hurdleNum << endl;

    // there is one genome segment in a non-maximal hurdles
    if (maxHurdle != hurdleNum) {
      left = hurdleIt->left-1;
      right = hurdleIt->right-1;
      cerr << "left=" << left << " right=" << right << endl;

      if (left < 0) {
	min1 = min2 = 0;
	cerr << " (this hurdle starts at the imaginary gene '0'. If the"
	     << " genomes are circular," << endl
	     << "  this may not be a real hurdle.)" << endl;
      }
      else {
	min1 = abs(consensusToRenamed[consensusG2[hurdleOnlyGenome[left]-1]])
	       -1;
	min2 = abs(inverseRenamedG2[min1+1])-1;
      }
      if (right >= consG2Size) {
	max1 = genome1.size()-1;
	max2 = genome2.size()-1;
	cerr << " (This hurdle ends at the imaginary genes '" << genome1.size()
	     << "' and '" << genome2.size() << "' in genomes 1 and 2" << endl
	     << "   respectively. If the genomes are circular, this may not be"
	     << " a real hurdle.)" << endl;
      }
      else {
	max1 = abs(consensusToRenamed[consensusG2[hurdleOnlyGenome[right]-1]])
	       -1;
	max2 = abs(inverseRenamedG2[max1+1])-1;
      }

      cerr << " In genome 1: From " << min1 << " to " << max1 << endl
	   << " >> ";
      if (min1 > max1) {
	pos = min1;
	min1 = max1;
	max1 = pos;
      }
      for (pos = min1; pos <= max1; ++pos) {
	cerr << genome1[pos] << " ";
      }
      cerr << endl
	   << " In genome 2: From " << min2 << " to " << max2 << endl
	   << " >> ";
      if (min2 > max2) {
	pos = min2;
	min2 = max2;
	max2 = pos;
      }
      for (pos = min2; pos <= max2; ++pos) {
	cerr << genome2[pos] << " ";
      }
      cerr << endl;
    }
    // there are two segments in maximal hurdles (left & right halves)
    else {
      cerr << "innerLeft=" << maxHInnerLeft << " innerRight="
	   << maxHInnerRight << endl;
      min1 = min2 = 0;
      max1 = genome1.size()-1;
      max2 = genome2.size()-1;
      innerMin1 = abs(consensusToRenamed[
		        consensusG2[hurdleOnlyGenome[maxHInnerLeft-1]-1]])-1;
      innerMax1 = abs(consensusToRenamed[
			consensusG2[hurdleOnlyGenome[maxHInnerRight-1]-1]])-1;
      innerMin2 = abs(inverseRenamedG2[innerMin1+1])-1;
      innerMax2 = abs(inverseRenamedG2[innerMax1+1])-1;
      cerr << " In genome 1: From " << min1 << " to " << innerMin1
	   << " and from " << innerMax1 << " to " << max1 << endl
	   << " >> ";

      if (min1 > innerMin1) {
	pos = min1;
	min1 = innerMin1;
	innerMin1 = pos;
      }
      for (pos = min1; pos <= innerMin1; ++pos) {
	cerr << genome1[pos] << " ";
      }
      cerr << "... ";
      if (innerMax1 > max1) {
	pos = innerMax1;
	innerMax1 = max1;
	max1 = pos;
      }
      for (pos = innerMax1; pos <= max1; ++pos) {
	cerr << genome1[pos] << " ";
      }
      cerr << endl;
      cerr << " In genome 2: From " << min2 << " to " << innerMin2
	   << " and from " << innerMax2 << " to " << max2 << endl
	   << " >> ";

      if (min2 > innerMin2) {
	pos = min2;
	min2 = innerMin2;
	innerMin2 = pos;
      }
      for (pos = min2; pos <= innerMin2; ++pos) {
	cerr << genome2[pos] << " ";
      }
      cerr << "... ";
      if (innerMax2 > max2) {
	pos = innerMax2;
	innerMax2 = max2;
	max2 = pos;
      }
      for (pos = innerMax2; pos <= max2; ++pos) {
	cerr << genome2[pos] << " ";
      }
      cerr << endl;
    }
    ++hurdleNum;
  }
}

// get a midpoint genome, but don't worry about returning the source of each
// gene
void SortSeq::getMidpointGenome(vector<int> &midGenome) {
  getIntermediateGenome(revs.size()/2, midGenome);
}

// get an intermediate genome after -midRev- reversals, but don't worry about
// returning the source of each gene
void SortSeq::getIntermediateGenome(int midRev, vector<int> &midGenome) {
  vector<int> source;
  source.reserve(100);
  getIntermediateGenome(midRev, midGenome, source);
}

void SortSeq::getMidpointGenome(vector<int> &midGenome,	vector<int> &source) {
  getIntermediateGenome(revs.size()/2, midGenome, source);
}

void SortSeq::getIntermediateGenome(int midRev, vector<int> &midGenome,
				    vector<int> &source) {
  int numRevs = revs.size(), delMidRev;
#ifdef OVERLY_OPTIMISTIC_INSERTION_IDEA
  vector< vector<InsConfiguration> > insConfigs;
  int insPos, insNum, insSign;
  int numRows, rowNum;
  InsConfiguration jointInsConfig;
  InsConfiguration::iterator posIt, posEnd, insConfIt, insConfEnd;
#else // OVERLY_OPTIMISTIC_INSERTION_IDEA
  hash_map<int,vector<int> >::const_iterator stickyIt,
    stickyEnd = stickyInsMap.end();
  vector<int>::const_iterator insIt, insEnd;
  vector<int>::const_reverse_iterator rInsIt, rInsEnd;
#endif // OVERLY_OPTIMISTIC_INSERTION_IDEA
  int delNum;
  int consPos, consGene;
  vector<int>::iterator consensusIt, consensusEnd;
  vector<int>::iterator segmentIt, segmentEnd;
  dPosLevel1_iterator dPosMapIt, dPosMapEnd;

  // boundary-check midRev and reverse it (g2->g1 easier to figure than g1->g2)
  if (midRev < 0)
    midRev = numRevs;
  else if (midRev > numRevs)
    midRev = 0;
  else
    midRev = numRevs - midRev;
  delMidRev = numRevs - midRev;

  midGenome.clear();
  source.clear();

#ifdef VERBOSE
  cerr << "MEDIAN GENOME" << endl;
#endif // VERBOSE


#ifdef OVERLY_OPTIMISTIC_INSERTION_IDEA
  getInsConfigurations(midRev, insConfigs);

  // condense -insConfigs- to a single InsConfiguration
  jointInsConfig.clear();
  numRows = insConfigs.size();
  for (rowNum = 0; rowNum < numRows; ++rowNum) {
    if (insConfigs[rowNum].size() != 1)
      cerr << endl << "ERROR in SortSeq.C: getIntermediateGenome()" << endl
	   << " - Multiple insertion configurations remaining" << endl
	   << "     : in row " << rowNum << endl << endl;
    posEnd = insConfigs[rowNum][0].end();
    for (posIt = insConfigs[rowNum][0].begin(); posIt != posEnd; ++posIt) {
      insPos = posIt->first;
      if (!(jointInsConfig.insert(IConfVal_t(insPos, posIt->second))).second)
	// true if insertion failed because insPos is already in jointInsConfig
	cerr << endl << "ERROR in SortSeq.C: getIntermediateGenome()" << endl
	     << " - Multiple insertion configurations" << endl
	     << "     : at position " << insPos << endl << endl;
    }
  }

  insConfEnd = jointInsConfig.end();
#endif // OVERLY_OPTIMISTIC_INSERTION_IDEA
  dPosMapEnd = deletePos[delMidRev].end();
  consensusEnd = revConsensusG2[midRev].end();
  consPos = 1;
  // store the median genome
  for (consensusIt = revConsensusG2[midRev].begin();
       consensusIt != consensusEnd; ++consensusIt) {
#ifdef OVERLY_OPTIMISTIC_INSERTION_IDEA
    // insertion block
    insConfIt = jointInsConfig.find(consPos);
    if (insConfIt != insConfEnd) {
      segmentEnd = insConfIt->second.end();
      for (segmentIt = insConfIt->second.begin(); segmentIt != segmentEnd;
	   ++segmentIt) {
	insNum = *segmentIt;
	insSign = (insNum < 0) ? -1 : 1;
	insNum
	  = insSign*renamedToGenome2[
		      consensusToRenamed[inserts[abs(insNum)-1].getLeft()]];
#ifdef VERBOSE
	cerr << insNum << " ";
#endif // VERBOSE
	midGenome.push_back(insNum);
	source.push_back(GENOME2_ONLY);
      }
    }
#ifdef VERBOSE
    cerr << " ";
#endif // VERBOSE
#endif // OVERLY_OPTIMISTIC_INSERTION_IDEA

    // deletion block
    dPosMapIt = deletePos[delMidRev].find(consPos);
    if (dPosMapIt != dPosMapEnd) {
      segmentEnd = dPosMapIt->second.end();
      for (segmentIt = dPosMapIt->second.begin(); segmentIt != segmentEnd;
	   ++segmentIt) {
	delNum = *segmentIt;
#ifdef VERBOSE
	cerr << delNum << " ";
#endif // VERBOSE
	midGenome.push_back(delNum);
	source.push_back(GENOME1_ONLY);
      }
    }
    // consensus gene (and insertions stuck to it)
    consGene = renamedToGenome2[consensusToRenamed[*consensusIt]];
#ifndef OVERLY_OPTIMISTIC_INSERTION_IDEA
    if ((stickyIt = stickyInsMap.find(*consensusIt)) != stickyEnd)
      for (insIt = stickyIt->second.begin(), insEnd = stickyIt->second.end();
	   insIt != insEnd; ++insIt) {
#ifdef VERBOSE
	cerr << *insIt << " ";
#endif // VERBOSE
	midGenome.push_back(*insIt);
	source.push_back(GENOME2_ONLY);
      }
#endif // not OVERLY_OPTIMISTIC_INSERTION_IDEA
#ifdef VERBOSE
    cerr << " " << consGene;
#endif // VERBOSE
    midGenome.push_back(consGene);
    source.push_back(BOTH_GENOMES);
#ifndef OVERLY_OPTIMISTIC_INSERTION_IDEA
    if ((stickyIt = stickyInsMap.find(-*consensusIt)) != stickyEnd)
      for (rInsIt = stickyIt->second.rbegin(),
	     rInsEnd = stickyIt->second.rend(); rInsIt != rInsEnd; ++rInsIt) {
#ifdef VERBOSE
	cerr << " " << -*rInsIt;
#endif // VERBOSE
	midGenome.push_back(-*rInsIt);
	source.push_back(GENOME2_ONLY);
      }
#endif // not OVERLY_OPTIMISTIC_INSERTION_IDEA
#ifdef VERBOSE
    cerr << endl;
#endif // VERBOSE
    ++consPos;
  }
#ifdef OVERLY_OPTIMISTIC_INSERTION_IDEA
  // last insertion block
  insConfIt = jointInsConfig.find(consPos);
  if (insConfIt != insConfEnd) {
    segmentEnd = insConfIt->second.end();
    for (segmentIt = insConfIt->second.begin(); segmentIt != segmentEnd;
	 ++segmentIt) {
      insNum = *segmentIt;
      insSign = (insNum < 0) ? -1 : 1;
      insNum
	= insSign*renamedToGenome2[
		    consensusToRenamed[inserts[abs(insNum)-1].getLeft()]];
#ifdef VERBOSE
      cerr << insNum << " ";
#endif // VERBOSE
      midGenome.push_back(insNum);
      source.push_back(GENOME2_ONLY);
    }
  }
#ifdef VERBOSE
  cerr << " ";
#endif // VERBOSE
#endif // OVERLY_OPTIMISTIC_INSERTION_IDEA
    // last deletion block
  dPosMapIt = deletePos[delMidRev].find(consPos);
  if (dPosMapIt != dPosMapEnd) {
    segmentEnd = dPosMapIt->second.end();
    for (segmentIt = dPosMapIt->second.begin(); segmentIt != segmentEnd;
	 ++segmentIt) {
      delNum = *segmentIt;
#ifdef VERBOSE
      cerr << delNum << " ";
#endif // VERBOSE
      midGenome.push_back(delNum);
      source.push_back(GENOME1_ONLY);
    }
  }
#ifndef OVERLY_OPTIMISTIC_INSERTION_IDEA
  // last sticky insertion block
  if ((stickyIt = stickyInsMap.find(consPos)) != stickyEnd)
    for (insIt = stickyIt->second.begin(), insEnd = stickyIt->second.end();
	 insIt != insEnd; ++insIt) {
#ifdef VERBOSE
      cerr << *insIt << " ";
#endif // VERBOSE
      midGenome.push_back(*insIt);
      source.push_back(GENOME2_ONLY);
    }
#endif // not OVERLY_OPTIMISTIC_INSERTION_IDEA
#ifdef VERBOSE
  cerr << endl;
#endif // VERBOSE
}


// Return the distribution of reversal sizes, where the size of the reversal
// is its length in the consensus genome
void SortSeq::getSimpleRevDist(map<int,int> &revSizeDistrib) const {
  revSizeDistrib.clear();
  vector<Pair>::const_iterator revIt;
  for (revIt = revs.begin(); revIt != revs.end(); ++revIt) {
    int revSize = revIt->getRight() - revIt->getLeft() + 1;
    map<int,int>::iterator sizeIt = revSizeDistrib.find(revSize);
    if (sizeIt == revSizeDistrib.end())
      revSizeDistrib[revSize] = 1;
    else
      ++(sizeIt->second);
  }
}


// Return all the possible configurations that inserted genes could be in after
// -midRev- reversals
//  -insConfigs- is a matrix of InsConfigurations in which each row contains
//  mutually exclusive possible configurations. Exactly one configuration from
//  each row must be chosen for the sorting sequence.
void SortSeq::getInsConfigurations(int midRev,
	        vector< vector<InsConfiguration> > &insConfigs) {
  int subsetNum, numSubsets = insSubsets.size();

  insConfigs.clear();
  insConfigs.reserve(numSubsets);

  vector<InsConfiguration> confGroup;
  InsConfiguration newConf;

  for (subsetNum = 0; subsetNum < numSubsets; ++subsetNum) {
    confGroup.clear();
    if (!mutExConfigs[subsetNum].empty()) {
      MutExConfig_t::iterator setIt, setEnd;
      for (setIt = mutExConfigs[subsetNum][midRev].begin(),
	     setEnd = mutExConfigs[subsetNum][midRev].end(); setIt != setEnd;
	   ++setIt) {
	newConf.clear();
	PositionSet_t::iterator posIt, posEnd;
	for (posIt = setIt->begin(), posEnd = setIt->end(); posIt != posEnd;
	     ++posIt)
	  newConf[*posIt] = combinedConfigs[subsetNum][midRev][*posIt];

	confGroup.push_back(newConf);
      }
    }
    else {
      if (mutExCycConfigs[subsetNum].empty())
	cerr << endl << "ERROR in SortSeq.C: getInsConfigurations()" << endl
	     << " - Malformed mutExConfigs/mutExCycConfigs data structures"
	     << endl
	     << "   neither contains subset " << subsetNum << endl
	     << "   an improper mem access is imminent" << endl << endl;
      vector<MutExConfig_t>::iterator dupIt, dupEnd;
      int dupNum = 0;
      for (dupIt = mutExCycConfigs[subsetNum][midRev].begin(),
	     dupEnd = mutExCycConfigs[subsetNum][midRev].end();
	   dupIt != dupEnd; ++dupIt, ++dupNum) {
	MutExConfig_t::iterator setIt, setEnd;
	for (setIt = dupIt->begin(), setEnd = dupIt->end(); setIt != setEnd;
	     ++setIt) {
	  newConf.clear();
	  PositionSet_t::iterator posIt, posEnd;
	  for (posIt = setIt->begin(), posEnd = setIt->end(); posIt != posEnd;
	       ++posIt)
	    newConf[*posIt]
	      = combinedCycConfigs[subsetNum][midRev][dupNum][*posIt];

	  confGroup.push_back(newConf);
	}
      }
    }
    newConf.clear();
    chooseConfigRandomly(newConf, confGroup);
    insConfigs.push_back(vector<InsConfiguration>(1, newConf));
  }
}

// choose one InsConfiguration randomly from -confGroup- and store in -newConf-
void SortSeq::chooseConfigRandomly(InsConfiguration &newConf,
				   const vector<InsConfiguration> &confGroup) {
  newConf = confGroup[rand() % confGroup.size()];
}

void SortSeq::computeInsConfigurations() {
  bool noCycle;
  set<int> currSubset, insGroup;
  set<int>::iterator setIt, setEnd, oldIt;
  vector< vector<int> >::iterator subsetsIt, subsetsEnd = insSubsets.end();
  InsConfiguration newConfig;

  int revNum, numRevs=revs.size();
  int numIns = inserts.size();

  // create num-of-insertions+1 by num-of-reversals+1 maps (first one empty)
  insertPos.clear();
  insertPos.resize(numIns+1, IPosLevel1_t(numRevs+1));

  combinedConfigs.clear();
  combinedConfigs.resize(insSubsets.size());
  mutExConfigs.clear();
  mutExConfigs.resize(insSubsets.size());
  combinedCycConfigs.clear();
  combinedCycConfigs.resize(insSubsets.size());
  mutExCycConfigs.clear();
  mutExCycConfigs.resize(insSubsets.size());

  if (numRevs==0) {
    cerr << "Using a zero-reversal genome" << endl;
    return;
  }
  if (numIns==0)
    return;

  int subsetNum = 0;
  for (subsetsIt = insSubsets.begin(); subsetsIt != subsetsEnd; ++subsetsIt) {
    currSubset.clear();
    currSubset.insert(subsetsIt->begin(), subsetsIt->end());

    // check if any of the remaining genes in subset have a cycle
    for (setIt = currSubset.begin(), setEnd = currSubset.end(), noCycle = true;
	 setIt != setEnd; ++setIt) {
      computeInsPositions(*setIt);
      noCycle = noCycle && noIPosCycle(*setIt);
    }

    // if there are no cycles, inserts will always clump together
    if (noCycle) {
      // make an InsConfiguration of the initial positions
      newConfig.clear();
      for (setIt = currSubset.begin(), setEnd = currSubset.end();
	   setIt != setEnd; ++setIt) {
	int insNum = *setIt;
	int pos = abs(insertPos[insNum][0].begin()->first);
	newConfig[pos] = vector<int>(1, insNum);
      }

      CombConfLevel1_t& subsetConfigs = combinedConfigs[subsetNum];
      subsetConfigs.push_back(newConfig);
      // make a series of InsConfigurations which are the combination of all
      // the different mutually exclusive InsConfigurations at each reversal
      // for the current insertion subset
      InsConfiguration::const_iterator mapIt, mapEnd;
      for (revNum = 0; revNum < numRevs; ++revNum) {
	newConfig.clear();
	for (mapIt = subsetConfigs[revNum].begin(), // iterate across
	       mapEnd = subsetConfigs[revNum].end();// insertion blocks of
	     mapIt != mapEnd; ++mapIt) {	      // the previous newConfig
	  if (mapIt->second.empty())
	    cerr << endl
		 << "ERROR in SortSeq.C: getInsConfigurations()" << endl
		 << " - Malformed 'combinedConfigs' data structure. An"
		 << " improper mem access is" << endl
		 << "   imminent" << endl << endl;
	  copyInsertionBlock(newConfig, revNum, mapIt);
	}
	subsetConfigs.push_back(newConfig);
      }

      // all of the final insertion positions are mutually exclusive, so put
      // each of the final insertion positions into a different set in
      // -mutExConfigs[numRevs]-
      PositionSet_t newConf;
      vector<MutExConfig_t>& subsetMutExConfigs = mutExConfigs[subsetNum];
      subsetMutExConfigs.resize(numRevs+1);
      for (mapIt = subsetConfigs[numRevs].begin(),
	     mapEnd = subsetConfigs[numRevs].end();
	   mapIt != mapEnd; ++mapIt) {
	int pos = mapIt->first;
	newConf.clear();
	newConf.insert(pos);
	subsetMutExConfigs[numRevs].insert(newConf);
      }

      // compute the mutually exclusive sets of insertion positions after each
      // reversal
      for (revNum = numRevs-1; revNum >= 0; --revNum) {
// 	cerr << "revNum=" << revNum;
	copyBackMutExConfigs(revs[numRevs-revNum-1], subsetConfigs[revNum],
			     subsetMutExConfigs[revNum+1],
			     subsetMutExConfigs[revNum]);
      }
    }
    else {	// if there are cycles in insertPos for the current subset
      IPosCycleInfo_t iPosCycleInfo;	// maps genes with cyclic graphs in
					// -insertPos- to all of their
					// division points

      // identify cycle genes and division points
      getIPosCycleInfo(iPosCycleInfo, currSubset);
#ifdef VERBOSE
      printIPosCycleInfo(iPosCycleInfo);
#endif // VERBOSE

      // build a combined config as we did for the acyclic case, but use an
      // array of InsConfigurations at each reversal and at reversals with
      // division points redouble the size of the array for each divided gene
//       CombCycConfigs_t combinedCycConfigs;
      getCombinedCycConfigs(subsetNum, iPosCycleInfo, currSubset);
    }

#ifndef VERBOSE
    // clear the subset from -insertPos- (unless you plan to print it later)
    for (setIt = currSubset.begin(), setEnd = currSubset.end();
	 setIt != setEnd; ++setIt)
      insertPos[*setIt].clear();
#endif // !VERBOSE

    ++subsetNum;
  }
#ifdef VERBOSE
    cerr << "Full version of insertPos:" << endl;
    printInsertPos();
#endif //VERBOSE

  insertPos.clear();

#ifdef VERBOSE
  printConfigs();
#endif // VERBOSE
}

// Ensure that the insertion -insVal- cannot reach an intermediate position
// in -insertPos- at reversal i from two different positions at reversal i-1
// (I believe cycles can only be created by unoriented reversals, i.e.,hurdles)
bool SortSeq::noIPosCycle(int insVal) {
  set<int> newVals;
  IPosLevel1_iterator revIt, revEnd;
  IPosLevel2_iterator mapIt, mapEnd;
  InsPosNode *node;

  // check for an -insertPos- map entry with two different positions before a
  // reversal leading to the same position after the reversal
  revEnd = insertPos[insVal].end()-1;
  for (revIt = insertPos[insVal].begin(); revIt != revEnd; ++revIt) {
    mapEnd = revIt->end();
    newVals.clear();
    for (mapIt = revIt->begin(); mapIt != mapEnd; ++mapIt) {
      node = &(mapIt->second);
      if (!((newVals.insert(abs(node->newPos1))).second)
	  // true if abs(newPos1) was already inserted into newVals
	  || node->twoNew
	  && !((newVals.insert(abs(node->newPos2))).second))
	  // true if abs(newPos2) was already inserted into newVals
	return false;
    }
  }
  return true;
}

// Return all the division-rejoin point pairs for the genes in -insSubset-
void SortSeq::getIPosCycleInfo(IPosCycleInfo_t &iPosCycleInfo,
			       const set<int> &insSubset) {
  set<int>::const_iterator pairSetIt, pairSetEnd;
  set<InsertPosPair_t> rejoinPoints;
  int maxPos = consensusG2.size()+2; // number higher than highest position

  iPosCycleInfo.clear();
  for (pairSetIt = insSubset.begin(), pairSetEnd = insSubset.end();
       pairSetIt != pairSetEnd; ++pairSetIt) {
    int insNum = *pairSetIt;

    // FIRST, get all the points in -insertPos- where two paths of the
    // insertion -insNum- rejoin (rejoin points)
    rejoinPoints.clear();
    set< pair<int,int> > newOldPairs;
    IPosLevel1_iterator revIt, revEnd;
    IPosLevel2_iterator mapIt, mapEnd;
    InsPosNode *node;

//     cerr << "rejoin points for insertion " << insNum << " (rev pos): ";
    int revNum = 0;    
    for (revIt = insertPos[insNum].begin(), revEnd = insertPos[insNum].end()-1;
	 revIt != revEnd; ++revIt) {
      // create a set with the positions before and after the reversal
      newOldPairs.clear();
      for (mapIt = revIt->begin(), mapEnd = revIt->end(); mapIt != mapEnd;
	   ++mapIt) {
	node = &(mapIt->second);
	int oldPos = abs(mapIt->first);
	int newPos = abs(node->newPos1);
	newOldPairs.insert(make_pair(newPos, oldPos));
	if (node->twoNew) {
	  oldPos = abs(mapIt->first);
	  newPos = abs(node->newPos2);
	  newOldPairs.insert(make_pair(newPos, oldPos));
	}
      }
      //  rejoin points are pairs with the same new position and different old
      //  positions
      set< pair<int,int> >::iterator pairSetIt = newOldPairs.begin();
      set< pair<int,int> >::iterator pairSetEnd = newOldPairs.end();
      while (pairSetIt != pairSetEnd) {
	int newPos = pairSetIt->first;
	++pairSetIt;
	if (pairSetIt != pairSetEnd && pairSetIt->first == newPos) {
// 	  cerr << "(" << revNum << " " << newPos << ") ";
	  rejoinPoints.insert(make_pair(revNum, newPos));
	  ++pairSetIt;
	}
      }
      ++revNum;
    }
//     cerr << endl;

    // SECOND, find division points
    set<InsertPosPair_t>::iterator rejoinIt, rejoinEnd;
    for (rejoinIt = rejoinPoints.begin(), rejoinEnd = rejoinPoints.end();
	 rejoinIt != rejoinEnd; ++rejoinIt) {
      // find the point in -insertPos- where the two paths leading into the
      // current rejoin point originally divided (division point)
      int leftPathPos, rightPathPos, newLeft, newRight;
      leftPathPos = rightPathPos = abs(rejoinIt->second);
      bool divPtFound = false;
      for (int revNum = rejoinIt->first; !divPtFound && revNum >= 0;
	   --revNum) {
// 	cerr << "rev " << revNum << ": ";
	newLeft = maxPos;
	newRight = -1;
	for (mapIt = insertPos[insNum][revNum].begin(),
	       mapEnd = insertPos[insNum][revNum].end();
	     !divPtFound && mapIt != mapEnd; ++mapIt) {
	  InsPosNode *node = &(mapIt->second);
	  int sourcePos = abs(mapIt->first);
	  int newPos = abs(node->newPos1);
// 	  cerr << sourcePos << "=" << newPos;
	  if (newPos == leftPathPos && sourcePos < newLeft)
	    newLeft = sourcePos;
	  if (newPos == rightPathPos && sourcePos > newRight)
	    newRight = sourcePos;
	  if (node->twoNew) {
	    newPos = abs(node->newPos2);
// 	    cerr << "," << newPos;
	    if (newPos == leftPathPos && sourcePos < newLeft)
	      newLeft = sourcePos;
	    if (newPos == rightPathPos && sourcePos > newRight)
	      newRight = sourcePos;
	  }
// 	  cerr << " ";
	}
// 	cerr << "L=" << newLeft << " R=" << newRight << endl;
	if (newLeft == newRight) {	// found the division point
	  iPosCycleInfo[revNum].insert	// so store it in -iPosCycleInfo-
	    /* (make_pair */(make_pair(insNum, newLeft/* ), *rejoinIt */));
	  divPtFound = true;
	}
	else {
	  if (newLeft == maxPos || newRight == -1)
	    cerr << endl << "ERROR in SortSeq.C: getIPosCycleInfo()" << endl
		 << " - 'insertPos' data structure is malformed." << endl
		 << "   The " << (newLeft==maxPos ? "left" : "right")
		 << " path back from the division point"
		 << " (insert=" << rejoinIt->first
		 << " position=" << rejoinIt->second << ")" << endl
		 << "   abruptly died at reversal " << revNum << endl
		 << "   (newLeft=" << newLeft
		 << " newRight=" << newRight
		 << " maxPos=" << maxPos << ")" << endl << endl;
	  leftPathPos = newLeft;
	  rightPathPos = newRight;
	}
      }
      if (!divPtFound)
	cerr << endl << "ERROR in SortSeq.C: getIPosCycleInfo()" << endl
	     << " - incomplete scan of 'insertPos' data structure" << endl
	     << "   I found the rejoin point of a cycle but can't find the"
	     << " division point to" << endl
	     << "   match." << endl << endl;
    }
  }
}

void SortSeq::printIPosCycleInfo(const IPosCycleInfo_t &iPosCycleInfo) {
  cerr << "IPosCycleInfo:" << endl;
  IPosCycleInfo_t::const_iterator ipcIt, ipcEnd;
  for (ipcIt = iPosCycleInfo.begin(), ipcEnd = iPosCycleInfo.end();
       ipcIt != ipcEnd; ++ipcIt) {
    cerr << " Rev " << ipcIt->first << endl;
    set<InsertPosPair_t>::const_iterator djIt, djEnd;
    for (djIt = ipcIt->second.begin(), djEnd = ipcIt->second.end();
	 djIt != djEnd; ++djIt) {
      cerr << "  Gene " << djIt->first
	   << " Pos " << djIt->second << endl;
    }
  }
}

void SortSeq::printConfigs() {
  cerr << "COMBINEDCONFIGS" << endl;
  CombConfigs_t::const_iterator ccBegin, ccIt, ccEnd;
  ccBegin = combinedConfigs.begin();
  for (ccIt = combinedConfigs.begin(), ccEnd = combinedConfigs.end();
       ccIt != ccEnd; ++ccIt) {
    cerr << " subset " << ccIt-ccBegin << ":" << endl;
    CombConfLevel1_t::const_iterator ccL1Begin, ccL1It, ccL1End;
    ccL1Begin = ccIt->begin();
    for (ccL1It = ccIt->begin(), ccL1End = ccIt->end(); ccL1It != ccL1End;
	 ++ccL1It) {
      cerr << "  rev " << ccL1It-ccL1Begin << ": ";
      InsConfiguration::const_iterator icIt, icEnd;
      for (icIt = ccL1It->begin(), icEnd = ccL1It->end(); icIt != icEnd;
	   ++icIt){
	cerr << icIt->first << "=";
	vector<int>::const_iterator vIt, vEnd;
	for (vIt = icIt->second.begin(), vEnd = icIt->second.end();
	     vIt != vEnd; ++vIt)
	  cerr << *vIt << " ";
	cerr << ", ";
    }
      cerr << endl;
    }
  }

  cerr << "MUTEXCONFIGS" << endl;
  vector<vector<MutExConfig_t> >::const_iterator mexBegin, mexIt, mexEnd;
  mexBegin = mutExConfigs.begin();
  for (mexIt = mutExConfigs.begin(), mexEnd = mutExConfigs.end();
       mexIt != mexEnd; ++mexIt) {
    cerr << " subset " << mexIt-mexBegin << ":" << endl;
    vector<MutExConfig_t>::const_iterator revBegin, revIt, revEnd;
    revBegin = mexIt->begin();
    for (revIt = mexIt->begin(), revEnd = mexIt->end(); revIt != revEnd;
	 ++revIt) {
      cerr << "  rev " << revIt-revBegin << ": ";
      MutExConfig_t::const_iterator setIt, setEnd;
      for (setIt = revIt->begin(), setEnd = revIt->end(); setIt != setEnd;
	   ++setIt) {
	PositionSet_t::const_iterator posIt, posEnd;
	for (posIt = setIt->begin(), posEnd = setIt->end(); posIt != posEnd;
	     ++posIt)
	  cerr << *posIt << " ";
	cerr << ", ";
      }
      cerr << endl;
    }
  }

  cerr << "COMBINEDCYCCONFIGS" << endl;
  CombCycConfigs_t::const_iterator cccBegin, cccIt, cccEnd;
  cccBegin = combinedCycConfigs.begin();
  for (cccIt = combinedCycConfigs.begin(), cccEnd = combinedCycConfigs.end();
       cccIt != cccEnd; ++cccIt) {
    cerr << " subset " << cccIt-cccBegin << ":" << endl;
    CycConfLevel1_t::const_iterator cccL1Begin, cccL1It, cccL1End;
    cccL1Begin = cccIt->begin();
    for (cccL1It = cccIt->begin(), cccL1End = cccIt->end();
	 cccL1It != cccL1End; ++cccL1It) {
      cerr << "  rev " << cccL1It-cccL1Begin << ": " << endl;
      CycConfLevel2_t::const_iterator cccL2Begin, cccL2It, cccL2End;
      cccL2Begin = cccL1It->begin();
      for (cccL2It = cccL1It->begin(), cccL2End = cccL1It->end();
	   cccL2It != cccL2End; ++cccL2It) {
	cerr << "   duplicate " << cccL2It-cccL2Begin << ": ";
	InsConfiguration::const_iterator icIt, icEnd;
	for (icIt = cccL2It->begin(), icEnd = cccL2It->end(); icIt != icEnd;
	     ++icIt){
	  cerr << icIt->first << "=";
	  vector<int>::const_iterator vIt, vEnd;
	  for (vIt = icIt->second.begin(), vEnd = icIt->second.end();
	       vIt != vEnd; ++vIt)
	    cerr << *vIt << " ";
	  cerr << ", ";
	}
	cerr << endl;
      }
    }
  }

  cerr << "MUTEXCYCCONFIGS" << endl;
  vector<vector<vector<MutExConfig_t> > >::const_iterator mexCBegin, mexCIt,
							  mexCEnd;
  mexCBegin = mutExCycConfigs.begin();
  for (mexCIt = mutExCycConfigs.begin(), mexCEnd = mutExCycConfigs.end();
       mexCIt != mexCEnd; ++mexCIt) {
    cerr << " subset " << mexCIt-mexCBegin << ":" << endl;
    vector<vector<MutExConfig_t> >::const_iterator revBegin, revIt, revEnd;
    revBegin = mexCIt->begin();
    for (revIt = mexCIt->begin(), revEnd = mexCIt->end();
	 revIt != revEnd; ++revIt) {
      cerr << "  rev " << revIt-revBegin << ":" << endl;
      vector<MutExConfig_t>::const_iterator dupBegin, dupIt, dupEnd;
      dupBegin = revIt->begin();
      for (dupIt = revIt->begin(), dupEnd = revIt->end(); dupIt != dupEnd;
	   ++dupIt) {
	cerr << "   duplicate " << dupIt-dupBegin << ": ";
	MutExConfig_t::const_iterator setIt, setEnd;
	for (setIt = dupIt->begin(), setEnd = dupIt->end(); setIt != setEnd;
	     ++setIt) {
	  PositionSet_t::const_iterator posIt, posEnd;
	  for (posIt = setIt->begin(), posEnd = setIt->end(); posIt != posEnd;
	       ++posIt)
	    cerr << *posIt << " ";
	  cerr << ", ";
	}
	cerr << endl;
      }
    }
  }
}

// ***** Helper function for -computeInsConfigurations()- *****
// Given the reversal -rev-, the InsConfiguration -insConfig- that the reversal
// acts on, and the set of mutually exclusive configurations -lastMutExConf-
// after the reversal, fill -newMutExConf- with the mutually exclusive
// configurations before the reversal
void SortSeq::copyBackMutExConfigs(const Pair &rev,
				   const InsConfiguration &insConfig,
				   const MutExConfig_t &lastMutExConf,
				   MutExConfig_t &newMutExConf) {
  InsConfiguration::const_iterator iCEnd = insConfig.end();
  int left = rev.getLeft()+1;
  int right = rev.getRight()+2;
//   cerr << " left=" << left << " right=" << right << endl;
  PositionSet_t newConf;
  MutExConfig_t::const_iterator configIt, configEnd;
  PositionSet_t::const_iterator posIt, posEnd;
  for (configIt = lastMutExConf.begin(),
	 configEnd = lastMutExConf.end();
       configIt != configEnd; ++configIt) {
    newConf.clear();
    for (posIt = configIt->begin(), posEnd = configIt->end(); posIt != posEnd;
	 ++posIt) {
      int pos = *posIt;
//       cerr << "pos=" << pos << endl;
      // pos outside reversal, so add position to new config
      if (pos < left || pos > right) {
	newConf.insert(pos);
      }
      // pos inside reversal, so add reflected position to new config number
      else if (pos > left && pos < right) {
	int newPos = left+right-pos;
	newConf.insert(newPos);
      }
      // position on edge of reversal
      else {
	// if pos could have come from the left, add -left- to new config
	if (insConfig.find(left) != iCEnd) {
	  newConf.insert(left);
	}
	// if pos could have come from the right, add -right- to new
	// config
	if (insConfig.find(right) != iCEnd) {
	  newConf.insert(right);
	}
      }
    }
    newMutExConf.insert(newConf);
//     cerr << " newConf: ";
//     for (posIt = newConf.begin(), posEnd = newConf.end(); posIt != posEnd;
// 	 ++posIt) {
//       cerr << *posIt << " ";
//     }
//     cerr << endl;
  }
}

// ***** Helper function for -computeInsConfigurations()- *****
void SortSeq::getCombinedCycConfigs(int subsetNum,
				    const IPosCycleInfo_t &iPosCycleInfo,
				    const set<int> &currSubset) {

  set<int>::const_iterator setIt, setEnd;
  IPosCycleInfo_t::const_iterator ipcInfoEnd = iPosCycleInfo.end();

  int revNum, numRevs = revs.size();
  CycConfLevel1_t& subsetCycConfigs = combinedCycConfigs[subsetNum];

  // make the first InsConfiguration with the insertions' initial positions
  subsetCycConfigs.resize(numRevs+1);
  subsetCycConfigs[0].resize(1);
  InsConfiguration *firstConfig = &(subsetCycConfigs[0][0]);
  for (setIt = currSubset.begin(), setEnd = currSubset.end();
       setIt != setEnd; ++setIt) {
    int insNum = *setIt;
    int pos = abs(insertPos[insNum][0].begin()->first);
    (*firstConfig)[pos].push_back(insNum);
  }
  firstConfig = NULL;

  // make a series of arrays of InsConfigurations which are the combination of
  //  all the different mutually exclusive InsConfigurations at each reversal
  //  for the current insertion subset
  // furthermore, at reversals with division points, increase the number of
  //  "duplicate" arrays
  for (revNum = 0; revNum < numRevs; ++revNum) {
    int growthFactor = 1;
    IPosCycleInfo_t::const_iterator divPoints = iPosCycleInfo.find(revNum);
    IPCycLevel1_t::const_iterator divPtIt, divPtEnd;
    map<int,int> numGenesAtPos;
    map<int,int>::iterator atPosIt, atPosEnd;

    // if this reversal contains division points, iterate across them and count
    // how many genes are at each position
    bool isDivPointRev = (divPoints != ipcInfoEnd);
    if (isDivPointRev) {
      divPtEnd = divPoints->second.end();
      for (divPtIt = divPoints->second.begin();
	   divPtIt != divPtEnd; ++divPtIt) {
	int divPtPos = divPtIt->second;
	atPosIt = numGenesAtPos.find(divPtPos);
	if (atPosIt != numGenesAtPos.end())
	  ++(atPosIt->second);		// increment count of existing position
	else
	  numGenesAtPos[divPtPos] = 1;	// add new position
      }
				// *****ERROR CHECK*****
      if (numGenesAtPos.empty())
	cerr << endl << "ERROR in SortSeq.C: getCombinedCycConfigs()" << endl
	     << " - malformed 'numGenesAtPos' data structure" << endl
	     << "   division points exist, but the structure is empty"
	     << endl << endl;
      else if (numGenesAtPos.size() > 2)
	cerr << endl << "ERROR in SortSeq.C: getCombinedCycConfigs()" << endl
	     << " - malformed 'numGenesAtPos' data structure" << endl
	     << "   structure contains more than two 'reversal endpoints'"
	     << endl << endl;
				// *****END CHECK*****

      for (atPosIt = numGenesAtPos.begin(), atPosEnd = numGenesAtPos.end();
	   atPosIt != atPosEnd; ++atPosIt) {
	// given X inserted genes on one side of the div point reversal,
	// anything from 0 to X of those genes can move to the other side of
	// the reversal, so the number of maps in -subsetCycConfigs- should
	// increase by a factor of X+1
	growthFactor *= (atPosIt->second+1);

				// *****ERROR CHECK*****
// 	cerr << "increase growth factor by factor of " << atPosIt->second+1
// 	     << endl;
	InsConfiguration::iterator insConfigIt, insConfigEnd;
	if ((insConfigIt = subsetCycConfigs[revNum][0].find(atPosIt->first))
	    == subsetCycConfigs[revNum][0].end())
	  cerr << endl << "ERROR in SortSeq.C: getCombinedCycConfigs()" << endl
	       << " - malformed 'combinedCycConfigs' data structure" << endl
	       << endl;
	else if (int(insConfigIt->second.size()) != atPosIt->second) {
	  cerr << endl << "ERROR in SortSeq.C: getCombinedCycConfigs()" << endl
	       << " - number of genes in 'numGenesAtPos[" << atPosIt->first
	       << "]' and" << endl
	       << "   'combinedCycConfigs[" << subsetNum << "]["
	       << revNum << "][0][" << atPosIt->first << "]' don't match ("
	       << atPosIt->second << "!=" << insConfigIt->second.size() << ")"
	       << endl;
	  cerr << "combinedCycConfigs[" << subsetNum << "]["
	       << revNum << "][0][" << atPosIt->first << "]: ";
	  vector<int>::iterator vIt, vEnd;
	  for (vIt = insConfigIt->second.begin(),
		 vEnd = insConfigIt->second.end(); vIt != vEnd; ++vIt)
	    cerr << *vIt << " ";
	  cerr << endl << endl;
	}
				// *****END CHECK*****
      }
//       cerr << "subset=";
//       for (setIt = currSubset.begin(), setEnd = currSubset.end();
// 	   setIt != setEnd; ++setIt)
// 	cerr << *setIt << " ";
//       cerr << "rev=" << revNum << " and growth factor=" << growthFactor
// 	   << endl;
    }

    int numDupls = subsetCycConfigs[revNum].size();
    subsetCycConfigs[revNum+1].resize(numDupls*growthFactor);
    for (int duplNum = 0; duplNum < numDupls; ++duplNum) {
      InsConfiguration::const_iterator mapIt, mapEnd;
      for (mapIt = subsetCycConfigs[revNum][duplNum].begin(),
	     mapEnd = subsetCycConfigs[revNum][duplNum].end();
	   mapIt != mapEnd; ++mapIt) { // iterate across previous newConfig
	if (mapIt->second.empty())
	  cerr << endl
	       << "ERROR in SortSeq.C: getCombinedCycConfigs()" << endl
	       << " - malformed 'combinedCycConfigs' data structure. An"
	       << " improper mem access is" << endl
	       << "   imminent" << endl << endl;
	// if this reversal has division points
	if (isDivPointRev) {
	  int currPos = mapIt->first;

	  // if the current position is in divPoints
	  if (numGenesAtPos.find(currPos) != atPosEnd) {
	    atPosIt = numGenesAtPos.begin();

	    // if there is only one div point position
	    if (numGenesAtPos.size() == 1) {
	      splitInsBlock_1D(subsetCycConfigs[revNum+1], duplNum, revNum,
			       mapIt);
	    }
	    // if this is the first of two div point positions
	    else if (atPosIt->first == currPos) {
	      splitInsBlock_2DFirst((++atPosIt)->second + 1,
				    subsetCycConfigs[revNum+1], duplNum,
				    revNum, mapIt);
	    }
	    // if this is the second div point position
	    else if ((++atPosIt)->first == currPos) {
	      splitInsBlock_2DSecond((--atPosIt)->second + 1,
				     subsetCycConfigs[revNum+1], duplNum,
				     revNum, mapIt);
	    }
	    else		// should never hit this case
	      cerr << endl
		   << "ERROR in SortSeq.C: getCombinedCycConfigs()" << endl
		   << " - malformed 'numGenesAtPos' data structure" << endl
		   << "   division point " << currPos << " at reversal "
		   << revNum << " found, then not found" << endl << endl;
	  }
	  else {		// current insertion block isn't in divPoints
	    spreadInsBlock(subsetCycConfigs[revNum+1], duplNum, revNum,
			   mapIt, growthFactor);
	  }
	}
	else {			// current reversal has no division points
	  copyInsertionBlock(subsetCycConfigs[revNum+1][duplNum], revNum,
			     mapIt);
	}
      }
    }
  }

  // erase -subsetCycConfigs- entries at the last reversal which don't
  // contain all the genes in the insertion subset
  eraseBadFinalEntries(currSubset.size(), subsetCycConfigs);

  // eliminate -subsetCycConfigs- subpaths that terminate in an erased
  // InsConfiguration entry
  eraseDeadSubpaths(subsetCycConfigs);

  // *****OUTPUT INTERMEDIATE RESULT*****
//   cerr << "subsetCycConfigs with cleared subpaths" << endl;
//   cccBegin = subsetCycConfigs.begin();
//   for (cccIt = cccBegin, cccEnd = subsetCycConfigs.end(); cccIt != cccEnd;
//        ++cccIt) {
//     cerr << " rev " << cccIt-cccBegin << ":" << endl;
//     CycConfLevel2_t::const_iterator duplBegin, duplIt, duplEnd;
//     duplBegin = cccIt->begin();
//     for (duplIt = duplBegin, duplEnd = cccIt->end(); duplIt != duplEnd;
// 	 ++duplIt) {
//       cerr << "  dupl " << duplIt-duplBegin << ": ";
//       InsConfiguration::const_iterator icIt, icEnd;
//       for (icIt = duplIt->begin(), icEnd = duplIt->end(); icIt != icEnd;
// 	   ++icIt){
// 	cerr << icIt->first << "=";
// 	vector<int>::const_iterator vIt, vEnd;
// 	for (vIt = icIt->second.begin(), vEnd = icIt->second.end();
// 	     vIt != vEnd; ++vIt)
// 	  cerr << *vIt << " ";
// 	cerr << ", ";
//       }
//       cerr << endl;
//     }
//   }
  // *****END OUTPUT INTERMEDIATE RESULT*****

  // all of the final insertion positions are mutually exclusive, so put
  // each of the final insertion positions into a different set in
  // -mutExCycConfigs[subsetNum][numRevs]-
  PositionSet_t newConf;
  vector< vector<MutExConfig_t> >& subsetMutExCycConfigs
    = mutExCycConfigs[subsetNum];
  subsetMutExCycConfigs.resize(numRevs+1);
  int numDupls = subsetCycConfigs[numRevs].size();
  subsetMutExCycConfigs[numRevs].resize(numDupls);
  for (int duplNum = 0; duplNum < numDupls; ++duplNum) {
    InsConfiguration::iterator mapIt, mapEnd;
    for (mapIt = subsetCycConfigs[numRevs][duplNum].begin(),
	   mapEnd = subsetCycConfigs[numRevs][duplNum].end();
	 mapIt != mapEnd; ++mapIt) {
      int pos = mapIt->first;
      newConf.clear();
      newConf.insert(pos);
      subsetMutExCycConfigs[numRevs][duplNum].insert(newConf);
    }
  }

  // compute the mutually exclusive sets of insertion positions after each
  // reversal
  for (revNum = numRevs-1; revNum >= 0; --revNum) {
//     cerr << "revNum=" << revNum << " ";
    int oldNumDups = numDupls;
    numDupls = subsetCycConfigs[revNum].size();
    subsetMutExCycConfigs[revNum].resize(numDupls);
    for (int duplNum = 0; duplNum < numDupls; ++duplNum) {
//       cerr << "duplNum=" << duplNum;
      if (oldNumDups == numDupls) {
	copyBackMutExConfigs(revs[numRevs-revNum-1],
			     subsetCycConfigs[revNum][duplNum],
			     subsetMutExCycConfigs[revNum+1][duplNum],
			     subsetMutExCycConfigs[revNum][duplNum]);
      }
      else {
	int growthFactor = oldNumDups / numDupls;
	int maxNext = (duplNum+1) * growthFactor;
	for (int nextRevDup = duplNum * growthFactor; nextRevDup < maxNext;
	     ++nextRevDup) {
// 	  cerr << " nextRevDup=" << nextRevDup;
	  copyBackMutExConfigs(revs[numRevs-revNum-1],
			       subsetCycConfigs[revNum][duplNum],
			       subsetMutExCycConfigs[revNum+1][nextRevDup],
			       subsetMutExCycConfigs[revNum][duplNum]);
	}
      }
    }
  }

  // JOEL: figure out if you need to write a function to erase empty or
  // duplicate InsConfigurations!
  // In fact, it is possible to get duplicate InsConfigurations. It would be
  // nice to eliminate duplicates, but I'm not going to do so at this point.
  // At worst, that means that the choice among the mutually exclusive
  // configurations is not quite uniformly at random. It should be nearly
  // uniform, though.
  {
  }
}

// ***** Helper function for -computeInsConfigurations()- *****
// Divide the insertion block in -mapIt- into all possible halves, sending the
// left half to the left and the right half to the right in -duplicatedConfigs-
void SortSeq::splitInsBlock_1D(CycConfLevel2_t &duplicatedConfigs,
			       int currDuplNum, int revNum,
			       InsConfiguration::const_iterator &mapIt) {
  int signedOldPos, leftPos, rightPos;
  InsPosNode *insNode;
  getInsConfigParams(mapIt, revNum, signedOldPos, insNode);

  vector<int> leftHalf, rightHalf;
  getLeftRightPos(insNode, mapIt, signedOldPos, leftPos, rightPos, leftHalf);

  int numDuplicates = leftHalf.size() + 1;
  int lowDupl = currDuplNum * numDuplicates;
  int highDupl = lowDupl + numDuplicates - 1;

  // copy the insertion block into the first duplicate
  copyInsertionBlock(duplicatedConfigs[lowDupl], leftHalf, leftPos, leftPos);

  // do the rest of the block copying except for the last duplicate
  for (int newDupl = lowDupl + 1; newDupl < highDupl; ++newDupl) {
    rightHalf.push_back(-leftHalf.back());
    leftHalf.pop_back();
    copyInsertionBlock(duplicatedConfigs[newDupl], leftHalf, leftPos, leftPos);
    copyInsertionBlock(duplicatedConfigs[newDupl], rightHalf, rightPos,
		       rightPos);
  }

  // copy the insertion block into the last duplicate
  rightHalf.push_back(-leftHalf.back());
  copyInsertionBlock(duplicatedConfigs[highDupl], rightHalf, rightPos,
		     rightPos);
}

// ***** Helper function for -computeInsConfigurations()- *****
// Divide the left insertion block in -mapIt- into all possible halves, sending
// the left half to the left and the right half to the right in
// -duplicatedConfigs-
void SortSeq::splitInsBlock_2DFirst(int secondBlockSize,
				    CycConfLevel2_t &duplicatedConfigs,
				    int currDuplNum, int revNum,
				    InsConfiguration::const_iterator &mapIt) {
  int signedOldPos, leftPos, rightPos;
  InsPosNode *insNode;
  getInsConfigParams(mapIt, revNum, signedOldPos, insNode);

  vector<int> leftHalf, rightHalf;
  getLeftRightPos(insNode, mapIt, signedOldPos, leftPos, rightPos, leftHalf);

  if (signedOldPos != leftPos)
    cerr << endl << "ERROR in SortSeq.C: splitInsBlock_2DFirst()" << endl
	 << " - the original position is not equal to the left position"
	 << endl << endl;

  int firstBlockSize = leftHalf.size() + 1;
  int firstDupl = currDuplNum * firstBlockSize * secondBlockSize;
  int newDupl, lowDupl, highDupl;
  int rowNum;

  // copy the block into the first "row" of duplicates
  highDupl = firstDupl + secondBlockSize;
  for (newDupl = firstDupl; newDupl < highDupl; ++newDupl)
    copyInsertionBlock(duplicatedConfigs[newDupl], leftHalf, leftPos, leftPos);

  // do the rest of the block copying except for the last "row"
  for (rowNum = 1; rowNum < firstBlockSize - 1; ++rowNum) {
    rightHalf.push_back(-leftHalf.back());
    leftHalf.pop_back();
    lowDupl = highDupl;
    highDupl += secondBlockSize;
    for (newDupl = lowDupl; newDupl < highDupl; ++newDupl) {
      copyInsertionBlock(duplicatedConfigs[newDupl], leftHalf, leftPos,
			 leftPos);
      copyInsertionBlock(duplicatedConfigs[newDupl], rightHalf, rightPos,
		       rightPos);
    }
  }

  // copy the block into the last "row" of duplicates
  lowDupl = highDupl;
  highDupl += secondBlockSize;
  rightHalf.push_back(-leftHalf.back());
  for (newDupl = lowDupl; newDupl < highDupl; ++newDupl)
    copyInsertionBlock(duplicatedConfigs[newDupl], rightHalf, rightPos,
		       rightPos);
}

// ***** Helper function for -computeInsConfigurations()- *****
// Divide the right insertion block in -mapIt- into all possible halves,
// sending the left half to the left and the right half to the right in
// -duplicatedConfigs-
void SortSeq::splitInsBlock_2DSecond(int firstBlockSize,
				     CycConfLevel2_t &duplicatedConfigs,
				     int currDuplNum, int revNum,
				     InsConfiguration::const_iterator &mapIt){
  int signedOldPos, leftPos, rightPos;
  InsPosNode *insNode;
  getInsConfigParams(mapIt, revNum, signedOldPos, insNode);

  vector<int> leftHalf, rightHalf;
  getLeftRightPos(insNode, mapIt, signedOldPos, leftPos, rightPos, leftHalf);

  if (signedOldPos != rightPos)
    cerr << endl << "ERROR in SortSeq.C: splitInsBlock_2DSecond()" << endl
	 << " - the original position is not equal to the right position"
	 << endl << endl;

  int secondBlockSize = leftHalf.size() + 1;
  int lowDupl = currDuplNum * firstBlockSize * secondBlockSize;
  int newDupl, highDupl;
  int colNum;

  // copy the block into the first "column" of duplicates
  highDupl = lowDupl + firstBlockSize*secondBlockSize;
  for (newDupl = lowDupl; newDupl < highDupl; newDupl += secondBlockSize)
    copyInsertionBlock(duplicatedConfigs[newDupl], leftHalf, leftPos, leftPos);

  // do the rest of the block copying except for the last "column"
  for (colNum = 1; colNum < secondBlockSize - 1; ++colNum) {
    rightHalf.insert(rightHalf.begin(), -leftHalf.front());
    leftHalf.erase(leftHalf.begin());
    ++lowDupl;
    for (newDupl = lowDupl; newDupl < highDupl; newDupl += secondBlockSize) {
      copyInsertionBlock(duplicatedConfigs[newDupl], leftHalf, leftPos,
			 leftPos);
      copyInsertionBlock(duplicatedConfigs[newDupl], rightHalf, rightPos,
		       rightPos);
    }
  }

  // copy the block into the last "column" of duplicates
  ++lowDupl;
  rightHalf.insert(rightHalf.begin(), -leftHalf.front());
  for (newDupl = lowDupl; newDupl < highDupl; newDupl += secondBlockSize)
    copyInsertionBlock(duplicatedConfigs[newDupl], rightHalf, rightPos,
		       rightPos);
}

// ***** Helper function for -computeInsConfigurations()- *****
// Copy the insertion block in -mapIt- into each of -growthFactor- number of
// duplicate InsConfigurations in -duplicatedConfigs-.
void SortSeq::spreadInsBlock(CycConfLevel2_t &duplicatedConfigs,
			     int currDuplNum, int revNum,
			     InsConfiguration::const_iterator &mapIt,
			     int growthFactor) {
  int lowDupl = currDuplNum * growthFactor;
  int highDupl = lowDupl + growthFactor - 1;
  for (int newDupl = lowDupl; newDupl <= highDupl; ++newDupl)
    copyInsertionBlock(duplicatedConfigs[newDupl], revNum, mapIt);
}

// ***** Helper function for -computeInsConfigurations()- *****
// Erase -subsetCycConfigs- entries at the last reversal which don't
// contain all the genes in the insertion subset
void SortSeq::eraseBadFinalEntries(int subsetSize,
				   CycConfLevel1_t &subsetCycConfigs) {
  if (subsetCycConfigs.empty())
    return;

  CycConfLevel1_t::reverse_iterator finalCycConf;
  finalCycConf = subsetCycConfigs.rbegin();
  CycConfLevel2_t::iterator dupIt, dupEnd;
  for (dupIt = finalCycConf->begin(), dupEnd = finalCycConf->end();
       dupIt != dupEnd; ++dupIt) {
    InsConfiguration::iterator insConfIt, insConfEnd, currIt;
    insConfIt = dupIt->begin();
    insConfEnd = dupIt->end();
    // erase insertion configuration entries with fewer genes than are in the
    // insertion subset
    while(insConfIt != insConfEnd) {
      currIt = insConfIt++;
      if (int(currIt->second.size()) < subsetSize)
	dupIt->erase(currIt);
    }
  }
}

// ***** Helper function for -computeInsConfigurations()- *****
// Eliminate -subsetCycConfigs- subpaths that terminate in an erased
// InsConfiguration entry
void SortSeq::eraseDeadSubpaths(CycConfLevel1_t &subsetCycConfigs) {
  // traverse -subsetCycConfigs- backwards from the next-to-last reversal and
  // eliminate intermediate InsConfiguration entries whose next positions have
  // all been erased
  bool removedConfig;
  int revNum, growthFactor;
  int oldDups, numDups;
  if (subsetCycConfigs.empty())
    return;
  numDups = subsetCycConfigs.back().size()-1;
  for (revNum = subsetCycConfigs.size()-2, removedConfig = true;
       removedConfig && revNum >= 0; --revNum) {
    oldDups = numDups;
    numDups = subsetCycConfigs[revNum].size()-1;
    growthFactor = (oldDups+1) / (numDups+1);
    removedConfig = false;
    for (int dupNum = 0; dupNum <= numDups; ++dupNum) {
      InsConfiguration::iterator insConfIt, insConfEnd, currIt;
      insConfIt = subsetCycConfigs[revNum][dupNum].begin();
      insConfEnd = subsetCycConfigs[revNum][dupNum].end();
      // erase insertion configuration entries that haven't got a successor in
      // the next higher reversal
      while(insConfIt != insConfEnd) {
	currIt = insConfIt++;
	if (currIt->second.empty())
	  cerr << endl << "ERROR in SortSeq.C: eraseDeadSubpaths()" << endl
	       << " - malformed InsConfiguration in -combinedCycConfigs-"
	       << endl << endl;
	int insNum = currIt->second.front();
	int currPos = currIt->first;
	if (insNum < 0) {
	  currPos = -currPos;
	  insNum = abs(insNum);
	}

	const InsPosNode *node = &(insertPos[insNum][revNum][currPos]);
	int nextPos1 = abs(node->newPos1);
	int nextPos2 = abs(node->newPos2);

	bool keepCurrent = false;
	int nextMax = (dupNum+1) * growthFactor;
	// if the current reversal is a division point, scan all the duplicates
	// the current duplicate leads to (if not, scan the same duplicate num)
	for (int nextRevDup = dupNum * growthFactor; nextRevDup < nextMax;
	     ++nextRevDup) {
	  if (subsetCycConfigs[revNum+1][nextRevDup].find(nextPos1)
	      != subsetCycConfigs[revNum+1][nextRevDup].end()
	      || ((node->twoNew)
		  && subsetCycConfigs[revNum+1][nextRevDup].find(nextPos2)
		  != subsetCycConfigs[revNum+1][nextRevDup].end())) {
	    keepCurrent = true;
	    break;
	  }
	}
	if (!keepCurrent) {
	  subsetCycConfigs[revNum][dupNum].erase(currIt);
	  removedConfig = true;
	}
      }
    }
//     if (!removedConfig && revNum > 0)
//       cerr << "finished early at reversal " << revNum << endl;
  }
}

// ***** Helper function for -computeInsConfigurations()- *****
// Copy the insertion block at -mapIt->second- into the InsConfiguration
// -newConfig-. Together -mapIt- and -revNum- provide the indices necessary
// to access the correct InsPosNode in -insertPos-
void SortSeq::copyInsertionBlock(InsConfiguration &newConfig, int revNum,
				 InsConfiguration::const_iterator &mapIt) {
  int signedOldPos;
  InsPosNode *insNode;
  getInsConfigParams(mapIt, revNum, signedOldPos, insNode);
  copyInsertionBlock(newConfig, mapIt->second, signedOldPos,
		     insNode->newPos1);
  if (insNode->twoNew)
    copyInsertionBlock(newConfig, mapIt->second, signedOldPos,
		       insNode->newPos2);
}

// ***** Helper function for -computeInsConfigurations()- *****
// Copy the insertion block -currBlock- from its previous position at
// -signedOldPos- into the InsConfiguration -newConfig- at position
// -signedNewPos-
// (The positions need to be signed in case of an in-place reversal)
void SortSeq::copyInsertionBlock(InsConfiguration &newConfig,
				 const vector<int> &currBlock,
				 int signedOldPos, int signedNewPos) {
  int newPos = abs(signedNewPos);
  vector<int> *newBlockPtr
    = &((newConfig.insert	// get pointer to insertion block at -newPos-
	 (InsConfiguration::value_type
	  (newPos, vector<int>()))).first->second);
  if (signedOldPos == signedNewPos) {	// copy -currBlock- to -newPos-
    newBlockPtr->insert(newBlockPtr->end(), currBlock.begin(),
			currBlock.end());
  }
  else {			// reverse -currBlock- and copy to -newPos-
    vector<int> revBlock = reverseGenomeSegment(currBlock);
    newBlockPtr->insert(newBlockPtr->end(), revBlock.begin(), revBlock.end());
  }
}

// ***** Helper function for -computeInsConfigurations()- *****
// Use the InsConfiguration element that -mapIt- points at to get the correct
// InsPosNode in -insertPos-, and the signed 'old' position that the inserted
// gene moves from
void SortSeq::getInsConfigParams(InsConfiguration::const_iterator &mapIt,
				 int revNum, int &signedOldPos,
				 InsPosNode* &insNode) {
  int firstIns = mapIt->second.front();
  int oldPos = mapIt->first;
  signedOldPos = (firstIns < 0) ? (-oldPos) : oldPos;
  firstIns = abs(firstIns);
  insNode = &(insertPos[firstIns][revNum][signedOldPos]);
}

// ***** Helper function for -computeInsConfigurations()- *****
// Figure out the left and right positions to move insertion block to
void SortSeq::getLeftRightPos(InsPosNode* &insNode,
			      InsConfiguration::const_iterator &mapIt,
			      int signedOldPos, int &leftPos, int &rightPos,
			      vector<int> &leftHalf) {
  if (!insNode->twoNew)
    cerr << endl << "ERROR in SortSeq.C: getLeftRightPos()" << endl
	 << " - although we are at a division point, the entry in 'insertPos'"
	 << " only goes one" << endl
	 << "   way" << endl << endl;

  leftPos = insNode->newPos1;
  rightPos = insNode->newPos2;
  if (abs(leftPos) > abs(rightPos)) {
    int temp = leftPos;
    leftPos = rightPos;
    rightPos = temp;
  }

  if (leftPos == signedOldPos)
    leftHalf = mapIt->second;
  else
    leftHalf = reverseGenomeSegment(mapIt->second);
}

ostream& operator<<(ostream& output, SortSeq& s) {
  output << "THIS DOESN'T REALLY WORK, BUT COULD EASILY BE FIXED" << endl;
  output << "Reversals:" << endl;
  int numRevs = s.revs.size();
  for(int i=0; i < numRevs; ++i)
    output << s.revs[i].getLeft() << "," << s.revs[i].getRight() << "  ";

  output << endl;

  return output;
}

#endif // SORTING_SEQUENCE
