#ifndef SORTSEQ_H
#define SORTSEQ_H
/*****************************************************************************
 *  FILE   : SortSeq.h
 *  AUTHOR : Joel V. Earnest-DeYoung
 *  DATE   : 11/2003
 *****************************************************************************
 *
 * Class for identifying sorting sequences between unequal genomes
 *
 *  Public Interface Methods:
 *   addRev()
 *
 * I'm using Krister and Mark's convention that
 *   deletions are genes in genome 1 but not in genome 2, and 
 *   insertions are genes in genome 2 but not genome 1.
 * What I call consensus genes are genes present in both genomes.
 * The reversal sequence -revs- converts the identity permutation into the
 *  consensus sequence -consensusG2-
 *
 *****************************************************************************/

// The SORTING_SEQUENCE flag turns on compilation of the code that records
// the sequence of operations from genome 1 to genome 2 as passed to multiMED
// #define SORTING_SEQUENCE

// The rest of this file is compiled only if the SORTING_SEQUENCE flag is
// turned on
#ifdef SORTING_SEQUENCE

// #define OVERLY_OPTIMISTIC_INSERTION_IDEA

#define GENOME1_ONLY	0
#define GENOME2_ONLY	1
#define BOTH_GENOMES	2

/*****************************************************************************
 * PLATFORM SPECIFIC STUFF
 *****************************************************************************/
#ifdef _MSC_VER	// MS Visual C++
#pragma warning (disable : 4786)	// Turn off dumb Visual C++ warning
#pragma warning (disable : 4503)	// Turn off dumb Visual C++ warning
#endif //_MSC_VER
/*****************************************************************************/

#include <ctime>
#include <iomanip>
#include <iostream>
#include <vector>
#include <set>
#include <utility>
#include <map>
#include <ext/hash_map>
#include <ext/hash_set>
#include "../graph/Pair.h"

#ifdef _MSC_VER
using namespace std;
#endif

class InsPosNode {
 public:
  InsPosNode();
  InsPosNode(int p1);
  InsPosNode(int p1, int p2);
  bool twoNew;
  int newPos1;
  int newPos2;
};

struct HurdleInfo {
  HurdleInfo(int l, int r, bool s);
  int left;
  int right;
  bool isSuperHurdle;
};

class SortSeq;
std::ostream& operator<<(std::ostream& output, SortSeq& s);

class SortSeq {
  friend std::ostream& operator<<(std::ostream& output, SortSeq& s);

 public:
  SortSeq();
  ~SortSeq();
  void clear();			// clear all data
  void setGenome1(int *seq, int len);
  void setGenome2(int *seq, int len);
  void setRenamedG2(int *seq, int len);
  void setConsensusG2(int *seq, int len);
  int getConsensusSize() const;
  void mapConsensusToRenamed(int consGene, int renGene);
				// map a gene in consensusG2 to renamedG2

  void addRev(Pair rev);	// add a reversal
  void addUnorRev(Pair rev);	// add an unoriented reversal (hurdle clearing)
  void addHurdle(int left, int right, bool isSuper);
				// add a hurdle
  void setMaximalHurdle(int innerLeft, int innerRight);
				// set inner extrema of maximal hurdle
  void setFortressRevPoint(int neoInnerLeft);
				// set left side of fortress-clearing reversal
				// (see description in SortSeq.C for details)

  void addInsert(Pair insert);	// add an insertion
  void addDelete(Pair del);	// add a deletion
  void setInsMatrix(int **iMat, int rows, int cols);
				// set the insertion matrix
  void addInsSubset(int *set);
				// add an insertion subset
  void computeConsensusPositions();

  void computeStickyInsMap();	// associate consensus genes and insertions
				// calculate all consensus gene positions
  void computeInsConfigurations();
				// calculate all configurations of insertions
  void computeDelPositions();	// calculate all possible delete positions
  int getNumRevs();		// return number of reversals
  int getNumInserts();		// return number of insertions
  int getNumDels();		// return number of deletions
  int getNumHurdles();		// return number of hurdles
  bool isFortress();		// true if consensusG2 is a fortress

  void printInsMatrix();
  void printInsSubsets();
  void printInsFinalLocs();
  void printGenomes();
  void printExecutedReversals();
  void printOriginalHurdles();

  void getMidpointGenome(std::vector<int> &midGenome, std::vector<int> &source);
  void getIntermediateGenome(int midRev, std::vector<int> &midGenome,
			     std::vector<int> &source);
  void getMidpointGenome(std::vector<int> &midGenome);
  void getIntermediateGenome(int midRev, std::vector<int> &midGenome);

  void getSimpleRevDist(std::map<int,int> &revSizeDistrib) const;


  // ADD A ROUTINE TO PRINT THE SORTING SEQUENCE
 private:
  void setGenome(std::vector<int> &genome, int *seq, int len);
				// set the value of a (private) genome sequence
  void reverseInsert(int pos, Pair &rev, int insNum, int revNum);
				// helper function for -computeInsPositions()-

  // maps insertion blocks to the position they lie left of in the consensus
  // genome. All insertions in an InsConfiguration should be in the same
  // insSubset
  typedef std::map<int, std::vector<int> > InsConfiguration;
  typedef InsConfiguration::value_type IConfVal_t;

  void getInsConfigurations(
    int midRev, std::vector< std::vector<InsConfiguration> > &insConfigs);
  void chooseConfigRandomly(InsConfiguration &newConf,
			    const std::vector<InsConfiguration> &confGroup);

  void printConfigs();		// print the -combinedConfigs-, -mutExConfigs-,
				// -combinedCycConfigs-, and -mutExCycConfigs-
				// data structures

  void printInsertPos();	// print the -insertPos- data structure
  void printDeletePos();	// print the -deletePos- data structure

  std::vector<int> genome1;		// original genome 1 sequence
  std::vector<int> genome2;		// original genome 2 sequence
  std::vector<int> renamedG2;	// duplication-free, renamed genome 2 sequence
  std::vector<int> consensusG2;	// renamed genome 2 sequence without insertions
  __gnu_cxx::hash_map<int,int> renamedToGenome2;
				// maps renamedG2 genes to genome2
  __gnu_cxx::hash_map<int,int> consensusToRenamed;
				// maps consensusG2 genes to renamedG2 genes

  std::vector<Pair> revs;	// turns identity permutation into consensusG2
  std::vector<Pair> dels;
  std::vector<Pair> inserts;

  std::vector< std::vector<int> > revConsensusG1;
				// the state of "consensusG1" after each
				// reversal
  std::vector< std::vector<int> > revConsensusG2;
				// the state of consensusG2 after each reversal

  std::vector< std::vector<int> > insSubsets;
				// sets of inserted genes which must be
				// together in genome2

  typedef __gnu_cxx::hash_map<int, __gnu_cxx::hash_set<int> > iisHashMap;
  iisHashMap insertionFinalLocs;// acceptable locations for insertions in
				// genome 1
  std::vector< std::vector<int> > insertMatrix;
  std::vector<HurdleInfo> hurdles;

  int firstHurdleRev, lastHurdleRev;
				// indices of 1st and last hurdle reversals in
				// revs (-1 means no hurdle revs)

  int maxHurdle;		// index of a maximal hurdle (-1 means no max
				// hurdle)
  int maxHInnerLeft;		// max index of left half of maximal hurdle
  int maxHInnerRight;		// min index of right half of maximal hurdle

  int fortressRevPoint;		// left index of fortress-clearing reversal
				// (-1 means consensusG2 is not a fortress)
				// (see description of setFortressRevPoint in
				//  SortSeq.C for more details)

  typedef std::map<int, InsPosNode> IPosLevel2_t;
  typedef std::vector<IPosLevel2_t> IPosLevel1_t;
  typedef std::vector<IPosLevel1_t> IPos_t;
  typedef IPosLevel2_t::iterator IPosLevel2_iterator;
  typedef IPosLevel1_t::iterator IPosLevel1_iterator;
  typedef IPos_t::iterator IPos_iterator;
  IPos_t insertPos;
  /*
  -insertPos- is a somewhat complex data structure
   It contains a matrix of maps with as many rows as there are inserted genes
    and as many columns as there are reversals (numIns x numRevs). For p,
    a possible position of insertion i before reversal r, the map at
    insertPos[i][r] maps the 1 or 2 positions which the reversal r could move
    insertion i to, starting from p. Additionally, the position will be stored
    as positive or negative if i has the same or opposite orientation as it did
    in genome 2, respectively.
   In -computeInsPositions()- all possible paths that each insertion could
    travel over the course of the reversals from genome 2 to genome 1 is
    computed. Then the -insertionFinalLocs- are used to keep only acceptable
    final locations for each insertion. Finally, working back from the final
    locations, the unacceptable intermediate locations are eliminated. Thus, we
    are left with all possible locations of each insertion after each reversal.
  */
                                                // maps consensusG2 genes to any
					        // insertions stuck to them
  __gnu_cxx::hash_map<int,std::vector<int> > stickyInsMap;

  typedef std::vector<InsConfiguration> CombConfLevel1_t;// indexed by reversal
  typedef std::vector<CombConfLevel1_t> CombConfigs_t;   // indexed by subset
  CombConfigs_t combinedConfigs;
  typedef std::vector<InsConfiguration> CycConfLevel2_t; // indexed by duplicate num
  typedef std::vector<CycConfLevel2_t> CycConfLevel1_t;  // indexed by reversal
  typedef std::vector<CycConfLevel1_t> CombCycConfigs_t; // indexed by subset
  CombCycConfigs_t combinedCycConfigs;

  typedef std::pair<int,int> InsertPosPair_t; // insert number and position of a
					      // division point

  typedef std::set<int> PositionSet_t;
  typedef std::set<PositionSet_t> MutExConfig_t;
  std::vector< std::vector<MutExConfig_t> > mutExConfigs;
				// indexed by subset, then by reversal
  std::vector< std::vector< std::vector<MutExConfig_t> > > mutExCycConfigs;
			// indexed by subset, then by reversal,
			// then by duplicate num
  /*
  -combinedConfigs- contains the 
  */

  typedef std::vector<int> dPosLevel2_t;
  typedef std::map<int, dPosLevel2_t> dPosLevel1_t;
  typedef std::vector<dPosLevel1_t> dPos_t;
  typedef dPosLevel2_t::iterator dPosLevel2_iterator;
  typedef dPosLevel1_t::iterator dPosLevel1_iterator;
  typedef dPosLevel1_t::value_type dPosLevel1_pair;
  typedef dPos_t::iterator dPos_iterator;
  dPos_t deletePos;
  // deletions in each position after each reversal
  //  deletePos[rev][pos] gives the deleted genes between positions pos-1 and
  //  pos in the consensus genome after reversal rev


  // **** HELPER FUNCTIONS FOR -computeInsConfigurations()- **** //
  void computeInsPositions(int insNum);
				// calculate all positions for the given insert

  // A rejoin point in -insertPos- is where an insertion moves to a particular
  //  position from two different previous positions.
  // A division point is the splitting point in -insertPos- of the two paths
  //  that lead to a rejoin point.
  // A division point is completely described by
  // - the reversal number it occurs at,
  // - the inserted gene that is divided, and
  // - the position of the gene prior to division
  // IPosCycleInfo_t maps reversals to division points.
  typedef std::set<InsertPosPair_t> IPCycLevel1_t;
  typedef std::map<int, IPCycLevel1_t> IPosCycleInfo_t; // indexed by reversal

  bool noIPosCycle(int insVal);	// check that -insertPos- doesn't have a cycle
				// at the -insVal- insertion
  void getIPosCycleInfo(IPosCycleInfo_t &iPosCycleInfo,
			const std::set<int> &insSubset);
  void printIPosCycleInfo(const IPosCycleInfo_t &iPosCycleInfo);
  void copyBackMutExConfigs(const Pair &rev, const InsConfiguration &insConfig,
			    const MutExConfig_t &lastMutExConf,
			    MutExConfig_t &newMutExConf);
  void getCombinedCycConfigs(int subsetNum,
			     const IPosCycleInfo_t &iPosCycleInfo,
			     const std::set<int> &currSubset);
  void splitInsBlock_1D(CycConfLevel2_t &duplicatedConfigs, int currDuplNum, 
			int revNum, InsConfiguration::const_iterator &mapIt);
  void splitInsBlock_2DFirst(int secondBlockSize,
			     CycConfLevel2_t &duplicatedConfigs,
			     int currDuplNum, int revNum,
			     InsConfiguration::const_iterator &mapIt);
  void splitInsBlock_2DSecond(int firstBlockSize,
			      CycConfLevel2_t &duplicatedConfigs,
			      int currDuplNum, int revNum,
			      InsConfiguration::const_iterator &mapIt);
  void spreadInsBlock(CycConfLevel2_t &duplicatedConfigs, int currDuplNum, 
		      int revNum, InsConfiguration::const_iterator &mapIt,
		      int growthFactor);
  void eraseBadFinalEntries(int subsetSize,
			    CycConfLevel1_t &subsetCycConfigs);
  void eraseDeadSubpaths(CycConfLevel1_t &subsetCycConfigs);
  void copyInsertionBlock(InsConfiguration &newConfig, int revNum,
			  InsConfiguration::const_iterator &mapIt);
  void copyInsertionBlock(InsConfiguration &newConfig,
			  const std::vector<int> &currBlock,
			  int signedOldPos, int signedNewPos);
  void getInsConfigParams(InsConfiguration::const_iterator &mapIt, int revNum,
			  int &signedOldPos, InsPosNode* &insNode);
  void getLeftRightPos(InsPosNode* &insNode,
		       InsConfiguration::const_iterator &mapIt,
		       int signedOldPos, int &leftPos, int &rightPos,
		       std::vector<int> &leftHalf);
  void getUniqueGeneConfig(const std::set<int> &uniqueSubset, int midRev,
			   InsConfiguration &currConfig);
  // ******************************************************* //
};

extern SortSeq *globalSortSequence;

#endif //ifdef SORTING_SEQUENCE
#endif //ifndef SORTSEQ_H
