/* Krister Swenson                                                 Winter 2007-8

A simple graph class that knows its Minimum Spanning Tree and can decompose
itself by the join decomposition.
It also knows its maximum clique or independent set (by brute force).
 ** uses the igraph library to compute all independent sets.
*/
#ifndef GRAPH_H
#define GRAPH_H

#include <iostream>
#include <ext/hash_map>
#include <ext/hash_set>
#include <utility>
#include <set>
#include <map>
#include <queue>
#include <vector>
#include <list>
#include <string>
#include "Vertex.h"
#include "dfmax.h"      //DIMACS code for getting maximum cliques.
#include "maxclique.h"  //Code for getting all maximal IS/cliques.
#include "miscg.h"      //Code for the maximum independent set on circle graphs.
#include "tools.h"      //Output functions for STL objects.
#include "/usr/include/igraph/igraph.h"

#define LARGEST 1000    //The largest vertex label that we will be given.




/*___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___
Type definitions:                                                             */


typedef __gnu_cxx::hash_map<int, Vertex> vertmap;     //Map label to Vertex.


  //Functions for hashing Vertex pointers:
  // ** not used at the moment.
namespace __gnu_cxx
{
template<>
struct hash<const Vertex*>
  {
  hash<unsigned> hasher;
  size_t operator()(const Vertex* v) const
    { return reinterpret_cast<size_t>(v) ; }
  };
}
struct eqVP
  {
  bool operator()(const Vertex* v1, const Vertex* v2) const
    { return *v1 == *v2; }
  };
typedef __gnu_cxx::hash_set< const Vertex*, __gnu_cxx::hash< Vertex* >, eqVP >
        vertexset;

typedef __gnu_cxx::hash_set<int> labelset;          //A set of Vertex labels.
std::ostream& operator<<(std::ostream& output, const labelset &vs);

typedef std::set<ipair> ipset;                      //A set of ipairs.
std::ostream& operator<<(std::ostream& output, const ipset &ps);


/*___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___/\___
Graph definition:                                                             */

struct MISlist;


class Graph
  {
    //Traverse the graph and print the edges.
    //  ** leaves things marked.
  friend std::ostream& operator<<(std::ostream& output, const Graph& g);

  public:
      //Create an undirected graph if -u- is true, otherwise, directed.
    Graph(bool u = true);
      //Innitialize a graph with a chord model (double occurance word).
    Graph(const std::vector<int> &cm);
      //Innitialize a graph with a chord model (double occurance word).
      // Use -chordTOinv- to set the name of the inversions.
    Graph(const std::vector<int> &cm,
          const std::vector<std::string> chordTOinv);
    Graph(const Graph &g);
    void operator=(const Graph &g);
    ~Graph();

    //Building the graph:
      //Add a single edge between -l1- and -l2-. Add the vertices if they don't
      // already exist.
    void addEdge(int l1, int l2);
    void addEdge(const Vertex *v, const Vertex *w);
      //Add an edge between existing vertices.
    void addEdgeExisting(Vertex *v, Vertex *w);
      //Remove an edge if it exists.
    void remEdge(int l1, int l2);
      //Remove an edge if it exists.
    void remEdge(Vertex *v, Vertex *w);
      //Add a single vertex labelled -l- with no edges.
    Vertex* addVertex(int l);
      //Add a single vertex named -n- and labelled -l-, with no edges.
    Vertex* addVertex(int l, const std::string &name);
      //Add a vertex labelled -l- adjacent to those in -adjlist-. Any vertex
      // that is in -adjlist- but does not exist will be created.  Ignore
      // self edges. If -checkdup- is false then don't check for duplicate
      // vertex labels.
    void addVertex(int l, const std::vector<int> &adjlist,bool checkdup = true);
      //Add a vertex labelled -l- adjacent to those in -adjset-. Any vertex
      // that is in -adjlist- but does not exist will be created.  Ignore
      // self edges. If -checkdup- is true then don't check for duplicate
      // vertex labels.
    void addVertex(int l, const __gnu_cxx::hash_set<int> &adjset,
                   bool checkdup = true);
      //Remove the vertex with label l.  Return the neighbors of -l-.
    __gnu_cxx::hash_set<int> remVertex(int l);
      //Remove all the neighbors of the given vertex.
    void remNeighbors(int l);
      //Build a bipartite graph from vertices labelled between l1-l2 to those
      // labelled r1-r2.
    void buildBipartite(int l1, int l2, int r1, int r2);
      //Build a clique among the given vertices.
    void buildClique(int v1, int v2);
      //Return the intersection of -g- and this graph.
    Graph getIntersection(const Graph &g) const;
      //Return the union of -g- and this graph.
    Graph getUnion(const Graph &g) const;


    //  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    //Querying the graph:   

      //Return true if the given graph has the same edge set as this one.
    bool operator==(const Graph &g);
      //Return true if there are no vertices in this graph.
    bool empty() const;
      //Return the number of vertices in the subgraph.
    int size() const;

      //Return true if this graph is a circle graph.
    bool isCG();
      //Get the maximum independent set of vertices.
    std::vector<int> getMIS();

      //Return true if the given edge exists in this graph (left to right).
    bool hasEdge(const ipair e) const;
    bool hasEdge(int l, int r) const;
    bool hasEdge(Vertex *v, Vertex *w) const;
      //Return true if the given vertex exists in this graph.
    bool hasVertex(int v) const;

      //Return a sublist of -l- that contains only the vertex labels that are
      // in this graph.
    std::list<int> getVertexSubset(const std::list<int> &l) const;
      //Return a chord model (double occurance word) for the given graph.
      // An empty model is returned if non exist.
    std::vector<int> getCM();

      //Return true if one vertex is reachable from the other.
    bool isReachable(Vertex &v, const Vertex &target);
      //Return a copy of the the connected components.
    std::vector<Graph> getComponents();
      //Get the subgraph comprised of the components with the given vertices.
    Graph getCompsWith(const std::list<int> &set);

      //Check whether the given chord model is consistent with this graph.
    bool checkCM(const std::list<int> &cm);
    bool checkCM(const std::vector<int> &cm);
      //Exhaustively search for a maximum independent set on the graph.
      //-goal- is the maximum size independent set to look for.
    std::list<int> getMISEX(int goal=1) const;
      //Get a list of all the independent sets in the graph.
    std::vector< std::list<int> > getAllIS();
      //Get the next maximal independent set in the graph.
    //std::list<int> getNextMalIS();
      //Get a list of all the maximal independent sets in the graph.
    std::vector< std::list<int> > getAllMalIS();
      //Get the next maximal independent set in the graph with the given
      //vertices in -whitelist- and without the given vertices in -blacklist-.
    //std::list<int> getMalISWith(const std::list<int> whitelist,
                                //const std::list<int> blacklist);
      //Reset datastructures related to maximal independent sets.
    void resetMalISinfo();
      //Return true if the given subset of vertices constitute an independent
      // set.
    bool isIS(const std::list<int> &vset) const;
      //Exhaustively search for a maximum clique on the graph.
      //-goal- is the maximum size clique to look for.
    std::list<int> getMCEX(int goal=1) const;
      //Find one IS from another IS by replacing in -is- a vertex at or after
      //  the one pointed to by -p- with one not in -is-.
      //Return true and have -p- point to the replaced value if a valid
      //  substitution was made.
    bool ISreplace(std::list<int> &is, std::list<int>::iterator &p) const;




  private:
      //Innitialize a graph with a chord model (double occurance word).
      // Use -chordTOinv- to set the name of the inversions.
    void init(const std::vector<int> &cm,
              const std::vector<std::string> chordTOinv);
    void copy(const Graph &g);
    //std::ostream& printGraph(std::ostream& output, Vertex *v);
      //Recursively decompose the graph with the join decomposition.
      //  Return true if this subgraph could be decomposed.
    bool decompose(std::vector<Graph> &d);
      //  ** assume that the graph is a single component.
    bool decompose(std::vector<Graph> &d, Graph &g, ipset &edgeS);
      //Return a split of this graph by the join decomposition.
      //In the new graphs the marker vertex will be labelled -l-.
    std::vector<Graph> split(ipset &testset, int &l);
      //Put each edge of the connected component connected to -v- into -s-.
      //  ** recursive
    void sEdges(Vertex &v, ipset &s);
      //Update the set of edges to be checked so that the edges in given
      // bipartite vertex parition will connect only to -l-.
    void updateEdgeSet(ipset &checkset, const labelset b1, const labelset b2,
                       int l);
      //Test for a split with -x- and -z- on one side and -y- on the other.
      // If there is a split then fill the sets -V1- and -V2- with vertex labels
      // and -V1B- and -V2B- with those that participate in the bipartite graph.
    bool testSplit(labelset &V1, labelset &V2, labelset &V1B, labelset &V2B,
                   Vertex &x, Vertex &z, Vertex &y);
      //Complement the graph in the neighborhood of -v-.
    void complement(const Vertex &v);
      //Return a subgraph of this graph on only those vertices in -set-.
    Graph getSubgraph(const labelset &set);
      //Recursively build the connected subraph from marked vertices.
      //Unmark them.
    void buildConSubgraph(Graph &g, Vertex &v);

      //Get a chord model (double occurance word) for the given graph.
      // Return false if there isn't a circle graph representation.
    bool getCG(std::list<int> &retval);
      //Recursively fill in the CM_ENUM array (of the appropriate size) with
      //all possible chord models.
    void fillCMenum(std::vector<int> &bucket);
    void fillCMenumh(std::list<int> &v, std::vector<int> &bucket);
      //Innitialize the buckets to each have value 2.
    std::vector<int> innitBucket(int size);
      //Reverse the subword given by the interval l to r.
    void reverseCM(std::list<int> &cm, std::list<int>::iterator l,
                                       std::list<int>::iterator r) const;
      //Add the given "vertex" to the word before l and after r.
    void addToCM(int v, std::list<int> &cm, std::list<int>::iterator l,
                                            std::list<int>::iterator r) const;
      //Remove the characters from the word before l and after r.
    void remFromCM(std::list<int> &cm, std::list<int>::iterator l,
                                       std::list<int>::iterator r) const;
      //Merge the give chord models through the markers.
    std::vector<int> mergeCMs(std::vector< std::list<int> > &cmlist);
      //Return the second highest marker in the given double occurance word.
    int nextHighestMark(const std::list<int> &doc) const;



      //Check whether the given vertex is consistent with -cc-.
    bool checkVertex(int v, const std::vector<int> &cc) const;

      //Return the number of chord models of the given size (# chords).
    int getEnumSize(int size);



      //Mark all vertices reachable from -v-.
    void markComponent(Vertex &v);
      //Unmark v's connected component induced by considering only marked
      //vertices.
    void clean(Vertex &v);
      //Unmark every vertex in the graph.
    void clean();

      //Print every vertex and a new line.
    void printVertex(std::ostream& output) const;
      //Print the decomposition given in -d-.
    void printDecomposition(std::ostream& output,
                            const std::vector<Graph> &d) const;


    vertmap vset;               //Vertices indexed by label.
    bool undirected;            //Marks the type of graph.
    int largest;                //The largest vertex label we will be given.
                                //Holds chord model enumeration for small cms.
    std::vector< std::vector < std::list <int> > > CM_ENUM;

    std::vector< std::list<int> > maxsets; //Vector of maximal ISs/cliques.
    bool setinit;                          //Flag for initialization of maxsets.
  };


#endif
